/******************************************************************************
* IritTrapErrors.cpp - implementation of functions for giving messages        *
* regarding type of error that occured in Irit.                               *
*******************************************************************************
* (C) Gershon Elber, Technion, Israel Institute of Technology                 *
******************************************************************************/

#include "NCSimMain.h"

#include "irit_sm.h"
#include "iritprsr.h"
#include "bool_lib.h"
#include "cagd_lib.h"
#include "geom_lib.h"
#include "symb_lib.h"
#include "misc_lib.h"
#include "mvar_lib.h"
#include "trim_lib.h"
#include "triv_lib.h"
#include "trng_lib.h"
#include "user_lib.h"
#include "grap_lib.h"

#pragma warning (disable: 4311)

#define MAX_CONSOLE_LINES 	        500

static void LclBoolFatalError(BoolFatalErrorType ErrID);

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Make error message box.		                                     M
*                                                                            *
* PARAMETERS:                                                                M
*   Msg:       Error message.                                                M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   IGIritError                                                              M
*****************************************************************************/
void IGIritError(const char *Msg)
{	
    NCSimMF -> Printf(NCSIM_LOG_ERROR, "Irit Error: %s\n", "Error");
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Traps Bool_lib errors right here. Call back function of bool_lib.	     *
*                                                                            *
* PARAMETERS:                                                                *
*   ErrID:    Error number in bool_lib library.                              *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void LclBoolFatalError(BoolFatalErrorType ErrID)
{
    const char
	*Str = BoolDescribeError(ErrID);

    NCSimMF -> Printf(NCSIM_LOG_ERROR,
		      "%s%s\n", "Irit BOOL lib: ", Str);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Traps Cagd_lib errors right here. Call back function of cagd_lib.	     M
*                                                                            *
* PARAMETERS:                                                                M
*   ErrID:    Error number in cagd_lib library.                              M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            M
* KEYWORDS:                                                                  M
*   CagdFatalError                                                           M
*****************************************************************************/
void CagdFatalError(CagdFatalErrorType ErrID)
{
    const char
	*Str = CagdDescribeError(ErrID);

    NCSimMF -> Printf(NCSIM_LOG_ERROR, "Irit CAGD Lib: %s\n", Str);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Traps Geom_lib errors right here. Call back function of geom_lib.	     M
*                                                                            *
* PARAMETERS:                                                                M
*   ErrID:    Error number in geom_lib library.                              M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            M
* KEYWORDS:                                                                  M
*   GeomFatalError                                                           M
*****************************************************************************/
void GeomFatalError(GeomFatalErrorType ErrID)
{    
    const char
	*Str = GeomDescribeError(ErrID);

    NCSimMF -> Printf(NCSIM_LOG_ERROR, "Irit GEOM Lib: %s\n", Str);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
* Default trap for Irit parser errors.					     M
*   This function prints the provided error message and dies.		     M
*                                                                            *
* PARAMETERS:                                                                M
*   ErrID:      Error type that was raised.                                  M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   IPFatalError, error handling                                             M
*****************************************************************************/
void IPFatalError(IPFatalErrorType ErrID)
{ 
    NCSimMF -> Printf(NCSIM_LOG_ERROR, "Irit PRSR Lib: %s\n",
		      IPDescribeError(ErrID));
}

/*****************************************************************************
* DESCRIPTION:                                                               M
* Trap Mdl_lib errors right here. Provides a default error handler for the   M
* mdl library. Gets an error description using MdlDescribeError, prints it   M
* and exit the program using exit.                                           M
*                                                                            *
* PARAMETERS:                                                                M
*   ErrID:      Error type that was raised.                                  M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   MdlFatalError, error handling                                            M
*****************************************************************************/
void MdlFatalError(MdlFatalErrorType ErrID)
{
    const char
	*Str = MdlDescribeError(ErrID);

    NCSimMF -> Printf(NCSIM_LOG_ERROR, "Irit MVAR Lib: %s\n", Str);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
* Trap Misc_lib errors right here. Provides a default error handler for the  M
* misc library. Gets an error description using MiscDescribeError, prints it M
* and exit the program using exit.                                           M
*                                                                            *
* PARAMETERS:                                                                M
*   ErrID:      Error type that was raised.                                  M
*   ErrDesc:    Possibly, an additional description on error.                M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   MiscFatalError, error handling                                           M
*****************************************************************************/
void MiscFatalError(MiscFatalErrorType ErrID, char *ErrDesc)
{
    const char
	*Str = MiscDescribeError(ErrID);

    NCSimMF -> Printf(NCSIM_LOG_ERROR, "Irit MISC Lib: %s, %s\n",
		      Str, ErrDesc);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Traps Mvar_lib errors right here. Call back function of trng_lib.	     M
*                                                                            *
* PARAMETERS:                                                                M
*   ErrID:    Error number in trng_lib library.                              M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            M
* KEYWORDS:                                                                  M
*   MvarFatalError, error handling                                           M
*****************************************************************************/
void MvarFatalError(MvarFatalErrorType ErrID)
{
    const char
	*Str = MvarDescribeError(ErrID);

    NCSimMF -> Printf(NCSIM_LOG_ERROR, "Irit MVAR Lib: %s\n", Str);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Traps Symb_lib errors right here. Call back function of symb_lib.	     M
*                                                                            *
* PARAMETERS:                                                                M
*   ErrID:    Error number in symb_lib library.                              M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            M
* KEYWORDS:                                                                  M
*   SymbFatalError, error handling                                           M
*****************************************************************************/
void SymbFatalError(SymbFatalErrorType ErrID)
{
    const char
	*Str = SymbDescribeError(ErrID);

    NCSimMF -> Printf(NCSIM_LOG_ERROR, "Irit SYMB Lib: %s\n", Str);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Traps Trim_lib errors right here. Call back function of trim_lib.	     M
*                                                                            *
* PARAMETERS:                                                                M
*   ErrID:    Error number in trim_lib library.                              M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            M
* KEYWORDS:                                                                  M
*   TrimFatalError, error handling                                           M
*****************************************************************************/
void TrimFatalError(TrimFatalErrorType ErrID)
{
    const char
	*Str = TrimDescribeError(ErrID);

    NCSimMF -> Printf(NCSIM_LOG_ERROR, "Irit TRIM Lib: %s\n", Str);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Traps Triv_lib errors right here. Call back function of triv_lib.	     M
*                                                                            *
* PARAMETERS:                                                                M
*   ErrID:    Error number in triv_lib library.                              M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            M
* KEYWORDS:                                                                  M
*   TrivFatalError, error handling                                           M
*****************************************************************************/
void TrivFatalError(TrivFatalErrorType ErrID)
{
    const char
	*Str = TrivDescribeError(ErrID);

    NCSimMF -> Printf(NCSIM_LOG_ERROR, "Irit TRIV Lib: %s\n", Str);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Traps Trng_lib errors right here. Call back function of trng_lib.	     M
*                                                                            *
* PARAMETERS:                                                                M
*   ErrID:    Error number in trng_lib library.                              M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            M
* KEYWORDS:                                                                  M
*   TrngFatalError, error handling                                           M
*****************************************************************************/
void TrngFatalError(TrngFatalErrorType ErrID)
{
    const char
	*Str = TrngDescribeError(ErrID);

    NCSimMF -> Printf(NCSIM_LOG_ERROR, "Irit TRNG Lib: %s\n", Str);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Traps User_lib errors right here. Call back function of user_lib.	     M
*                                                                            *
* PARAMETERS:                                                                M
*   ErrID:    Error number in user_lib library.                              M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            M
* KEYWORDS:                                                                  M
*   UserFatalError                                                           M
*****************************************************************************/
void UserFatalError(UserFatalErrorType ErrID)
{
    const char
	*Str = UserDescribeError(ErrID);

    NCSimMF -> Printf(NCSIM_LOG_ERROR, "Irit USER Lib: %s\n", Str);
}
