/******************************************************************************
* NCSimOGL.cpp - Main entry point of OGL drawing class in NCSim.	      *
*******************************************************************************
* (C) Gershon Elber					                      *
*******************************************************************************
* Written by Gershon Elber, Dec 2006.					      *
******************************************************************************/

#include "iritprsr.h"
#include "attribut.h"

#include "NCSimMain.h"
#include "NCSimOGL.h"

#include "RealSpinCtrl.h"

#ifdef IRIT_DOUBLE
#   define OGL_GL_VRTX		glVertex3d
#   define OGL_GL_VRTX_V	glVertex3dv
#   define OGL_GL_NORMAL	glNormal3d
#   define OGL_GL_NORMAL_V	glNormal3dv
#else
#   define OGL_GL_VRTX		glVertex3f
#   define OGL_GL_VRTX_V	glVertex3fv
#   define OGL_GL_NORMAL	glNormal3f
#   define OGL_GL_NORMAL_V	glNormal3fv
#endif /* IRIT_FLOAT */

#define SCALE_FACTOR_MODIFIER	0.01

#define NCSIM_GET_ZBUF_DISP_LIST(GridX, GridY) \
	(OGLDispListBase + NCSIM_OGL_DISP_LIST_ZBUF + \
	 (GridY) * NCSIM_MAX_ZBUF_DISP_LIST + (GridX))

enum { /* Open GL Display list indices. */
    NCSIM_OGL_DISP_LIST_NONE = 0,
    NCSIM_OGL_DISP_LIST_GNRL,
    NCSIM_OGL_DISP_LIST_GPATH,
    NCSIM_OGL_DISP_LIST_STOCK,
    NCSIM_OGL_DISP_LIST_PART,
    NCSIM_OGL_DISP_LIST_TOOL,

    NCSIM_OGL_DISP_LIST_ZBUF = 10,

    NCSIM_OGL_DISP_LIST_RANGE = 110 /* Allow 100 display lists in Zbuffer. */
};

static void IGDrawOneObject(IPObjectStruct *PObj, IrtHmgnMatType Mat);
static void RenderOneNCZBufferTriangle(IrtRType X1,
				       IrtRType Y1,
				       IrtRType Z1,
				       IrtRType X2,
				       IrtRType Y2,
				       IrtRType Z2,
				       IrtRType X3,
				       IrtRType Y3,
				       IrtRType Z3);
static void RenderNCZBufferQuadStrip(IrtRType X1,
				     IrtRType Y1,
				     IrtRType Z1,
				     IrtRType X2,
				     IrtRType Y2,
				     IrtRType Z2,
				     int LastQuad);
static void RealSpinUpdateValCB(double Val, wxWindowID Id);

bool NCSimOGLClass::GlblOGLInitialized = false;

/*****************************************************************************
* AUXILIARY:								     *
*    Event table for NCSimOGLClass					     *
*****************************************************************************/
BEGIN_EVENT_TABLE(NCSimOGLClass, wxGLCanvas)
    EVT_SIZE(NCSimOGLClass::OnSize)
    EVT_PAINT(NCSimOGLClass::OnPaint)
    EVT_ERASE_BACKGROUND(NCSimOGLClass::OnEraseBackground)
    EVT_KEY_DOWN(NCSimOGLClass::OnKeyDown)
    EVT_KEY_UP(NCSimOGLClass::OnKeyUp)
    EVT_MOUSE_EVENTS(NCSimOGLClass::OnMouseEvent)
END_EVENT_TABLE()

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Constructor of the OGL drawing class.                                    M
*                                                                            *
* PARAMETERS:                                                                M
*   Parent:   Window of the OGL drawing window.                              M
*   Id:       Id assigned to window.                                         M
*   Pos:      Position of window.                                            M
*   Size:     Size of window.                                                M
*   Style:    OGL Window modifier styles.                                    M
*   Name:     Windows class name.                                            M
*                                                                            *
* RETURN VALUE:                                                              M
*   NoValue                                                                  M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimOGLClass::NCSimOGLClass                                             M
*****************************************************************************/
NCSimOGLClass::NCSimOGLClass(wxWindow *Parent,
			     wxWindowID Id,
			     const wxPoint &Pos,
			     const wxSize &Size,
			     long Style,
			     const wxString &Name):
	wxGLCanvas(Parent,
		   Id,
		   NULL,
		   Pos,
		   Size,
		   Style | wxFULL_REPAINT_ON_RESIZE,
		   Name),
	ScrnViewportBias(-1),
	ScrnWidth(-1),
	ScrnHeight(-1),
	LastKey(0),
	MouseX(-1),
	MouseY(-1),
	MouseKey(-1),
	LastEvent(IG_EVENT_NONE),
	OGLDispListBase(-1),
	ForceDispListBuild(true),
	DrawAxes(true),
	DrawTool(true),
	DrawToolPath(true),
	DrawStock(true),
	DrawFinalPart(true),
	DrawOtherGeom(true),
	SkipFrameRate(10),
	SkippedFrames(10),
        AxesSize(100.0),
	OGLParams(NULL)
{
    NCSIM_DEBUG_FUNC_ENTRY

    int i, j;

    /* Explicitly create a new rendering context instance for this canvas. */
    Context = new wxGLContext(this);
    /* Test if OGL context is properly initialized - not well documented.  */
    if (Context == NULL) {
        fprintf(stderr, "Open GL context creation failed\n");
	exit(1);
    }

    SetCurrent(*Context);

    /* Must select a function to draw polys. */
    IGDrawPolyFuncPtr = IGDrawPoly;

    /* A verification we are not invading other name spaces. */
    assert(wxID_NCSIM_OGL_PARAM_LAST - wxID_NCSIM_OGL_PARAMS <
						   wxID_NCSIM_OGL_PARAMS_SIZE);
   
    for (i = 0; i < NCSIM_OGL_DISP_LIST_ZBUF; i++) {
        for (j = 0; j < NCSIM_OGL_DISP_LIST_ZBUF; j++)
	    ZBufDispListActive[i][j] = -1;	  /* Mark as uninitialized. */
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Constructor of the OGL drawing class.                                    M
*                                                                            *
* PARAMETERS:                                                                M
*   Parent:   Window of the OGL drawing window.                              M
*   Other:    context to share OGL resources with.                           M
*   Id:       Id assigned to window.                                         M
*   Pos:      Position of window.                                            M
*   Size:     Size of window.                                                M
*   Style:    OGL Window modifier styles.                                    M
*   Name:     Windows class name.                                            M
*                                                                            *
* RETURN VALUE:                                                              M
*   NoValue                                                                  M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimOGLClass::NCSimOGLClass                                             M
*****************************************************************************/
NCSimOGLClass::NCSimOGLClass(wxWindow *Parent,
			     const NCSimOGLClass *Other,
			     wxWindowID Id,
			     const wxPoint &Pos,
			     const wxSize &Size,
			     long Style,
			     const wxString &Name):
	wxGLCanvas(Parent,
		   Id,
		   NULL,
		   Pos,
		   Size,
		   Style | wxFULL_REPAINT_ON_RESIZE,
		   Name),
	ScrnViewportBias(-1),
	ScrnWidth(-1),
	ScrnHeight(-1),
	LastKey(0),
	MouseX(-1),
	MouseY(-1),
	MouseKey(-1),
	LastEvent(IG_EVENT_NONE),
	OGLDispListBase(-1),
	ForceDispListBuild(true),
	DrawAxes(true),
	DrawTool(true),
	DrawToolPath(true),
	DrawStock(true),
	DrawFinalPart(true),
	DrawOtherGeom(true),
	SkipFrameRate(10),
	SkippedFrames(10),
        AxesSize(100.0),
	OGLParams(NULL)
{
    NCSIM_DEBUG_FUNC_ENTRY

    int i, j;

    /* Explicitly create a new rendering context instance for this canvas. */
    Context = new wxGLContext(this);
    SetCurrent(*Context);

    /* A verification we are not invading other name spaces. */
    assert(wxID_NCSIM_OGL_PARAM_LAST - wxID_NCSIM_OGL_PARAMS <
						   wxID_NCSIM_OGL_PARAMS_SIZE);

    for (i = 0; i < NCSIM_OGL_DISP_LIST_ZBUF; i++) {
        for (j = 0; j < NCSIM_OGL_DISP_LIST_ZBUF; j++)
	    ZBufDispListActive[i][j] = -1;	  /* Mark as uninitialized. */
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Class destructor                                                         M
*                                                                            *
* PARAMETERS:                                                                M
*   None                                                                     M
*                                                                            *
* RETURN VALUE:                                                              M
*   NoValue                                                                  M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimOGLClass::~NCSimOGLClass                                            M
*****************************************************************************/
NCSimOGLClass::~NCSimOGLClass()
{
    NCSIM_DEBUG_FUNC_ENTRY

    if (OGLParams != NULL)
	delete OGLParams;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Converts screen coordinates to object space coordinates.                 M
*                                                                            *
* PARAMETERS:                                                                M
*   Sx, Sy:   Input screen coordinates.                                      M
*   Ix, Iy:   Output object space coordinates.                               M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimOGLClass::Screen2ObjSpcCoords                                       M
*****************************************************************************/
void NCSimOGLClass::Screen2ObjSpcCoords(int Sx,
				        int Sy,
				        IrtRType *Ox,
				        IrtRType *Oy)
{
    NCSIM_DEBUG_FUNC_ENTRY

    NCSimulationClass
	*CrntSim = NCSimMF -> CrntSim;
    IrtRType
	MaxSize = IRIT_MAX(ScrnWidth, ScrnHeight);

    if (ScrnWidth > ScrnHeight)
        Sy += ScrnViewportBias;
    else
        Sx += ScrnViewportBias;

    /* Mapping to 0, 1. */
    *Ox = Sx / MaxSize;
    *Oy = 1.0 - Sy / MaxSize;

    /* Mapping to -1, 1. */
    *Ox = 2.0 * *Ox - 1.0;
    *Oy = 2.0 * *Oy - 1.0;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Pop up the rendering-parameters' setting dialog.                         M
*                                                                            *
* PARAMETERS:                                                                M
*   None                                                                     M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimOGLClass::SetOGLParams                                              M
*****************************************************************************/
void NCSimOGLClass::SetOGLParams()
{
    NCSIM_DEBUG_FUNC_ENTRY

    if (OGLParams == NULL)
        OGLParams = new NCSimOGLParamClass(this, NCSimMF, wxID_ANY,
					   wxT("Rendering Parameters"));

    OGLParams -> Show();
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Restart OGL state for a new simulation.                                  M
*                                                                            *
* PARAMETERS:                                                                M
*   None                                                                     M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimOGLClass::NCSimOGLInitSim                                           M
*****************************************************************************/
void NCSimOGLClass::NCSimOGLInitSim()
{
    NCSIM_DEBUG_FUNC_ENTRY

    int i, j;

    if (NCSimMF -> UseOGLDispLists) {
        for (i = 0; i < NCSIM_OGL_DISP_LIST_ZBUF; i++) {
	    for (j = 0; j < NCSIM_OGL_DISP_LIST_ZBUF; j++)
	        ZBufDispListActive[i][j] = -1;    /* Mark as uninitialized. */
	}
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Update OGL state when the current simulation page has been changed.      M
*                                                                            *
* PARAMETERS:                                                                M
*   None                                                                     M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimOGLClass::NCSimPageChange                                           M
*****************************************************************************/
void NCSimOGLClass::NCSimPageChange()
{
    NCSIM_DEBUG_FUNC_ENTRY

    NCSimulationClass
	*CrntSim = NCSimMF -> CrntSim;

    ForceDispListBuild = true;

    /* Make sure we recreate the tesselation, if has one. */
    if (CrntSim -> PObjGnrl != NULL)
        IGActiveFreePolyIsoAttribute(CrntSim -> PObjGnrl,
				     TRUE, TRUE, FALSE, FALSE);

    if (CrntSim -> GPathGeom != NULL)
        IGActiveFreePolyIsoAttribute(CrntSim -> GPathGeom,
				     TRUE, TRUE, FALSE, FALSE);

    if (CrntSim -> StockWireGeom != NULL)
        IGActiveFreePolyIsoAttribute(CrntSim -> StockWireGeom,
				     TRUE, TRUE, FALSE, FALSE);

    if (CrntSim -> FinalPartGeom != NULL)
        IGActiveFreePolyIsoAttribute(CrntSim -> FinalPartGeom,
				     TRUE, TRUE, FALSE, FALSE);

    if (CrntSim -> Tools[CrntSim -> ToolIdx].Geom != NULL)
        IGActiveFreePolyIsoAttribute(CrntSim -> Tools[CrntSim -> ToolIdx].Geom,
				     TRUE, TRUE, FALSE, FALSE);

    Refresh();
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Draw one object.                                                         *
*                                                                            *
* PARAMETERS:                                                                *
*   PObj: N.S.F.I.                                                           *
*   Mat:  N.S.F.I.                                                           *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void IGDrawOneObject(IPObjectStruct *PObj, IrtHmgnMatType Mat)
{
    NCSIM_DEBUG_FUNC_ENTRY

    IGDrawObject(PObj);
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Render one object.  If it has a display list - use it.  Otherwise.       *
* construct one now on the fly.                                              *
*                                                                            *
* PARAMETERS:                                                                *
*   PObj:    Object to render.                                               *
*   OglDispListIndex:  The index of this object's display list.              *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
void NCSimOGLClass::RenderOneObject(IPObjectStruct *PObj,
				    int OGLDispListIndex)
{
    NCSIM_DEBUG_FUNC_ENTRY

    IrtHmgnMatType Mat;
    int OGLDispList = AttrGetObjectIntAttrib(PObj, "OGLDL");

    if (ForceDispListBuild || IP_ATTR_IS_BAD_INT(OGLDispList)) {
        MatGenUnitMat(Mat);

        glNewList(OGLDispListBase + OGLDispListIndex,
		  GL_COMPILE_AND_EXECUTE);

	IPTraverseObjListHierarchy(PObj, Mat, IGDrawOneObject);

	glEndList();

	AttrSetObjectIntAttrib(PObj, "OGLDL",
			       OGLDispListBase + OGLDispListIndex);
    }
    else {
        glCallList(OGLDispList);
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Remove all OGL info from the given object that is about to be deleted.   *
*                                                                            *
* PARAMETERS:                                                                *
*   PObj:    Object to free all OGL related info.                            *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
void NCSimOGLClass::CleanOGLObject(IPObjectStruct *PObj)
{
    NCSIM_DEBUG_FUNC_ENTRY

    int OGLDispList = AttrGetObjectIntAttrib(PObj, "OGLDL");

    if (!IP_ATTR_IS_BAD_INT(OGLDispList)) {
        glDeleteLists(OGLDispList, 1);

	AttrFreeObjectAttribute(PObj, "OGLDL");
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Delete all OGL info regarding the Zbuffer display lists.		     *
*                                                                            *
* PARAMETERS:                                                                *
*   None			                                             *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
void NCSimOGLClass::CleanOGLZBuf()
{
    NCSIM_DEBUG_FUNC_ENTRY

    int i, j, GridSizeX, GridSizeY, GridCellSizeX, GridCellSizeY;
    NCSimulationClass
	*CrntSim = NCSimMF -> CrntSim;

    CrntSim -> ZBufferGetGridSize(&GridSizeX, &GridSizeY,
				  &GridCellSizeX, &GridCellSizeY);

    if (NCSimMF -> UseOGLDispLists) {
        for (i = 0; i < GridSizeX; i++) {
	    for (j = 0; j < GridSizeY; j++) {
	        glDeleteLists(ZBufDispListActive[i][j], 1);
	    }
	}
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Handles the rendering of the stock side and bottom walls.                M
*                                                                            *
* PARAMETERS:                                                                M
*   DrawSides:  True to draw the sides as well, false to draw just a bottom. M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimOGLClass::RenderStockBottomWalls                                    M
*****************************************************************************/
void NCSimOGLClass::RenderStockBottomWalls(bool DrawSides)
{
    NCSIM_DEBUG_FUNC_ENTRY

    NCSimulationClass
	*CrntSim = NCSimMF -> CrntSim;
    GMBBBboxStruct
	*B = &CrntSim -> StockBBox;
    int i, j, k;
    IrtPtType Crnrs[8];
    IrtVecType N;

    /* Create the eight corners of the box. */
    for (k = 0; k < 2; k++) {
	for (j = 0; j < 2; j++) {
	    for (i = 0; i < 2; i++) {
	        int Idx = i + j * 2 + k * 4;

		Crnrs[Idx][0] = i == 0 ? B -> Min[0] : B -> Max[0];
		Crnrs[Idx][1] = j == 0 ? B -> Min[1] : B -> Max[1];
		Crnrs[Idx][2] = k == 0 ? B -> Min[2] : B -> Max[2];
	    }
	}
    }

    IRIT_VEC_RESET(N);

    N[2] = 1.0;
    RenderOneRectangle(Crnrs[0], Crnrs[1], Crnrs[3], Crnrs[2], N);   /* Bot. */
    N[2] = 0.0;

    if (DrawSides) {
        N[1] = 1.0;
	RenderOneRectangle(Crnrs[0], Crnrs[4], Crnrs[5], Crnrs[1], N);/*YMin.*/
	N[1] = -1.0;
	RenderOneRectangle(Crnrs[2], Crnrs[3], Crnrs[7], Crnrs[6], N);/*YMax.*/
	N[1] = 0.0;

	N[0] = 1.0;
	RenderOneRectangle(Crnrs[0], Crnrs[2], Crnrs[6], Crnrs[4], N);/*XMin.*/
	N[0] = -1.0;
	RenderOneRectangle(Crnrs[1], Crnrs[5], Crnrs[7], Crnrs[3], N);/*XMax.*/
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Traverse all active geometry and render it to the OGL window.            M
*                                                                            *
* PARAMETERS:                                                                M
*   Mat:  View transformation matrix to render the geometry from.            M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimOGLClass::RenderAxes                                                M
*****************************************************************************/
void NCSimOGLClass::RenderAxes()
{
    glDisable(GL_LIGHTING);
    glLineWidth(2);

    glBegin(GL_LINES);

    glColor3f(1.0, 0.0, 0.0);
    OGL_GL_VRTX(0.0, 0.0, 0.0);
    OGL_GL_VRTX(AxesSize, 0.0, 0.0);

    glColor3f(0.0, 1.0, 0.0);
    OGL_GL_VRTX(0.0, 0.0, 0.0);
    OGL_GL_VRTX(0.0, AxesSize, 0.0);

    glColor3f(0.0, 0.0, 1.0);
    OGL_GL_VRTX(0.0, 0.0, 0.0);
    OGL_GL_VRTX(0.0, 0.0, AxesSize);

    glEnd();

    glLineWidth(1);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Traverse all active geometry and render it to the OGL window.            M
*                                                                            *
* PARAMETERS:                                                                M
*   Mat:  View transformation matrix to render the geometry from.            M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimOGLClass::RenderGeometry                                            M
*****************************************************************************/
void NCSimOGLClass::RenderGeometry(IrtHmgnMatType Mat)
{
    NCSIM_DEBUG_FUNC_ENTRY

    int i, j;
    GLfloat GLViewMat[16];
    NCSimulationClass
	*CrntSim = NCSimMF -> CrntSim;

    for (i = 0; i < 4; i++)
        for (j = 0; j < 4; j++)
	    GLViewMat[i * 4 + j] = (float) Mat[i][j];

    /* Push this matrix as well in the hierarcy and update IGGlblCrntViewMat */
    /* to current accumulated viewing transformation.			     */
    glPushMatrix();
    glMultMatrixf(GLViewMat);

    if (DrawAxes)
        RenderAxes();

    /* Do we have anything to render? */
    if (CrntSim == NULL ||
	(CrntSim -> FinalPartGeom == NULL &&
	 CrntSim -> PObjGnrl == NULL &&
	 CrntSim -> GPathGeom == NULL &&
	 CrntSim -> StockWireGeom == NULL)) {
        glPopMatrix();
	return;
    }

    if (CrntSim -> PObjGnrl != NULL && DrawOtherGeom)
        RenderOneObject(CrntSim -> PObjGnrl, NCSIM_OGL_DISP_LIST_GNRL);

    if (CrntSim -> GPathGeom != NULL && DrawToolPath)
        RenderOneObject(CrntSim -> GPathGeom, NCSIM_OGL_DISP_LIST_GPATH);

    if (CrntSim -> StockWireGeom != NULL && DrawStock) {
	IGSetColorRGB(CrntSim -> StockColor);
        RenderOneObject(CrntSim -> StockWireGeom, NCSIM_OGL_DISP_LIST_STOCK);
    }

    if (CrntSim -> FinalPartGeom != NULL && DrawFinalPart)
        RenderOneObject(CrntSim -> FinalPartGeom, NCSIM_OGL_DISP_LIST_PART);

    if (CrntSim -> Tools[CrntSim -> ToolIdx].Geom != NULL && DrawTool) {
        glPushMatrix();

        /* The tool might be animated: */
        if (CrntSim -> AnimCrntState == NCSIM_ANIM_STATE_RUN ||
	    CrntSim -> AnimCrntState == NCSIM_ANIM_STATE_PAUSE) {
	    for (i = 0; i < 4; i++)
	        for (j = 0; j < 4; j++)
		    GLViewMat[i * 4 + j] = i == j;

	    for (j = 0; j < 3; j++)
	        GLViewMat[12 + j] = CrntSim -> AnimToolPosition[j];

	    glMultMatrixf(GLViewMat);

	    if (CrntSim -> AnimGCodeInfo != NULL &&
		CrntSim -> AnimGCodeInfo -> GCodeType ==
					    IP_NC_GCODE_LINE_MOTION_G0FAST)
	        IGSetColorRGB(CrntSim -> ToolFastColor);
	    else
	        IGSetColorRGB(CrntSim -> ToolCutColor);

	    RenderOneObject(CrntSim -> Tools[CrntSim -> ToolIdx].Geom, 
			    NCSIM_OGL_DISP_LIST_TOOL);
	}
	else {
	    IGSetColorRGB(CrntSim -> ToolIdleColor);

	    glTranslated(CrntSim -> StockBBox.Max[0],
			 CrntSim -> StockBBox.Max[1],
			 CrntSim -> StockBBox.Max[2] * 2 -
			     CrntSim -> StockBBox.Min[2]);

	    RenderOneObject(CrntSim -> Tools[CrntSim -> ToolIdx].Geom,
			    NCSIM_OGL_DISP_LIST_TOOL);
	}

	glPopMatrix();
    }

    if (CrntSim -> NCZBufferOn && DrawStock) {
	int XMin, YMin, XMax, YMax, GridSizeX, GridSizeY,
	    GridCellSizeX, GridCellSizeY,
	    n = 0;

	glEnable(GL_LIGHTING);
	if (CrntSim -> StockTransp > 0)
	    IGSetTranspObj(CrntSim -> StockTransp);
	IGSetColorRGB(CrntSim -> StockColor);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glTranslated(0.0, 0.0, -IG_POLYGON_Z_TRANS);

	XMin = CrntSim -> RedrawCellXMin;
	XMax = CrntSim -> RedrawCellXMax;
	YMin = CrntSim -> RedrawCellYMin;
	YMax = CrntSim -> RedrawCellYMax;
	CrntSim -> ZBufferGetGridSize(&GridSizeX, &GridSizeY,
				      &GridCellSizeX, &GridCellSizeY);

	for (i = 0; i < GridSizeX; i++) {
	    for (j = 0; j < GridSizeY; j++) {
		if (NCSimMF -> UseOGLDispLists) {
	            if (ForceDispListBuild ||
			ZBufDispListActive[i][j] < 0 ||
			(i >= XMin && i <= XMax && j >= YMin && j <= YMax)) {
		        /* This is a New display list, or this cell was     */
		        /* modified and was tagged as 'Active'.	            */
		        glNewList(ZBufDispListActive[i][j] = 
			              NCSIM_GET_ZBUF_DISP_LIST(i, j),
				  GL_COMPILE_AND_EXECUTE);

			n += CrntSim -> ZBufferGetCellQuadStrip(i, j,
						     RenderNCZBufferQuadStrip);

			glEndList();
		    }
		    else {
		        /* Use the existing display list. */
		        glCallList(ZBufDispListActive[i][j]);
			n++;
		    }
		}
		else {
		    n += CrntSim -> ZBufferGetCellQuadStrip(i, j,
						     RenderNCZBufferQuadStrip);

		}
	    }
	}

	CrntSim -> ZBufferUpdateCellsToRedraw(true);

	if (n > 0) {
	    /* Render the bottom walls. */
	    RenderStockBottomWalls();
	}

	glPopMatrix();

	glMatrixMode(GL_MODELVIEW);
	IGSetTranspObj(0.0);
	glDisable(GL_LIGHTING);
    }

    glPopMatrix();

    ForceDispListBuild = false;
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Renders one triangle from the NC ZBuffer, given its three vertices'	     *
* coefficients.  This function is a call back function from the NC ZBuffer.  *
*                                                                            *
* PARAMETERS:                                                                *
*   X1, Y1, Z1:   Coefficient of first triangle vertex.                      *
*   X2, Y2, Z2:   Coefficient of second triangle vertex.                     *
*   X3, Y3, Z3:   Coefficient of third triangle vertex.                      *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void RenderOneNCZBufferTriangle(IrtRType X1,
				       IrtRType Y1,
				       IrtRType Z1,
				       IrtRType X2,
				       IrtRType Y2,
				       IrtRType Z2,
				       IrtRType X3,
				       IrtRType Y3,
				       IrtRType Z3)
{
    NCSIM_DEBUG_FUNC_ENTRY

    IrtVecType V1, V2, V;

    if (IGGlblCountNumPolys)
        IGGlblNumPolys++;

    if (IGGlblDrawStyle == IG_STATE_DRAW_STYLE_SOLID) {
	glBegin(GL_POLYGON);

	V1[0] = X2 - X1;
	V1[1] = Y2 - Y1;
	V1[2] = Z2 - Z1;
	V2[0] = X3 - X1;
	V2[1] = Y3 - Y1;
	V2[2] = Z3 - Z1;
	IRIT_CROSS_PROD(V, V1, V2);
	IRIT_VEC_SAFE_NORMALIZE(V);
	if (V[2] > 0)
	    OGL_GL_NORMAL(V[0], V[1], V[2]);
	else
	    OGL_GL_NORMAL(-V[0], -V[1], -V[2]);
    }
    else {
        glBegin(GL_LINE_STRIP);
	OGL_GL_VRTX(X3, Y3, Z3);
    }

    OGL_GL_VRTX(X1, Y1, Z1);
    OGL_GL_VRTX(X2, Y2, Z2);
    OGL_GL_VRTX(X3, Y3, Z3);

    glEnd();
}
/*****************************************************************************
* DESCRIPTION:                                                               *
*   Renders one quad from the NC ZBuffer, given its two new vertices	     *
* vertices' coefficients.  This function is a call back function from the    *
* NC ZBuffer.								     *
*                                                                            *
* PARAMETERS:                                                                *
*   X1, Y1, Z1:   Coefficient of first new quad vertex.                      *
*   X2, Y2, Z2:   Coefficient of second new quad vertex.                     *
*   LastQuad:     TRUE for last quad.			                     *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void RenderNCZBufferQuadStrip(IrtRType X1,
				     IrtRType Y1,
				     IrtRType Z1,
				     IrtRType X2,
				     IrtRType Y2,
				     IrtRType Z2,
				     int LastQuad)
{
    NCSIM_DEBUG_FUNC_ENTRY

    IRIT_STATIC_DATA int
	ActiveQuad = FALSE;
    IRIT_STATIC_DATA IrtRType
	LastX1 = IRIT_INFNTY,
        LastY1 = IRIT_INFNTY,
        LastZ1 = IRIT_INFNTY,
        LastX2 = IRIT_INFNTY,
        LastY2 = IRIT_INFNTY,
        LastZ2 = IRIT_INFNTY;
    IrtVecType V1, V2, V;

    if (IGGlblDrawStyle == IG_STATE_DRAW_STYLE_SOLID) {
        if (!ActiveQuad) {
	    if (LastX1 == IRIT_INFNTY) {
	        /* Really first pair - just keep the pair. */
	        LastX1 = X1;
		LastY1 = Y1;
		LastZ1 = Z1;
	        LastX2 = X2;
		LastY2 = Y2;
		LastZ2 = Z2;
		return;
	    }
	    else {
	        /* The second pair - we have the first quad. */
		ActiveQuad = TRUE;
		glBegin(GL_QUAD_STRIP);

		V1[0] = X2 - X1;
		V1[1] = Y2 - Y1;
		V1[2] = Z2 - Z1;
		V2[0] = LastX1 - X1;
		V2[1] = LastY1 - Y1;
		V2[2] = LastZ1 - Z1;
		IRIT_CROSS_PROD(V, V1, V2);
		IRIT_VEC_SAFE_NORMALIZE(V);
		if (V[2] > 0)
		    OGL_GL_NORMAL(V[0], V[1], V[2]);
		else
		    OGL_GL_NORMAL(-V[0], -V[1], -V[2]);

		OGL_GL_VRTX(LastX1, LastY1, LastZ1);
		OGL_GL_VRTX(LastX2, LastY2, LastZ2);
	    }
	}
	else {
	    V1[0] = X2 - X1;
	    V1[1] = Y2 - Y1;
	    V1[2] = Z2 - Z1;
	    V2[0] = LastX1 - X1;
	    V2[1] = LastY1 - Y1;
	    V2[2] = LastZ1 - Z1;
	    IRIT_CROSS_PROD(V, V1, V2);
	    IRIT_VEC_SAFE_NORMALIZE(V);
	    if (V[2] > 0)
	        OGL_GL_NORMAL(V[0], V[1], V[2]);
	    else
	        OGL_GL_NORMAL(-V[0], -V[1], -V[2]);
	}

	OGL_GL_VRTX(X1, Y1, Z1);
	OGL_GL_VRTX(X2, Y2, Z2);

	if (IGGlblCountNumPolys)
	    IGGlblNumPolys++;

	if (LastQuad) {
	    glEnd();
	    ActiveQuad = FALSE;
	    LastX1 = IRIT_INFNTY;
	    LastY1 = IRIT_INFNTY;
	    LastZ1 = IRIT_INFNTY;
	    LastX2 = IRIT_INFNTY;
	    LastY2 = IRIT_INFNTY;
	    LastZ2 = IRIT_INFNTY;
	}
	else {
	    LastX1 = X1;
	    LastY1 = Y1;
	    LastZ1 = Z1;
	}
    }
    else {
        glBegin(GL_LINES);

	OGL_GL_VRTX(X1, Y1, Z1); 
	OGL_GL_VRTX(X2, Y2, Z2); 

	if (IGGlblCountNumPolys)
	    IGGlblNumPolys++;

        if (!ActiveQuad) {
	    ActiveQuad = TRUE;
	}
	else {
	    OGL_GL_VRTX(LastX1, LastY1, LastZ1); 
	    OGL_GL_VRTX(X1, Y1, Z1); 
	    OGL_GL_VRTX(LastX2, LastY2, LastZ2); 
	    OGL_GL_VRTX(X2, Y2, Z2); 
	}

	if (LastQuad) {
	    glEnd();
	    ActiveQuad = FALSE;
	}
	else {
	    LastX1 = X1;
	    LastY1 = Y1;
	    LastZ1 = Z1;
	    LastX2 = X2;
	    LastY2 = Y2;
	    LastZ2 = Z2;
	}
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Renders one rectangle, given its four vertices and normal.               M
*                                                                            *
* PARAMETERS:                                                                M
*   Pt1, Pt2, Pt3, Pt4:   The four vertices of the rectangle polygon.        M
*   N:			  The normal of the polygon.                         M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimOGLClass::RenderOneRectangle                                        M
*****************************************************************************/
void NCSimOGLClass::RenderOneRectangle(IrtPtType Pt1,
				       IrtPtType Pt2,
				       IrtPtType Pt3,
				       IrtPtType Pt4,
				       IrtPtType N)
{
    NCSIM_DEBUG_FUNC_ENTRY

    if (IGGlblDrawStyle == IG_STATE_DRAW_STYLE_SOLID) {
        glBegin(GL_POLYGON);

	OGL_GL_NORMAL_V(N);
    }
    else {
        glBegin(GL_LINE_STRIP);

	OGL_GL_VRTX_V(Pt4);
    }

    OGL_GL_VRTX_V(Pt1);
    OGL_GL_VRTX_V(Pt2);
    OGL_GL_VRTX_V(Pt3);
    OGL_GL_VRTX_V(Pt4);

    glEnd();
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Handles OGL rendering events.                                            M
*                                                                            *
* PARAMETERS:                                                                M
*   None                                                                     M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimOGLClass::RenderCrntSim                                             M
*****************************************************************************/
void NCSimOGLClass::RenderCrntSim()
{
    NCSIM_DEBUG_FUNC_ENTRY

    if (NCSimMF == NULL || NCSimMF -> CrntSim == NULL)
	return;

    wxPaintDC
	dc(this);

    /* Init OpenGL once, but after SetCurrent. */
    InitGL();

    IGGlblNumPolys = 0;
    IGUpdateFPS(TRUE);

    glDrawBuffer(IGGlblDoDoubleBuffer ? GL_BACK : GL_FRONT);
    glClearColor((float) (IGGlblBackGroundColor[0] / 255.0),
		 (float) (IGGlblBackGroundColor[1] / 255.0),
		 (float) (IGGlblBackGroundColor[2] / 255.0),
		 1.0);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(-1.0, 1.0, -1.0, 1.0, IGGlblZMinClip, IGGlblZMaxClip);

    glEnable(GL_NORMALIZE);   /* We do scaling extensively here. */

    if ((IGGlblAntiAliasing == IG_STATE_ANTI_ALIAS_ON &&
	 IGGlblDrawStyle == IG_STATE_DRAW_STYLE_WIREFRAME) ||
	(IGGlblAntiAliasing == IG_STATE_ANTI_ALIAS_BLEND &&
	 IGGlblDrawStyle == IG_STATE_DRAW_STYLE_SOLID)) {
        glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	glEnable(GL_LINE_SMOOTH);
	glEnable(GL_POINT_SMOOTH);
	glEnable(GL_BLEND);
    }
    else {
        glBlendFunc(GL_ONE, GL_ZERO);
	glDisable(GL_LINE_SMOOTH);
	glDisable(GL_POINT_SMOOTH);
	glDisable(GL_BLEND);
    }

    if (IGGlblBackFaceCull)
        glEnable(GL_CULL_FACE);
    else
        glDisable(GL_CULL_FACE);
    glCullFace(GL_FRONT);           /* In irit we have reversed orientation. */

    if (IGGlblDepthCue)
        glEnable(GL_FOG);
    else
        glDisable(GL_FOG);

    if (IGGlblDrawStyle == IG_STATE_DRAW_STYLE_SOLID) {
        glEnable(GL_LIGHTING);
	glEnable(GL_DEPTH_TEST);
    }
    else {
        glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);
    }

    switch (IGGlblViewMode) {        /* Update the current view. */
	case IG_VIEW_ORTHOGRAPHIC:
	    IRIT_GEN_COPY(IGGlblCrntViewMat, IPViewMat, sizeof(IrtHmgnMatType));
	    break;
	case IG_VIEW_PERSPECTIVE:
	    MatMultTwo4by4(IGGlblCrntViewMat, IPViewMat, IPPrspMat);
	    break;
    }
    MatInverseMatrix(IGGlblCrntViewMat, IGGlblInvCrntViewMat);

    IGSetLightSource(NULL, NULL, -1);              /* Refresh light sources. */

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    if (!IGGlblManipulationActive &&
	IGGlblDoDoubleBuffer &&
	IGGlblAntiAliasing != IG_STATE_ANTI_ALIAS_OFF &&
	IGGlblDrawStyle == IG_STATE_DRAW_STYLE_SOLID) {
        static double
	    Jitter4[4][2] = {
		{  -0.57, -0.53 },
		{   0.57,  0.53 },
		{  -0.53,  0.57 },
		{   0.53, -0.57 }
	    };
	GLint ViewPort[4];
	int j;

	glGetIntegerv(GL_VIEWPORT, ViewPort);
	glClear(GL_ACCUM_BUFFER_BIT);

	for (j = 0; j < 4; j++) {
	    IrtHmgnMatType JitteredCrntViewMat, Mat;

	    if (j == 0)      /* Clear screen/set background & clear ZBuffer. */
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	    MatGenMatTrans(Jitter4[j][0] / ViewPort[2],
			   Jitter4[j][1] / ViewPort[3],
			   0.0, Mat);
	    MatMultTwo4by4(JitteredCrntViewMat, IGGlblCrntViewMat, Mat);

	    RenderGeometry(JitteredCrntViewMat);

	    glAccum(GL_ACCUM, 0.25);
	}

	glAccum(GL_RETURN, 1);
    }
    else {
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	RenderGeometry(IGGlblCrntViewMat);
    }

    glFlush();

    SwapBuffers();

    IGUpdateFPS(FALSE);
    NCSimMF -> SetFPSNumPolysStatus(IGGlblFramePerSec, IGGlblNumPolys);

    NCSimMF -> CrntSim -> SaveView(IPViewMat);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Traverse all active geometry and pick closest location to mouse.         M
*                                                                            *
* PARAMETERS:                                                                M
*   x, y: Mouse position.						     M
*   Mat:  View transformation matrix of geometry.		             M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimOGLClass::PickGeometry                                              M
*****************************************************************************/
void NCSimOGLClass::PickGeometry(int x, int y)
{
    NCSIM_DEBUG_FUNC_ENTRY

    int i, MinPlIsPolyline;
    IrtRType HitDepth, MinDist,
        AllMinDist = IRIT_INFNTY;
    IrtVecType RayDir;
    IrtPtType Pt1, Pt2, MinPt, AllMinPt, PickPos;
    IrtHmgnMatType InvMat;
    IPPolygonStruct *MinPl;
    IPObjectStruct *PTmp;
    NCSimulationClass
	*CrntSim = NCSimMF -> CrntSim;

    /* Do we have anything to render? */
    if (CrntSim == NULL ||
	(CrntSim -> FinalPartGeom == NULL &&
	 CrntSim -> PObjGnrl == NULL &&
	 CrntSim -> GPathGeom == NULL &&
	 CrntSim -> StockWireGeom == NULL))
	return;

    if (!MatInverseMatrix(IGGlblCrntViewMat, InvMat))
        return;

    Screen2ObjSpcCoords(x, y, &PickPos[0], &PickPos[1]);
    PickPos[2] = 0.0;
    MatMultPtby4by4(Pt1, PickPos, InvMat);
    PickPos[2] = 1.0;
    MatMultPtby4by4(Pt2, PickPos, InvMat);
    IRIT_PT_SUB(RayDir, Pt2, Pt1);

    /* Find out the closest object in the display list to the picked line. */
    IGGlblMinPickDist = IRIT_INFNTY;
    if (CrntSim -> GPathGeom != NULL) {
	for (i = 0;
	     (PTmp = IPListObjectGet(CrntSim -> GPathGeom, i++)) != NULL; ) {
	    if (!IP_HAS_BBOX_OBJ(PTmp)) {
	        GMBBBboxStruct
		    *BBox = GMBBComputeBboxObject(PTmp);

		IRIT_PT_COPY(PTmp -> BBox[0], BBox -> Min);
		IRIT_PT_COPY(PTmp -> BBox[1], BBox -> Max);
		IP_SET_BBOX_OBJ(PTmp);
	    }

	    MinDist = IGFindMinimalDist(PTmp, &MinPl, MinPt,
					&MinPlIsPolyline, Pt1, RayDir,
					&HitDepth);

	    if (AllMinDist > MinDist) {
	        AllMinDist = MinDist;
		IRIT_PT_COPY(AllMinPt, MinPt);
	    }
	}
    }

    if (AllMinDist < IRIT_INFNTY) {
	NCSimPrintf(NCSIM_LOG_INFO, "Picked toolpath location %f %f %f.\n",
		    AllMinPt[0], AllMinPt[1], AllMinPt[2]);
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Handles OGL window mouse events.	                                     M
*                                                                            *
* PARAMETERS:                                                                M
*   Event:  The event's details.                                             M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimOGLClass::OnMouseEvent                                              M
*****************************************************************************/
void NCSimOGLClass::OnMouseEvent(wxMouseEvent &Event)
{
    NCSIM_DEBUG_FUNC_ENTRY

    int x = Event.GetPosition().x,
	y = Event.GetPosition().y;

    if (Event.ShiftDown() && Event.LeftDown()) {
        /* It is a pick request. */
        PickGeometry(x, y);
	return;
    }

    if (Event.LeftDown()) {
	CaptureMouse();
	MouseKey = MOUSE_LEFT_DOWN;
	LastEvent = IG_EVENT_ROTATE;
    }
    else if (Event.LeftUp()) {
	MouseKey = MOUSE_KEY_NONE;
	ReleaseMouse();
    }
    else if (Event.MiddleDown()) {
	CaptureMouse();
	MouseKey = MOUSE_MIDDLE_DOWN;
	LastEvent = IG_EVENT_SCALE;
    }
    else if (Event.MiddleUp()) {
	MouseKey = MOUSE_KEY_NONE;
	ReleaseMouse();
    }
    else if (Event.RightDown()) {
	CaptureMouse();
	MouseKey = MOUSE_RIGHT_DOWN;
	LastEvent = IG_EVENT_TRANSLATE;
    }
    else if (Event.RightUp()) {
	MouseKey = MOUSE_KEY_NONE;
	ReleaseMouse();
    }
    else if (Event.Dragging()) {
        IrtRType ChangeFactors[2];

        switch (MouseKey) {
	    case MOUSE_LEFT_DOWN:
	    case MOUSE_MIDDLE_DOWN:
	    case MOUSE_RIGHT_DOWN:
		ChangeFactors[0] = (x - MouseX) * IGGlblChangeFactor;
		ChangeFactors[1] = (MouseY - y) * IGGlblChangeFactor;
		if (LastEvent == IG_EVENT_SCALE) {
		    ChangeFactors[0] *= SCALE_FACTOR_MODIFIER;
		    ChangeFactors[1] *= SCALE_FACTOR_MODIFIER;
		}

		if (IGDefaultProcessEvent(LastEvent, ChangeFactors))
		    Refresh();
	        break;
	    case MOUSE_KEY_NONE:
	    default:
	        break;
	}
    }

    MouseX = x;
    MouseY = y;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Handles OGL window (re)paint events.                                     M
*                                                                            *
* PARAMETERS:                                                                M
*   Event:  The event's details.                                             M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimOGLClass::OnPaint                                                   M
*****************************************************************************/
void NCSimOGLClass::OnPaint(wxPaintEvent &WXUNUSED(Event))
{
    NCSIM_DEBUG_FUNC_ENTRY

    RenderCrntSim();
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Handles OGL window (re)size events.                                      M
*                                                                            *
* PARAMETERS:                                                                M
*   Event:  The event's details.                                             M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimOGLClass::OnSize                                                    M
*****************************************************************************/
void NCSimOGLClass::OnSize(wxSizeEvent &Event)
{
    /* Set GL viewport (not called by wxGLCanvas::OnSize on all platforms). */
    GetClientSize(&ScrnWidth, &ScrnHeight);
   
    if (ScrnWidth > ScrnHeight) {
        ScrnViewportBias = (ScrnWidth - ScrnHeight) / 2;

        glViewport(0, -ScrnViewportBias, ScrnWidth, ScrnWidth);
    }
    else {
        ScrnViewportBias = (ScrnHeight - ScrnWidth) / 2;

	glViewport(-ScrnViewportBias, 0, ScrnHeight, ScrnHeight);
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Handles OGL window background redraw events.                             M
*                                                                            *
* PARAMETERS:                                                                M
*   Event:  The event's details.                                             M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimOGLClass::OnEraseBackground                                         M
*****************************************************************************/
void NCSimOGLClass::OnEraseBackground(wxEraseEvent &WXUNUSED(Event))
{
    NCSIM_DEBUG_FUNC_ENTRY

    /* Do nothing, to avoid flashing. */
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Handles OGL initialization events.                                       M
*                                                                            *
* PARAMETERS:                                                                M
*   None                                                                     M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimOGLClass::InitGL                                                    M
*****************************************************************************/
void NCSimOGLClass::InitGL()
{
    NCSIM_DEBUG_FUNC_ENTRY

    IRIT_STATIC_DATA IrtVecType
        WhiteColor = { 1.0, 1.0, 1.0 };
    int i;
    GLfloat FogColor[4];

    if (!GlblOGLInitialized)
        GlblOGLInitialized = true;
    else
        return;

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glOrtho(0.0, 1.0, 0.0, 1.0, IGGlblZMinClip, IGGlblZMaxClip);

    glRasterPos2i(0, 0);

    glEnable(GL_COLOR_MATERIAL);
    glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);

    glFrontFace(GL_CCW);

    for (i = 0; i < 3; i++)
	FogColor[i] = (float) IGGlblBackGroundColor[i] / 255.0;
    FogColor[3] = (float) 1.0;
    glFogfv(GL_FOG_COLOR, FogColor);
    glFogi(GL_FOG_MODE, GL_LINEAR);

    glClearColor(IGGlblBackGroundColor[0] / 255.0,
		 IGGlblBackGroundColor[1] / 255.0,
		 IGGlblBackGroundColor[2] / 255.0,
		 1.0);
    glClearAccum(0.0, 0.0, 0.0, 0.0);

    for (i = 0; i < IGShadeParam.NumOfLightSrcs; i++)
        IGSetLightSource(IGShadeParam.LightPos[i], WhiteColor, i);

    IGGlblCountNumPolys = IGGlblCountFramePerSec = TRUE;
    IGGlblRelLowresFineNess = 1.0;
    IGGlblPllnFineness = NCSIM_DEFAULT_PLLN_OPTI_FINENESS;

    OGLDispListBase = glGenLists(NCSIM_OGL_DISP_LIST_RANGE);
    assert(OGLDispListBase != 0);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   handles OGL window key down events.                                      M
*                                                                            *
* PARAMETERS:                                                                M
*   Event:  The event's details.                                             M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimOGLClass::OnKeyDown                                                 M
*****************************************************************************/
void NCSimOGLClass::OnKeyDown(wxKeyEvent &Event)
{
    NCSIM_DEBUG_FUNC_ENTRY

    LastKey = Event.GetKeyCode();
    if (LastKey == 0)
	return;

    Event.Skip();
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   handles OGL window key up events.                                        M
*                                                                            *
* PARAMETERS:                                                                M
*   Event:  The event's details.                                             M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimOGLClass::OnKeyUp                                                   M
*****************************************************************************/
void NCSimOGLClass::OnKeyUp(wxKeyEvent &Event)
{
    NCSIM_DEBUG_FUNC_ENTRY

    Event.Skip();
}

/*****************************************************************************
* AUXILIARY:								     *
*    Event table for NCSimOGLParamClass				             *
*****************************************************************************/
BEGIN_EVENT_TABLE(NCSimOGLParamClass, wxDialog)
    EVT_TEXT(wxID_NCSIM_OGL_PARAM_LINE_WIDTH, OnTextChange)
    EVT_TEXT(wxID_NCSIM_OGL_PARAM_FRAME_SKIP, OnTextChange)
    EVT_CHECKBOX(wxID_NCSIM_OGL_PARAM_DRAW_AXES, OnCheckBox)
    EVT_CHECKBOX(wxID_NCSIM_OGL_PARAM_DRAW_TOOL, OnCheckBox)
    EVT_CHECKBOX(wxID_NCSIM_OGL_PARAM_DRAW_TPATH, OnCheckBox)
    EVT_CHECKBOX(wxID_NCSIM_OGL_PARAM_DRAW_STOCK, OnCheckBox)
    EVT_CHECKBOX(wxID_NCSIM_OGL_PARAM_DRAW_FINAL_PART, OnCheckBox)
    EVT_CHECKBOX(wxID_NCSIM_OGL_PARAM_DRAW_OTHER_GEOM, OnCheckBox)
    EVT_BUTTON(wxID_NCSIM_OGL_PARAM_ORTHO_PERSP, OnButton)
    EVT_BUTTON(wxID_NCSIM_OGL_PARAM_SOLID_WIRE, OnButton)
    EVT_BUTTON(wxID_NCSIM_OGL_PARAM_LIGHT0CLR, OnButton)
    EVT_BUTTON(wxID_NCSIM_OGL_PARAM_LIGHT1CLR, OnButton)
    EVT_BUTTON(wxID_NCSIM_OGL_PARAM_BG_COLOR, OnButton)
END_EVENT_TABLE()

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Sets the given button background color to Clr (and compensate the        *
* foreground accordingly.)                                                   *
*                                                                            *
* PARAMETERS:                                                                *
*   Btn:  Button to update its background color.                             *
*   Clr:  New color to use.				                     *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void SetButtonBGColor(wxWindow *Btn, int *Clr)
{
    NCSIM_DEBUG_FUNC_ENTRY

    Btn -> SetBackgroundColour(wxColour(Clr[0],
					Clr[1],
					Clr[2]));
    if (Clr[0] + Clr[1] + Clr[2] > 374)
	Btn -> SetForegroundColour(wxColour(0, 0, 0));
    else
	Btn -> SetForegroundColour(wxColour(255, 255, 255));
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Constructor for the interactive setup control of NC loaded file.         M
*                                                                            *
* PARAMETERS:                                                                M
*   NCSimOGL:    Simulation OGL we should create this page for.		     M
*   Parent:      Frame window.					             M
*   Id:          ID assigned to window. 		                     M
*   Title:       Window's title.		       		             M
*   Flags:       Window styles' flags.				             M
*                                                                            *
* RETURN VALUE:                                                              M
*   NoValue                                                                  M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimOGLParamClass::NCSimOGLParamClass                                   M
*****************************************************************************/
NCSimOGLParamClass::NCSimOGLParamClass(NCSimOGLClass *NCSimOGL,
				       wxWindow *Parent,
				       wxWindowID Id,
				       const wxString &Title,
				       int Flags):
	wxDialog(Parent,
		 Id,
		 Title,
		 wxDefaultPosition,
		 wxDefaultSize,
		 Flags | wxCAPTION)
{
    NCSIM_DEBUG_FUNC_ENTRY

    int i;
    wxBoxSizer *Sizer,
	*TopSizer = new wxBoxSizer(wxVERTICAL);
    wxButton *Btn;
    wxCheckBox *CBtn;
    wxStaticText *Ln;
    wxSpinCtrl *SCtrl;
    RealSpinCtrlClass *RSpin;
    wxArrayString Strs1, Strs2;
    char Line[IRIT_LINE_LEN];

    /* Update local light sources info. */
    for (i = 0; i < 4; i++) {
        Light0Pos[i] = IGShadeParam.LightPos[0][i];
        Light1Pos[i] = IGShadeParam.LightPos[1][i];
    }
    for (i = 0; i < 3; i++) {
	Light0Clr[i] = Light1Clr[i] = 255;
    }

    ThisNCSimOGL = NCSimOGL;  /* Keep a ref. to NCSimOGLClass we belong to. */

    SetSizer(TopSizer);

    /* Row 1. */
    Sizer = new wxBoxSizer(wxHORIZONTAL);

    Btn = new wxButton(this, wxID_NCSIM_OGL_PARAM_ORTHO_PERSP,
	IGGlblViewMode == IG_VIEW_ORTHOGRAPHIC ? wxT("Orthographic")
					       : wxT("Perspective"));
    Btn -> SetToolTip(NCSimToolTip(wxID_NCSIM_OGL_PARAM_ORTHO_PERSP));
    Sizer -> Add(Btn, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    Btn = new wxButton(this, wxID_NCSIM_OGL_PARAM_SOLID_WIRE,
	IGGlblDrawStyle == IG_STATE_DRAW_STYLE_SOLID ? wxT("Solid")
						     : wxT("WireFrame"));
    Btn -> SetToolTip(NCSimToolTip(wxID_NCSIM_OGL_PARAM_SOLID_WIRE));
    Sizer -> Add(Btn, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    Ln = new wxStaticText(this, wxID_ANY, wxT("Line\nWidth:"));
    Ln -> SetToolTip(NCSimToolTip(wxID_NCSIM_OGL_PARAM_LINE_WIDTH));
    Sizer -> Add(Ln, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    sprintf(Line, "%d", IGGlblLineWidth);
    SCtrl = new wxSpinCtrl(this, wxID_NCSIM_OGL_PARAM_LINE_WIDTH,
			   NCSimChars2wxString(Line),
			   wxDefaultPosition, wxSize(40, 20),
			   wxTE_RICH | wxSP_ARROW_KEYS | wxSP_WRAP, 1, 100, 1);
    SCtrl -> SetToolTip(NCSimToolTip(wxID_NCSIM_OGL_PARAM_LINE_WIDTH));
    Sizer -> Add(SCtrl, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    TopSizer -> Add(Sizer, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    /* Row 2. */
    Sizer = new wxBoxSizer(wxHORIZONTAL);

    Btn = new wxButton(this, wxID_NCSIM_OGL_PARAM_BG_COLOR,
		       wxT("Bkgrnd Color"));
    SetButtonBGColor(Btn, IGGlblBackGroundColor);
    Btn -> SetToolTip(NCSimToolTip(wxID_NCSIM_OGL_PARAM_BG_COLOR));
    Sizer -> Add(Btn, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    Ln = new wxStaticText(this, wxID_ANY, wxT("Sensitivity\nof Mouse:"));
    Ln -> SetToolTip(NCSimToolTip(wxID_NCSIM_OGL_PARAM_MOUSE_SENS));
    Sizer -> Add(Ln, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    RSpin = new RealSpinCtrlClass(&IGGlblChangeFactor,
				  0.001, 1000, 0.01, 0.001,
				  this, wxID_NCSIM_OGL_PARAM_MOUSE_SENS,
				  wxDefaultPosition, wxSize(45, 20));
    RSpin -> SetUpdateFunction(RealSpinUpdateValCB);
    RSpin -> SetToolTip(NCSimToolTip(wxID_NCSIM_OGL_PARAM_MOUSE_SENS));
    Sizer -> Add(RSpin, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    Ln = new wxStaticText(this, wxID_ANY, wxT("Frame\nSkip:"));
    Ln -> SetToolTip(NCSimToolTip(wxID_NCSIM_OGL_PARAM_FRAME_SKIP));
    Sizer -> Add(Ln, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    sprintf(Line, "%d", NCSimOGL -> SkipFrameRate);
    SCtrl = new wxSpinCtrl(this, wxID_NCSIM_OGL_PARAM_FRAME_SKIP,
			   NCSimChars2wxString(Line),
			   wxDefaultPosition, wxSize(50, 20),
			   wxTE_RICH | wxSP_ARROW_KEYS | wxSP_WRAP, 1, 1000,
			   10);
    SCtrl -> SetToolTip(NCSimToolTip(wxID_NCSIM_OGL_PARAM_FRAME_SKIP));
    Sizer -> Add(SCtrl, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    TopSizer -> Add(Sizer, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    /* Row 3. */
    Sizer = new wxBoxSizer(wxHORIZONTAL);

    Btn = new wxButton(this, wxID_NCSIM_OGL_PARAM_LIGHT0CLR, wxT("Light0 Color"));
    Btn -> SetToolTip(NCSimToolTip(wxID_NCSIM_OGL_PARAM_LIGHT0POS));
    Sizer -> Add(Btn, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    RSpin = new RealSpinCtrlClass(&Light0Pos[0], -1000, 1000, 0.1, 0.01,
				  this, wxID_NCSIM_OGL_PARAM_LIGHT0POS,
				  wxDefaultPosition, wxSize(45, 20));
    RSpin -> SetUpdateFunction(RealSpinUpdateValCB);
    RSpin -> SetToolTip(NCSimToolTip(wxID_NCSIM_OGL_PARAM_LIGHT0POS));
    Sizer -> Add(RSpin, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    RSpin = new RealSpinCtrlClass(&Light0Pos[1], -1000, 1000, 0.1, 0.01,
				  this, wxID_NCSIM_OGL_PARAM_LIGHT0POS,
				  wxDefaultPosition, wxSize(45, 20));
    RSpin -> SetUpdateFunction(RealSpinUpdateValCB);
    RSpin -> SetToolTip(NCSimToolTip(wxID_NCSIM_OGL_PARAM_LIGHT0POS));
    Sizer -> Add(RSpin, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    RSpin = new RealSpinCtrlClass(&Light0Pos[2], -1000, 1000, 0.1, 0.01,
				  this, wxID_NCSIM_OGL_PARAM_LIGHT0POS,
				  wxDefaultPosition, wxSize(45, 20));
    RSpin -> SetUpdateFunction(RealSpinUpdateValCB);
    RSpin -> SetToolTip(NCSimToolTip(wxID_NCSIM_OGL_PARAM_LIGHT0POS));
    Sizer -> Add(RSpin, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    RSpin = new RealSpinCtrlClass(&Light0Pos[3], 0, 1000, 0.1, 0.01,
				  this, wxID_NCSIM_OGL_PARAM_LIGHT0POS,
				  wxDefaultPosition, wxSize(45, 20));
    RSpin -> SetUpdateFunction(RealSpinUpdateValCB);
    RSpin -> SetToolTip(NCSimToolTip(wxID_NCSIM_OGL_PARAM_LIGHT0POS));
    Sizer -> Add(RSpin, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    TopSizer -> Add(Sizer, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    /* Row 4. */
    Sizer = new wxBoxSizer(wxHORIZONTAL);

    Btn = new wxButton(this, wxID_NCSIM_OGL_PARAM_LIGHT1CLR, wxT("Light1 Color"));
    Btn -> SetToolTip(NCSimToolTip(wxID_NCSIM_OGL_PARAM_LIGHT1POS));
    Sizer -> Add(Btn, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    RSpin = new RealSpinCtrlClass(&Light1Pos[0], -1000, 1000, 0.1, 0.01,
				  this, wxID_NCSIM_OGL_PARAM_LIGHT1POS,
				  wxDefaultPosition, wxSize(45, 20));
    RSpin -> SetUpdateFunction(RealSpinUpdateValCB);
    RSpin -> SetToolTip(NCSimToolTip(wxID_NCSIM_OGL_PARAM_LIGHT1POS));
    Sizer -> Add(RSpin, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    RSpin = new RealSpinCtrlClass(&Light1Pos[1], -1000, 1000, 0.1, 0.01,
				  this, wxID_NCSIM_OGL_PARAM_LIGHT1POS,
				  wxDefaultPosition, wxSize(45, 20));
    RSpin -> SetUpdateFunction(RealSpinUpdateValCB);
    RSpin -> SetToolTip(NCSimToolTip(wxID_NCSIM_OGL_PARAM_LIGHT1POS));
    Sizer -> Add(RSpin, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    RSpin = new RealSpinCtrlClass(&Light1Pos[2], -1000, 1000, 0.1, 0.01,
				  this, wxID_NCSIM_OGL_PARAM_LIGHT1POS,
				  wxDefaultPosition, wxSize(45, 20));
    RSpin -> SetUpdateFunction(RealSpinUpdateValCB);
    RSpin -> SetToolTip(NCSimToolTip(wxID_NCSIM_OGL_PARAM_LIGHT1POS));
    Sizer -> Add(RSpin, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    RSpin = new RealSpinCtrlClass(&Light1Pos[3], 0, 1000, 0.1, 0.01,
				  this, wxID_NCSIM_OGL_PARAM_LIGHT1POS,
				  wxDefaultPosition, wxSize(45, 20));
    RSpin -> SetUpdateFunction(RealSpinUpdateValCB);
    RSpin -> SetToolTip(NCSimToolTip(wxID_NCSIM_OGL_PARAM_LIGHT1POS));
    Sizer -> Add(RSpin, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    TopSizer -> Add(Sizer, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    /* Row 5. */
    Sizer = new wxBoxSizer(wxHORIZONTAL);

    Ln = new wxStaticText(this, wxID_ANY, wxT("Draw:"));
    Sizer -> Add(Ln, 0, wxALIGN_CENTER_VERTICAL | wxALL, 2);

    CBtn = new wxCheckBox(this, wxID_NCSIM_OGL_PARAM_DRAW_AXES, wxT("Axes"));
    CBtn -> Set3StateValue(NCSimOGL -> DrawAxes ? wxCHK_CHECKED
						: wxCHK_UNCHECKED);
    CBtn -> SetToolTip(NCSimToolTip(wxID_NCSIM_OGL_PARAM_DRAW_AXES));
    Sizer -> Add(CBtn, 0, wxALIGN_CENTER_VERTICAL | wxALL, 2);

    CBtn = new wxCheckBox(this, wxID_NCSIM_OGL_PARAM_DRAW_TOOL, wxT("Tool"));
    CBtn -> Set3StateValue(NCSimOGL -> DrawTool ? wxCHK_CHECKED
						: wxCHK_UNCHECKED);
    CBtn -> SetToolTip(NCSimToolTip(wxID_NCSIM_OGL_PARAM_DRAW_TOOL));
    Sizer -> Add(CBtn, 0, wxALIGN_CENTER_VERTICAL | wxALL, 2);

    CBtn = new wxCheckBox(this, wxID_NCSIM_OGL_PARAM_DRAW_TPATH,
			  wxT("TPath"));
    CBtn -> Set3StateValue(NCSimOGL -> DrawToolPath ? wxCHK_CHECKED
						    : wxCHK_UNCHECKED);
    CBtn -> SetToolTip(NCSimToolTip(wxID_NCSIM_OGL_PARAM_DRAW_TPATH));
    Sizer -> Add(CBtn, 0, wxALIGN_CENTER_VERTICAL | wxALL, 2);

    CBtn = new wxCheckBox(this, wxID_NCSIM_OGL_PARAM_DRAW_STOCK,
			  wxT("Stock"));
    CBtn -> Set3StateValue(NCSimOGL -> DrawStock ? wxCHK_CHECKED
						 : wxCHK_UNCHECKED);
    CBtn -> SetToolTip(NCSimToolTip(wxID_NCSIM_OGL_PARAM_DRAW_STOCK));
    Sizer -> Add(CBtn, 0, wxALIGN_CENTER_VERTICAL | wxALL, 2);

    CBtn = new wxCheckBox(this, wxID_NCSIM_OGL_PARAM_DRAW_FINAL_PART,
			  wxT("Part"));
    CBtn -> Set3StateValue(NCSimOGL -> DrawFinalPart ? wxCHK_CHECKED
						     : wxCHK_UNCHECKED);
    CBtn -> SetToolTip(NCSimToolTip(wxID_NCSIM_OGL_PARAM_DRAW_FINAL_PART));
    Sizer -> Add(CBtn, 0, wxALIGN_CENTER_VERTICAL | wxALL, 2);

    CBtn = new wxCheckBox(this, wxID_NCSIM_OGL_PARAM_DRAW_OTHER_GEOM,
			  wxT("Other"));
    CBtn -> Set3StateValue(NCSimOGL -> DrawOtherGeom ? wxCHK_CHECKED
						     : wxCHK_UNCHECKED);
    CBtn -> SetToolTip(NCSimToolTip(wxID_NCSIM_OGL_PARAM_DRAW_OTHER_GEOM));
    Sizer -> Add(CBtn, 0, wxALIGN_CENTER_VERTICAL | wxALL, 2);

    TopSizer -> Add(Sizer, 0, wxALIGN_CENTER_VERTICAL | wxALL, 2);

    /* Row 5. */
    Sizer = new wxBoxSizer(wxHORIZONTAL);

    Ln = new wxStaticText(this, wxID_ANY,
			  wxT("Mouse Sensitivity\nTransformations:"));
    Ln -> SetToolTip(NCSimToolTip(wxID_NCSIM_OGL_PARAM_MOUSE_SENSITIVITY));
    Sizer -> Add(Ln, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    RSpin = new RealSpinCtrlClass(&IGGlblChangeFactor, 0, 1000, 0.1, 0.01,
				  this, wxID_NCSIM_OGL_PARAM_MOUSE_SENSITIVITY,
				  wxDefaultPosition, wxSize(55, 20));
    RSpin -> SetUpdateFunction(RealSpinUpdateValCB);
    RSpin -> SetToolTip(NCSimToolTip(wxID_NCSIM_OGL_PARAM_MOUSE_SENSITIVITY));
    Sizer -> Add(RSpin, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    Ln = new wxStaticText(this, wxID_ANY,
			  wxT("Axes Size:"));
    Ln -> SetToolTip(NCSimToolTip(wxID_NCSIM_OGL_PARAM_AXES_SIZE));
    Sizer -> Add(Ln, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    RSpin = new RealSpinCtrlClass(&ThisNCSimOGL -> AxesSize, 0, 1000, 1, 0.1,
				  this, wxID_NCSIM_OGL_PARAM_AXES_SIZE,
				  wxDefaultPosition, wxSize(65, 20));
    RSpin -> SetUpdateFunction(RealSpinUpdateValCB);
    RSpin -> SetToolTip(NCSimToolTip(wxID_NCSIM_OGL_PARAM_AXES_SIZE));
    Sizer -> Add(RSpin, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    TopSizer -> Add(Sizer, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    /* Last Row. */
    Sizer = new wxBoxSizer(wxHORIZONTAL);

    Btn = new wxButton(this, wxID_OK, wxT("OK"));
    Sizer -> Add(Btn, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    TopSizer -> Add(Sizer, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    /* Done. */
    TopSizer -> Fit(this);
    TopSizer -> SetSizeHints(this);
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Call back function of a RealSpin object, after its value being modified. *
*                                                                            *
* PARAMETERS:                                                                *
*   Val:   New values.							     *
*   Id:    ID of modified window.					     *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void RealSpinUpdateValCB(double Val, wxWindowID Id)
{
    NCSIM_DEBUG_FUNC_ENTRY

    NCSimOGLClass
	*OGLWin = NCSimMF -> OGLWin;

    switch (Id) {
	case wxID_NCSIM_OGL_PARAM_MOUSE_SENS:
	    break;
	case wxID_NCSIM_OGL_PARAM_AXES_SIZE:
	case wxID_NCSIM_OGL_PARAM_LIGHT0POS:
	case wxID_NCSIM_OGL_PARAM_LIGHT1POS:
	    OGLWin -> OGLParams -> UpdateLightSource(0);
	    OGLWin -> OGLParams -> UpdateLightSource(1);

	    NCSimMF -> Refresh();	                 /* Force a redraw. */
	    break;
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Update OGL light source index Index.                                     M
*                                                                            *
* PARAMETERS:                                                                M
*   Index:  Of light source to update in OGL: 0 or 1.                        M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimOGLParamClass::UpdateLightSource                                    M
*****************************************************************************/
void NCSimOGLParamClass::UpdateLightSource(int Index)
{
    NCSIM_DEBUG_FUNC_ENTRY

    int i;
    float LightPos[4];
    IrtRType LightClr[3];

    switch (Index) {
	default:
	case 0:
	    for (i = 0; i < 4; i++)
	        LightPos[i] = Light0Pos[i];
	    for (i = 0; i < 3; i++)
	        LightClr[i] = Light0Clr[i] / 255.0;
	    break;
	case 1:
	    for (i = 0; i < 4; i++)
	        LightPos[i] = Light1Pos[i];
	    for (i = 0; i < 3; i++)
	        LightClr[i] = Light1Clr[i] / 255.0;
	    break;
    }
    IGSetLightSource(LightPos, LightClr, Index);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Traps events from text controllers.                                      M
*                                                                            *
* PARAMETERS:                                                                M
*   Event: N.S.F.I.                                                          M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimOGLParamClass::OnTextChange                                         M
*****************************************************************************/
void NCSimOGLParamClass::OnTextChange(wxCommandEvent &Event)
{
    NCSIM_DEBUG_FUNC_ENTRY

    int i;
    wxString
	Str = Event.GetString();
    const char *s;

    switch (Event.GetId()) {
	case wxID_NCSIM_OGL_PARAM_LINE_WIDTH:
	    s = NCSimwxString2Chars(Str);
	    if (sscanf(s, "%d", &i) == 1) {
	        IGGlblLineWidth = i;
		ThisNCSimOGL -> ForceDispListBuild = true;
	    }
	    break;
	case wxID_NCSIM_OGL_PARAM_FRAME_SKIP:
	    s = NCSimwxString2Chars(Str);
	    if (sscanf(s, "%d", &i) == 1) {
	        ThisNCSimOGL -> SkipFrameRate =
		    ThisNCSimOGL -> SkippedFrames = i;
	    }
	    break;
    }

    NCSimMF -> Refresh();		                 /* Force a redraw. */
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Traps events from check box.                                             M
*                                                                            *
* PARAMETERS:                                                                M
*   Event: N.S.F.I.                                                          M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimOGLParamClass::OnCheckBox                                           M
*****************************************************************************/
void NCSimOGLParamClass::OnCheckBox(wxCommandEvent &Event)
{
    NCSIM_DEBUG_FUNC_ENTRY

    switch (Event.GetId()) {
	case wxID_NCSIM_OGL_PARAM_DRAW_AXES:
	    ThisNCSimOGL -> DrawAxes = Event.IsChecked();
	    break;
	case wxID_NCSIM_OGL_PARAM_DRAW_TOOL:
	    ThisNCSimOGL -> DrawTool = Event.IsChecked();
	    break;
	case wxID_NCSIM_OGL_PARAM_DRAW_TPATH:
	    ThisNCSimOGL -> DrawToolPath = Event.IsChecked();
	    break;
	case wxID_NCSIM_OGL_PARAM_DRAW_STOCK:
	    ThisNCSimOGL -> DrawStock = Event.IsChecked();
	    break;
	case wxID_NCSIM_OGL_PARAM_DRAW_FINAL_PART:
	    ThisNCSimOGL -> DrawFinalPart = Event.IsChecked();
	    break;
	case wxID_NCSIM_OGL_PARAM_DRAW_OTHER_GEOM:
	    ThisNCSimOGL -> DrawOtherGeom = Event.IsChecked();
	    break;
    }

    NCSimMF -> Refresh();		                 /* Force a redraw. */
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Traps events from a button.		                                     M
*                                                                            *
* PARAMETERS:                                                                M
*   Event: N.S.F.I.                                                          M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimOGLParamClass::OnButton                                             M
*****************************************************************************/
void NCSimOGLParamClass::OnButton(wxCommandEvent &Event)
{
    NCSIM_DEBUG_FUNC_ENTRY

    wxButton
	*Btn = (wxButton *) Event.GetEventObject();

    if (Event.GetId() == wxID_NCSIM_OGL_PARAM_ORTHO_PERSP) {
        if (IGGlblViewMode == IG_VIEW_ORTHOGRAPHIC)
	    IGGlblViewMode = IG_VIEW_PERSPECTIVE;
	else
	    IGGlblViewMode = IG_VIEW_ORTHOGRAPHIC;

	Btn -> SetLabel(IGGlblViewMode ==
				   IG_VIEW_ORTHOGRAPHIC ? wxT("Orthographic")
						        : wxT("Perspective"));
    }
    else if (Event.GetId() == wxID_NCSIM_OGL_PARAM_SOLID_WIRE) {
        if (IGGlblDrawStyle == IG_STATE_DRAW_STYLE_SOLID)
	    IGGlblDrawStyle = IG_STATE_DRAW_STYLE_WIREFRAME;
	else
	    IGGlblDrawStyle = IG_STATE_DRAW_STYLE_SOLID;

	Btn -> SetLabel(IGGlblDrawStyle ==
				 IG_STATE_DRAW_STYLE_SOLID ? wxT("Solid")
							   : wxT("WireFrame"));

	ThisNCSimOGL -> ForceDispListBuild = true;
    }
    else if (Event.GetId() == wxID_NCSIM_OGL_PARAM_BG_COLOR ||
	     Event.GetId() == wxID_NCSIM_OGL_PARAM_LIGHT0CLR ||
	     Event.GetId() == wxID_NCSIM_OGL_PARAM_LIGHT1CLR) {
        wxColourData ClrData;
	int i, *Clr;
	IrtRType c;

	switch (Event.GetId()) {
	    case wxID_NCSIM_OGL_PARAM_BG_COLOR:
	        Clr = IGGlblBackGroundColor;
	        break;
	    case wxID_NCSIM_OGL_PARAM_LIGHT0CLR:
	        Clr = Light0Clr;
	        break;
	    case wxID_NCSIM_OGL_PARAM_LIGHT1CLR:
	        Clr = Light1Clr;
	        break;
	}

	c = 255.0 / (IRIT_MAX(IRIT_MAX(Clr[0], Clr[1]), Clr[2]) * 15.0);
	ClrData.SetChooseFull(true);
	ClrData.SetColour(wxColour(Clr[0], Clr[1], Clr[2]));
	for (i = 0; i < 16; i++) {
	    ClrData.SetCustomColour(i,
				    wxColour(Clr[0] * i * c,
					     Clr[1] * i * c,
					     Clr[2] * i * c));
	}

	wxColourDialog ClrDialog(this, &ClrData);
	if (ClrDialog.ShowModal() == wxID_OK) {
	    wxColourData
	        RetData = ClrDialog.GetColourData();
	    wxColour
	        wxColour = RetData.GetColour();

	    Clr[0] = wxColour.Red();
	    Clr[1] = wxColour.Green();
	    Clr[2] = wxColour.Blue();
	}

	SetButtonBGColor(Btn, Clr);

	switch (Event.GetId()) {
	    case wxID_NCSIM_OGL_PARAM_BG_COLOR:
	        break;
	    case wxID_NCSIM_OGL_PARAM_LIGHT0CLR:
	        UpdateLightSource(0);
	        break;
	    case wxID_NCSIM_OGL_PARAM_LIGHT1CLR:
	        UpdateLightSource(1);
	        break;
	}
    }

    NCSimMF -> Refresh();		                 /* Force a redraw. */
}
