/******************************************************************************
* NCSimZBuffer.h - Zbuffer simulation core.				      *
*******************************************************************************
* (C) Gershon Elber					                      *
*******************************************************************************
* Written by Gershon Elber, Dec 2006.					      *
******************************************************************************/

#include <stdio.h>

#include "irit_sm.h"
#include "grap_lib.h"

#include "NCSimZBuffer.h"
#include "NCSimOGL.h"
#include "NCSimulation.h"
#include "NCSimMain.h"

#define NCSIM_MIN_ACCUM_STEPS 5

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Constructor of the zbuffer simulator.                                    M
*                                                                            *
* PARAMETERS:                                                                M
*   ThisSim:	  The simulation that uses this Z buffer.                    M
*   ZBufferSize:  Size of this new Z buffer (Maximal size in either X or Y). M
*   StockBoundingBox:	  Dimensions of the stock to use here.	             M
*                                                                            *
* RETURN VALUE:                                                              M
*   NoValue                                                                  M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimZBufferClass::NCSimZBufferClass                                     M
*****************************************************************************/
NCSimZBufferClass::NCSimZBufferClass(class NCSimulationClass *ThisSim,
				     int ZBufferSize,
				     GMBBBboxStruct *StockBoundingBox)
{
    NCSIM_DEBUG_FUNC_ENTRY

    PObjGnrl = NULL;
    FinalPartGeom = NULL;

    TheSim = ThisSim;

    ZBufSize = ZBufXSize = ZBufYSize = ZBufferSize;

    StockWireGeom = NULL;
    StockBBox = *StockBoundingBox;
    StockColor[0] = 200;
    StockColor[1] = 175;
    StockColor[2] = 120;
    StockTransp = 0.0;

    AccumulatedVolume = AccumulatedTime = 0.0;
    AccumulatedValidStep = 0;

    ZBuf = NULL;

    UpdateObj2ScreenMap();

    ReCreateStockGeometry();
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Destructor of the  zbuffer simulator.                                    M
*                                                                            *
* PARAMETERS:                                                                M
*   None								     M
*                                                                            *
* RETURN VALUE:                                                              M
*   NoValue                                                                  M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimZBufferClass::~NCSimZBufferClass                                    M
*****************************************************************************/
NCSimZBufferClass::~NCSimZBufferClass()
{
    NCSIM_DEBUG_FUNC_ENTRY

    /* Free all the geometry. */
    if (PObjGnrl != NULL) {
        NCSimMF -> OGLWin -> CleanOGLObject(PObjGnrl);
        IPFreeObjectList(PObjGnrl);
    }

    if (FinalPartGeom != NULL) {
        NCSimMF -> OGLWin -> CleanOGLObject(FinalPartGeom);
	IPFreeObjectList(FinalPartGeom);
    }

    if (StockWireGeom != NULL) {
        NCSimMF -> OGLWin -> CleanOGLObject(StockWireGeom);
	IPFreeObjectList(StockWireGeom);
    }

    if (ZBuf != NULL) {
        NCSimMF -> OGLWin -> CleanOGLZBuf();
	INCRndrDestroy(ZBuf);
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Recreate the stock geometry from this object's parameters.               M
*                                                                            *
* PARAMETERS:                                                                M
*   None                                                                     M
*                                                                            *
* RETURN VALUE:                                                              M
*   bool:    True if geometry was (re)created, false otherwise.              M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimZBufferClass::ReCreateStockGeometry                                 M
*****************************************************************************/
bool NCSimZBufferClass::ReCreateStockGeometry()
{
    NCSIM_DEBUG_FUNC_ENTRY

    if (IRIT_PT_EQ_ZERO(StockBBox.Min) && IRIT_PT_EQ_ZERO(StockBBox.Max))
	return false;

    UpdateObj2ScreenMap();

    if (StockWireGeom != NULL)
	IPFreeObjectList(StockWireGeom);

    StockWireGeom = PrimGenBOXWIREObject(StockBBox.Min,
				         StockBBox.Max[0] - StockBBox.Min[0],
				         StockBBox.Max[1] - StockBBox.Min[1],
					 StockBBox.Max[2] - StockBBox.Min[2]);

    return true;
}


/*****************************************************************************
* DESCRIPTION:                                                               M
*   Update the mapping values from object to screen space.	             M
*                                                                            *
* PARAMETERS:                                                                M
*   None                                                                     M
*                                                                            *
* RETURN VALUE:                                                              M
*   void				                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimZBufferClass::UpdateObj2ScreenMap                                   M
*****************************************************************************/
void NCSimZBufferClass::UpdateObj2ScreenMap()
{
    NCSIM_DEBUG_FUNC_ENTRY

    if (IRIT_PT_EQ_ZERO(StockBBox.Min) && IRIT_PT_EQ_ZERO(StockBBox.Max))
	return;

    Obj2ScreenScale[0] = ZBufXSize / (StockBBox.Max[0] - StockBBox.Min[0]);
    Obj2ScreenScale[1] = ZBufYSize / (StockBBox.Max[1] - StockBBox.Min[1]);
    Obj2ScreenScale[2] = 1.0;

    Obj2ScreenTrans[0] = -StockBBox.Min[0];
    Obj2ScreenTrans[1] = -StockBBox.Min[1];
    Obj2ScreenTrans[2] = 0.0;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Initialize the Z buffer.				                     M
*                                                                            *
* PARAMETERS:                                                                M
*   GridSize:        Number of cells in ZBuffer grid, along finer direction. M
*   Tools:	     The tool we expect to use in this simulation.	     M
*   UniformAspectRatio:  TRUE to make pixel squares.			     M
*                                                                            *
* RETURN VALUE:                                                              M
*   bool:    TRUE if geometry was (re)created, false otherwise.              M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimZBufferClass::ZBufferInit	                                     M
*****************************************************************************/
bool NCSimZBufferClass::ZBufferInit(int GridSize,
				    NCSimToolClass *Tools,
				    CagdBType UniformAspectRatio)
{
    NCSIM_DEBUG_FUNC_ENTRY

    static IRndrColorType
	BackgrCol = { 0.0, 0.0, 0.0, 0.0 };
    int i, GridSizeX, GridSizeY, X1, X2, Y1, Y2;

    if (UniformAspectRatio) {
	IrtRType Dx, Dy;

	Dx = StockBBox.Max[0] - StockBBox.Min[0];
	Dy = StockBBox.Max[1] - StockBBox.Min[1];
	if (Dx > Dy) {
	    ZBufXSize = ZBufSize;
	    ZBufYSize = (int) (ZBufSize * Dy / Dx + 0.5);
	}
	else {
	    ZBufXSize = (int) (ZBufSize * Dx / Dy + 0.5);
	    ZBufYSize = ZBufSize;
	}

	UpdateObj2ScreenMap();
    }

    if (NCSimMF -> UseOGLDispLists) {
	if (ZBufXSize > ZBufYSize) {
	    GridSizeX = GridSize;
	    GridSizeY = IRIT_MAX(1, GridSize * ZBufYSize / ZBufXSize);
	}
	else {
	    GridSizeY = GridSize;
	    GridSizeX = IRIT_MAX(1, GridSize * ZBufXSize / ZBufYSize);
	}
    }
    else {
	GridSizeY = GridSizeX = 1;
    }

    /* Create proper masks for all the tools in this simulation. */
    for (i = 0; i < NCSIM_MAX_NUM_OF_TOOLS; i++) {
        if (Tools[i].Type != NCSIM_TOOL_TYPE_NONE &&
	    Tools[i].Geom != NULL)
	    ZBufferInitToolMask(&Tools[i]);
    }

    if (ZBuf != NULL)
	INCRndrDestroy(ZBuf);
    ZBuf = INCRndrInitialize(ZBufXSize, ZBufYSize, GridSizeX, GridSizeY,
			     StockBBox.Min, StockBBox.Max, TRUE);

    /* Render the initial geometry into the Zbuffer if has any. */
    if (StockGeomInit.Type != NCSIM_STOCK_GEOM_INIT_NONE)
        ZBufferInitStockGeom();

    /* Reset active cells by doing a dummy query. */
    ZBufferGetActiveCells(&X1, &Y1, &X2, &Y2);
    ZBufferUpdateCellsToRedraw(true);

    return true;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Renders the current initial stock geometry into the Zbuffer.             M
*                                                                            *
* PARAMETERS:                                                                M
*   None                                                                     M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimZBufferClass::ZBufferInitStockGeom                                  M
*****************************************************************************/
void NCSimZBufferClass::ZBufferInitStockGeom()
{
    NCSIM_DEBUG_FUNC_ENTRY

    IRIT_STATIC_DATA CagdVType
	Origin = { 0.0, 0.0, 0.0 };
    int i, j, VrtcsRvrsd, Faces, StockChanged;
    IrtRType *ZVals, MinZ, BndryMinZ;
    IRndrZBufferCmpType OldZBufCmp;
    CagdSrfStruct *Srf, *TSrf;
    CagdVType Dir;
    CagdRType Height;
    IrtHmgnMatType Mat1, Mat2;
    IrtVecType Vecs[4];
    IPObjectStruct *PObj, *PTris;

    switch (StockGeomInit.Type) {
	default:
	case NCSIM_STOCK_GEOM_INIT_NONE:
	    return;
	case NCSIM_STOCK_GEOM_INIT_PLANE:
	    if (StockGeomInit.Plane.ABCD[2] == 0.0) {
	        NCSimMF -> Printf(NCSIM_LOG_ERROR,
		    "Vertical plane to initialize stock geometry ignored\n");
	        return;
	    }

	    /* Create a polygon at the XY coordinates of the bbox and set */
	    /* the Z values to satisfy the plane constraints.		  */
	    IRIT_PT2D_COPY(Vecs[0], StockBBox.Min);
	    IRIT_PT2D_COPY(Vecs[2], StockBBox.Max);
	    Vecs[1][0] = Vecs[0][0];
	    Vecs[1][1] = Vecs[2][1];
	    Vecs[3][0] = Vecs[2][0];
	    Vecs[3][1] = Vecs[0][1];

	    /* Set Z values. */
	    for (i = 0; i < 4; i++)
	        Vecs[i][2] = (StockGeomInit.Plane.ABCD[0] * Vecs[i][0] +
			      StockGeomInit.Plane.ABCD[1] * Vecs[i][1] +
			      StockGeomInit.Plane.ABCD[3]) /
			     -StockGeomInit.Plane.ABCD[2];

	    PObj = IPGenPOLYObject(PrimGenPolygon4Vrtx(Vecs[0], Vecs[1],
						       Vecs[2], Vecs[3],
						       NULL, &VrtcsRvrsd,
						       NULL));
	    Faces = NCSIM_ZBUF_FRONT_FACE | NCSIM_ZBUF_BACK_FACE;
	    break;
        case NCSIM_STOCK_GEOM_INIT_CYLIN_CAP:
        case NCSIM_STOCK_GEOM_INIT_CYLIN_CAVITY:
	    IRIT_VEC_SUB(Dir, StockGeomInit.Cylin.Pt2, StockGeomInit.Cylin.Pt1);
	    Height = IRIT_PT_LENGTH(Dir);

	    TSrf = CagdPrimCylinderSrf(Origin, StockGeomInit.Cylin.Rad, Height,
				       TRUE, CAGD_PRIM_CAPS_NONE);
	    GMGenMatrixZ2Dir(Mat1, Dir);
	    MatGenMatTrans(StockGeomInit.Cylin.Pt1[0],
			   StockGeomInit.Cylin.Pt1[1],
			   StockGeomInit.Cylin.Pt1[2], Mat2);
	    MatMultTwo4by4(Mat1, Mat1, Mat2);
	    Srf = CagdSrfMatTransform(TSrf, Mat1);
	    CagdSrfFree(TSrf);

	    PObj = IPGenPOLYObject(IPSurface2Polygons(Srf, FALSE,
						      GeomApproxTolerance,
						      FALSE, FALSE, TRUE));
	    CagdSrfFree(Srf);

	    Faces = StockGeomInit.Type == NCSIM_STOCK_GEOM_INIT_CYLIN_CAP ?
				NCSIM_ZBUF_BACK_FACE : NCSIM_ZBUF_FRONT_FACE;
	    break;
	case NCSIM_STOCK_GEOM_INIT_SPHERE_CAP:
	case NCSIM_STOCK_GEOM_INIT_SPHERE_CAVITY:
	    Srf = CagdPrimSphereSrf(StockGeomInit.Sphere.Cntr,
				    StockGeomInit.Sphere.Rad, TRUE);

	    PObj = IPGenPOLYObject(IPSurface2Polygons(Srf, FALSE,
						      GeomApproxTolerance,
						      FALSE, FALSE, TRUE));
	    CagdSrfFree(Srf);

	    Faces = StockGeomInit.Type == NCSIM_STOCK_GEOM_INIT_SPHERE_CAP ?
				 NCSIM_ZBUF_BACK_FACE : NCSIM_ZBUF_FRONT_FACE;
	    break;
	case NCSIM_STOCK_GEOM_INIT_GEN_SRF:
	    PObj = StockGeomInit.GenSrf.Geom;

	    Faces = NCSIM_ZBUF_FRONT_FACE | NCSIM_ZBUF_BACK_FACE;
	    break;
    }

    /* Render PObj into the ZBuffer, regardless of the current Z values. */
    assert(PObj != NULL);
    OldZBufCmp = INCRndrSetZCmp(ZBuf, IRNDR_ZBUFFER_ALWAYS);

    PTris = GMConvertPolysToTriangles(PObj);
    if (StockGeomInit.Type != NCSIM_STOCK_GEOM_INIT_GEN_SRF)
        IPFreeObject(PObj);

    AttrSetObjectIntAttrib(PTris, "_TRANSFORMED", TRUE);
    ZBufferRenderObj(PTris, NULL, Faces);

    IPFreeObject(PTris);

    /* Examine the new, updated, Z buffer for too low values. */
    ZVals = (IrtRType *) IritMalloc(sizeof(IrtRType) * ZBufXSize);
    MinZ = BndryMinZ = IRIT_INFNTY;
    for (i = 0; i < ZBufYSize; i++) {
	INCRndrGetLineDepth(ZBuf, 0, ZBufXSize - 1, i, ZVals);
	for (j = 0; j < ZBufXSize; j++) {
	    if (MinZ > ZVals[j])
	        MinZ = ZVals[j];
	    if (i == 0 ||
		i == ZBufYSize - 1 ||
		j == 0 ||
		j == ZBufXSize - 1) {
	        if (BndryMinZ > ZVals[j])
		    BndryMinZ = ZVals[j];
	    }
	}
    }
    IritFree(ZVals);
    StockChanged = FALSE;
    if (MinZ < StockBBox.Min[2] &&
	NCSimMF -> PopupYesNo("Initial stock geometry is below stock bbox.\nCorrect stock bbox?", NCSIM_LOG_WARN)) {
        StockBBox.Min[2] = MinZ - 1.0;
	StockChanged = TRUE;
    }

    if (BndryMinZ < StockBBox.Max[2] &&
	NCSimMF -> PopupYesNo("Initial stock geometry penetrates stock bbox.\nCorrect stock bbox?", NCSIM_LOG_WARN)) {
        StockBBox.Max[2] = BndryMinZ - 1.0;
        StockBBox.Min[2] = StockBBox.Max[2] - 1.0;
	StockChanged = TRUE;
    }

    if (StockChanged) {
        ReCreateStockGeometry();
	if (TheSim -> AllInfo != NULL)
	    TheSim -> AllInfo -> Refresh();
    }

    /* Coerce the boundary of the Z buffer to go back to bbox height. */
    for (i = 0; i < ZBufXSize; i++) {
        INCRndrPutPixel(ZBuf, i, 0, StockBBox.Min[2]);
        INCRndrPutPixel(ZBuf, i, ZBufYSize - 1, StockBBox.Min[2]);
    }
    for (i = 1; i < ZBufYSize; i++) {
        INCRndrPutPixel(ZBuf, 0, i, StockBBox.Min[2]);
        INCRndrPutPixel(ZBuf, ZBufXSize - 1, i, StockBBox.Min[2]);
    }

    INCRndrSetZCmp(ZBuf, OldZBufCmp);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Initialize the tool mask of the given tool.		                     M
*                                                                            *
* PARAMETERS:                                                                M
*   Tool:	To initialize its Zbuffer mask.				     M
*                                                                            *
* RETURN VALUE:                                                              M
*   void								     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimZBufferClass::ZBufferInitToolMask                                   M
*****************************************************************************/
void NCSimZBufferClass::ZBufferInitToolMask(NCSimToolClass *Tool)
{
    NCSIM_DEBUG_FUNC_ENTRY

    int x, y, XMin, YMin, XMax, YMax;
    IrtRType *Mask,
	*ZValues = (IrtRType *) IritMalloc(sizeof(IrtRType) * ZBufXSize
						            * ZBufYSize);
    IrtPtType ToolPos;
    INCZBufferPtrType ToolZBuf;

    if (Tool -> Type == NCSIM_TOOL_TYPE_NONE ||
	Tool -> Geom == NULL ||
	(IRIT_PT_EQ_ZERO(StockBBox.Min) && IRIT_PT_EQ_ZERO(StockBBox.Max))) {
	IritFree(ZValues);
	return;
    }

    /* Create a Zbuffer to render the tool and create the mask into.       */
    /* Temporary replace our main Zbuffer, ZBuf, with this one.            */
    ToolZBuf = INCRndrInitialize(ZBufXSize, ZBufYSize, 1, 1,
				 StockBBox.Min, StockBBox.Max, TRUE);
    IRIT_SWAP(INCZBufferPtrType, ToolZBuf, ZBuf);

    /* Render the tool at the middle of the Zbuffer at the bottom level.   */
    IRIT_PT2D_BLEND(ToolPos, StockBBox.Min, StockBBox.Max, 0.5);
    ToolPos[2] = StockBBox.Min[2];

    if (Tool -> Mask != NULL) {    /* Make sure the mask is not used here. */
        IritFree(Tool -> Mask);
	Tool -> Mask = NULL;
    }
    ZBufferRenderNCTool(Tool, ToolPos);

    /* Fetch the Z buffer amd search for all affected pixels - these are    */
    /* our sought mask.  Build a bbox around them and save all as mask.     */
    /*   Make sure to skip boundary that might be coerced to the bottom.    */
    ZBufferGetCellZVals(0, 0, ZValues, &XMin, &YMin, &XMax, &YMax);
    XMin = YMin = IRIT_MAX_INT;
    XMax = YMax = -IRIT_MAX_INT;
    for (y = 1; y < ZBufYSize - 1; y++) {
        int y1 = y * ZBufXSize;

	for (x = 1; x < ZBufXSize - 1; x++) {
	    if (!IRIT_APX_EQ(ZValues[x + y1], StockBBox.Max[2])) {
	        if (XMin > x)
		    XMin = x;
		if (XMax < x)
		    XMax = x;
		if (YMin > y)
		    YMin = y;
		if (YMax < y)
		    YMax = y;
	    }
	}
    }

    /* Create the mask. */
    Tool -> MaskX = XMax - XMin + 1;
    Tool -> MaskY = YMax - YMin + 1;
    Tool -> Mask = Mask = (IrtRType *) IritMalloc(sizeof(IrtRType) *
					      Tool -> MaskX * Tool -> MaskY);

    for (y = YMin; y <= YMax; y++) {
        int y1 = y * ZBufXSize;

	for (x = XMin; x <= XMax; x++) {
	    *Mask++ = IRIT_APX_EQ(ZValues[x + y1], StockBBox.Max[2]) ?
			  IRIT_INFNTY : (ZValues[x + y1] - StockBBox.Min[2]);
	}
    }

    IRIT_SWAP(INCZBufferPtrType, ToolZBuf, ZBuf);
    INCRndrDestroy(ToolZBuf);
    IritFree(ZValues);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Render the tool into the Z buffer at the designated tool position.       M
*                                                                            *
* PARAMETERS:                                                                M
*   Tool:         The tool to render.					     M
*   ToolPos:	  The current tool position.				     M
*                                                                            *
* RETURN VALUE:                                                              M
*   bool:    True if geometry was rendered, false otherwise.                 M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimZBufferClass::ZBufferRenderNCTool                                   M
*****************************************************************************/
bool NCSimZBufferClass::ZBufferRenderNCTool(NCSimToolClass *Tool,
					    IrtPtType ToolPos)
{
    NCSIM_DEBUG_FUNC_ENTRY

    int j;
    IPObjectStruct *ToolGeom;

    if (ZBuf == NULL)
	return false;

    if (Tool -> Mask != NULL) {
        int ZBufPos[2];
	IrtRType Volume;

	for (j = 0; j < 2; j++)
	    ZBufPos[j] = (int) (0.5 + (ToolPos[j] + Obj2ScreenTrans[j]) *
							  Obj2ScreenScale[j]);

	Volume = INCRndrPutMask(ZBuf, ZBufPos, ToolPos[2], Tool -> Mask,
				Tool -> MaskX, Tool -> MaskY);

	/* Check if exceeding material removal rate. */
	if (TheSim != NULL && Volume > 0) {
	    AccumulatedTime += TheSim -> NCSimStep * TheSim -> AnimMultFactor;
	    AccumulatedVolume += Volume;

	    if (++AccumulatedValidStep >= NCSIM_MIN_ACCUM_STEPS) {
	        /* Removed volume is returned in pixels, in XY.  Convert     */
	        /* to object space values, per second.                       */
		TheSim -> MaterialRemovalRate = AccumulatedVolume /
			(Obj2ScreenScale[0] * Obj2ScreenScale[1] *
			                                      AccumulatedTime);

		if (TheSim -> MaterialRemovalRate >
		                           TheSim -> MaxMaterialRemovalRate) {
		    if (TheSim -> AnimGCodeInfo -> GCodeLineNumber !=
			                TheSim -> LastErrorGCodeLineNumber) {
		        NCSimPrintf(NCSIM_LOG_ERROR,
			    "Material removal rate too high (%.3g) in \"%s\", line %d, \"%s\"%s%s.\n",
			    TheSim -> MaterialRemovalRate,
			    TheSim -> NCFileName,
			    TheSim -> AnimGCodeInfo -> StreamLineNumber,
			    TheSim -> AnimGCodeInfo -> Line,
			    TheSim -> PauseOnError ? ", paused" : "",
			    TheSim -> UpdateFeedRates ? ", updating feedrate" : "");

			TheSim -> LastErrorGCodeLineNumber =
			           TheSim -> AnimGCodeInfo -> GCodeLineNumber;
		    }

		    if (TheSim -> UpdateFeedRates) {
		        IrtRType
			    NewFD = TheSim -> AnimGCodeInfo -> FeedRate *
					    TheSim -> MaxMaterialRemovalRate /
					    TheSim -> MaterialRemovalRate;

			TheSim -> AnimGCodeInfo  -> UpdatedFeedRate =
			    IRIT_MIN(TheSim -> AnimGCodeInfo -> UpdatedFeedRate,
				     NewFD);
			    
			/* Round the feedrate down, to a slower feedrate. */
			TheSim -> AnimGCodeInfo  -> UpdatedFeedRate = 0.1 *
			    ((int) (TheSim -> AnimGCodeInfo  -> UpdatedFeedRate * 10.0));
		    }

		    if (TheSim -> PauseOnError)
		        TheSim -> AnimState(NCSIM_ANIM_STATE_PAUSE);
		}

		AccumulatedTime = AccumulatedVolume = 0.0;
		AccumulatedValidStep = 0;
	    }
	}

	return true;
    }
    else if (Tool -> Geom != NULL) {
        ToolGeom = AttrGetObjectObjAttrib(Tool -> Geom, "ToolBottom");

	return ZBufferRenderObj(ToolGeom, ToolPos);
    }
    else
        return false;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Render the given object into the Z buffer at the designated position.    M
*                                                                            *
* PARAMETERS:                                                                M
*   PObj:         The object to render.					     M
*   Pos:	  Possible position to reposition the object. Can be NULL.   M
*   DrawFrontBackFaces:  Mask to control front/back face drawings.           M
*                                                                            *
* RETURN VALUE:                                                              M
*   bool:    True if geometry was rendered, false otherwise.                 M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimZBufferClass::ZBufferRenderObj                                      M
*****************************************************************************/
bool NCSimZBufferClass::ZBufferRenderObj(IPObjectStruct *PObj,
					 IrtPtType Pos,
					 int DrawFrontBackFaces)
{
    NCSIM_DEBUG_FUNC_ENTRY

    int i, j;
    IPPolygonStruct *Pl;

    INCRndrBeginObject(ZBuf, PObj);

    for (Pl = PObj -> U.Pl; Pl != NULL; Pl = Pl -> Pnext) {
        IrtPtType OrigCoord[3];
	IPVertexStruct *V;

	/* Update the triangle with translation, after keeping original. */
	for (i = 0, V = Pl -> PVertex; i < 3; i++, V = V -> Pnext) {
	    IRIT_PT_COPY(OrigCoord[i], V -> Coord);

	    if (Pos != NULL) /* If we have a repositioning value, do it now. */
	        IRIT_PT_ADD(V -> Coord, V -> Coord, Pos);

	    /* And also map to screen space in XY. */
	    for (j = 0; j < 2; j++)
	        V -> Coord[j] = (V -> Coord[j] + Obj2ScreenTrans[j])
							 * Obj2ScreenScale[j];
	}

	if (Pl -> Plane[2] > 0) {
	    if (DrawFrontBackFaces & NCSIM_ZBUF_FRONT_FACE)
	        INCRndrPutTriangle(ZBuf, Pl);
	}
	else {
	    if (DrawFrontBackFaces & NCSIM_ZBUF_BACK_FACE)
	        INCRndrPutTriangle(ZBuf, Pl);
	}

	/* Restore original positions. */
	for (i = 0, V = Pl -> PVertex; i < 3; i++, V = V -> Pnext) {
	    IRIT_PT_COPY(V -> Coord, OrigCoord[i]);
	}
    }

    INCRndrEndObject(ZBuf);

    return true;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Fetch the Z buffer grid's dimensions.		                     M
*                                                                            *
* PARAMETERS:                                                                M
*   GridSizeX, GridSizeY:          The grid dimensions.			     M
*   GridCellSizeX, GridCellSizeY:  Cell sizes.			             M
*                                                                            *
* RETURN VALUE:                                                              M
*   void			                                             M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimZBufferClass::ZBufferGetGridSize                                    M
*****************************************************************************/
void NCSimZBufferClass::ZBufferGetGridSize(int *GridSizeX,
					   int *GridSizeY,
					   int *GridCellSizeX,
					   int *GridCellSizeY)
{
    NCSIM_DEBUG_FUNC_ENTRY

    if (ZBuf == NULL) {
        *GridSizeX = *GridSizeY = *GridCellSizeX = *GridCellSizeY = 0;
	return;
    }

    INCRndrGetZbufferGridCellMaxSize(ZBuf, GridSizeX, GridSizeY,
				     GridCellSizeX, GridCellSizeY);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Fetch the Z buffer Z values of the prescribed grid cell.                 M
*                                                                            *
* PARAMETERS:                                                                M
*   GridX, GridY:  The desired Cell to fetch its Z values.		     M
*   ZValues:	   Buffer, sufficiently large to hold all the Z values.      M
*   XMin, YMin:    Minimal dimensions of fetched grid cell.		     M
*   XMax, YMax:    Maximal dimensions of fetched grid cell.		     M
*                                                                            *
* RETURN VALUE:                                                              M
*   bool:	TRUE if successful, FALSE otherwise.		             M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimZBufferClass::ZBufferGetCellZVals                                   M
*****************************************************************************/
bool NCSimZBufferClass::ZBufferGetCellZVals(int GridX,
					    int GridY,
					    IrtRType *ZValues,
					    int *XMin,
					    int *YMin,
					    int *XMax,
					    int *YMax)
{
    NCSIM_DEBUG_FUNC_ENTRY

    if (ZBuf == NULL) {
        *XMin = *YMin = 1;
	*XMax = *YMax = 0;
	return false;
    }

    return INCRndrGetZbufferGridCell(ZBuf, GridX, GridY, ZValues,
				     XMin, YMin, XMax, YMax) ? true : false;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Fetch the Z buffer geometry of the prescribed grid cell, as triangles.   M
* The triangles are send via a call back function.  The triangles are 	     M
* returns so every new triangle shares two vertices with the previous	     M
* triangle, forming triangle-fans as much as possible.		             M
*                                                                            *
* PARAMETERS:                                                                M
*   GridX, GridY:  The desired Cell to fetch its triangles.		     M
*   TriFunc:       Call back function to invoke with every triangle.         M
*                                                                            *
* RETURN VALUE:                                                              M
*   int:	Number of triangles processed.			             M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimZBufferClass::ZBufferGetCellTris                                    M
*****************************************************************************/
int NCSimZBufferClass::ZBufferGetCellTris(int GridX,
					  int GridY,
					  NCSimZBufTriangleCBFuncType TriFunc)
{
    NCSIM_DEBUG_FUNC_ENTRY

    int GridSizeX, GridSizeY, GridCellSizeX, GridCellSizeY,
	NCZBufXMin, NCZBufYMin, NCZBufXMax, NCZBufYMax,
	n = 0;
    IrtRType *ZValues;

    if (ZBuf == NULL)
	return 0;	     /* No Zbuffer and hence no triangles either... */

    ZBufferGetGridSize(&GridSizeX, &GridSizeY, &GridCellSizeX, &GridCellSizeY);
    ZValues = (IrtRType *) IritMalloc(sizeof(IrtRType) * GridCellSizeX
				      		       * GridCellSizeY);

    if (ZBufferGetCellZVals(GridX, GridY, ZValues,
			    &NCZBufXMin, &NCZBufYMin,
			    &NCZBufXMax, &NCZBufYMax)) {
	int i, j,
	    NCZBufDx = NCZBufXMax - NCZBufXMin + 1,
	    NCZBufDy = NCZBufYMax - NCZBufYMin + 1;
	IrtRType XMin, YMin, Dx, Dy, x, y, x1, y1, PixelDx, PixelDy;

	/* Compute cells size and position, in real object coordinates. */
	Dx = (StockBBox.Max[0] - StockBBox.Min[0]) / GridSizeX;
	Dy = (StockBBox.Max[1] - StockBBox.Min[1]) / GridSizeY;
	XMin = Dx * GridX + StockBBox.Min[0];
	YMin = Dy * GridY + StockBBox.Min[1];

	/* Compute the x/y object space step for each x/y pixel step. */
	PixelDx = Dx / (NCZBufXMax - NCZBufXMin);
	PixelDy = Dy / (NCZBufYMax - NCZBufYMin);

	for (j = 0; j < NCZBufDy - 1; j++) {
	    y = YMin + PixelDy * j;
	    y1 = y + PixelDy;
	    for (i = 0; i < NCZBufDx - 1; i++) {
	        x = XMin + PixelDx * i; 
		x1 = x + PixelDx;
	        TriFunc( x, y1, ZValues[NCZBufDx * (j + 1) + i],
			 x,  y, ZValues[NCZBufDx * j + i],
			x1, y1, ZValues[NCZBufDx * (j + 1) + i + 1]);
	        TriFunc(x1, y1, ZValues[NCZBufDx * (j + 1) + i + 1],
			 x,  y, ZValues[NCZBufDx * j + i],
			x1,  y, ZValues[NCZBufDx * j + i + 1]);
	    }
	}
	n = (NCZBufDx - 1) * (NCZBufDy - 1) * 2;
    }

    IritFree(ZValues);

    return n;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Fetch the Z buffer geometry of the prescribed grid cell, as quad strips. M
* The geometry is send via a call back function, as pairs of vertices each   M
* time, defining a new quad.					             M
*                                                                            *
* PARAMETERS:                                                                M
*   GridX, GridY:  The desired Cell to fetch its triangles.		     M
*   QSFunc:        Call back function to invoke and render the quad strips.  M
*                                                                            *
* RETURN VALUE:                                                              M
*   int:	Number of polygons processed.			             M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimZBufferClass::ZBufferGetCellQuadStrip                               M
*****************************************************************************/
int NCSimZBufferClass::ZBufferGetCellQuadStrip(int GridX,
					       int GridY,
					  NCSimZBufQuadStripCBFuncType QSFunc)
{
    NCSIM_DEBUG_FUNC_ENTRY

    int GridSizeX, GridSizeY, GridCellSizeX, GridCellSizeY,
	NCZBufXMin, NCZBufYMin, NCZBufXMax, NCZBufYMax,
	n = 0;
    IrtRType *ZValues;

    if (ZBuf == NULL)
	return 0;	     /* No Zbuffer and hence no triangles either... */

    ZBufferGetGridSize(&GridSizeX, &GridSizeY, &GridCellSizeX, &GridCellSizeY);
    ZValues = (IrtRType *) IritMalloc(sizeof(IrtRType) * GridCellSizeX
				      		       * GridCellSizeY);

    if (ZBufferGetCellZVals(GridX, GridY, ZValues,
			    &NCZBufXMin, &NCZBufYMin,
			    &NCZBufXMax, &NCZBufYMax)) {
	int i, j,
	    NCZBufDx = NCZBufXMax - NCZBufXMin + 1,
	    NCZBufDy = NCZBufYMax - NCZBufYMin + 1;
	IrtRType XMin, YMin, Dx, Dy, x, y, y1, PixelDx, PixelDy;

	/* Compute cells size and position, in real object coordinates. */
	Dx = (StockBBox.Max[0] - StockBBox.Min[0]) / GridSizeX;
	Dy = (StockBBox.Max[1] - StockBBox.Min[1]) / GridSizeY;
	XMin = Dx * GridX + StockBBox.Min[0];
	YMin = Dy * GridY + StockBBox.Min[1];

	/* Compute the x/y object space step for each x/y pixel step. */
	PixelDx = Dx / (NCZBufXMax - NCZBufXMin);
	PixelDy = Dy / (NCZBufYMax - NCZBufYMin);

	for (j = 0; j < NCZBufDy - 1; j++) {
	    y = YMin + PixelDy * j;
	    y1 = y + PixelDy;
	    for (i = 0; i < NCZBufDx; i++) {
	        x = XMin + PixelDx * i; 
	        QSFunc(x, y1, ZValues[NCZBufDx * (j + 1) + i],
		       x,  y, ZValues[NCZBufDx * j + i],
		       i == NCZBufDx - 1);
	    }
	}
	n = (NCZBufDx - 1) * (NCZBufDy - 1);
    }

    IritFree(ZValues);

    return n;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Fetch the Z buffer's cell indices of cells that were modified since the  M
* last redraw, and keep to for the next time we redraw the screen.	     M
*                                                                            *
* PARAMETERS:                                                                M
*   Reset:	True to reset the cells that need redraw.	 	     M
*                                                                            *
* RETURN VALUE:                                                              M
*   void								     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimZBufferClass::ZBufferUpdateCellsToRedraw                            M
*****************************************************************************/
void NCSimZBufferClass::ZBufferUpdateCellsToRedraw(bool Reset)
{
    NCSIM_DEBUG_FUNC_ENTRY

    int XMin, YMin, XMax, YMax;

    if (Reset) {
        RedrawCellXMin = RedrawCellYMin = IRIT_MAX_INT;
	RedrawCellXMax = RedrawCellYMax = -IRIT_MAX_INT;
	return;
    }

    ZBufferGetActiveCells(&XMin, &YMin, &XMax, &YMax);

    if (RedrawCellXMin > XMin)
        RedrawCellXMin = XMin;
    if (RedrawCellXMax < XMax)
        RedrawCellXMax = XMax;
    if (RedrawCellYMin > YMin)
        RedrawCellYMin = YMin;
    if (RedrawCellYMax < YMax)
        RedrawCellYMax = YMax;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Fetch the Z buffer's cell indices of cells that were modified since the  M
* last call to this function, region that is denoted active region.          M
*                                                                            *
* PARAMETERS:                                                                M
*   XMin, YMin:       Minimal dimensions of active region.		     M
*   XMax, YMax:       Maximal dimensions of active region.		     M
*                                                                            *
* RETURN VALUE:                                                              M
*   bool:	TRUE if successful, FALSE otherwise.		             M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimZBufferClass::ZBufferGetActiveCells                                 M
*****************************************************************************/
bool NCSimZBufferClass::ZBufferGetActiveCells(int *XMin,
					      int *YMin,
					      int *XMax,
					      int *YMax)
{
    NCSIM_DEBUG_FUNC_ENTRY

    int RetVal;
    IrtRType MaterialRemoved;

    if (ZBuf == NULL) {
	/* No active cells as we have no ZBuffer. */
	*XMin = *YMin = IRIT_MAX_INT;
	*XMax = *YMax = -IRIT_MAX_INT;
	return FALSE;
    }

    RetVal = INCRndrGetActiveCells(ZBuf, XMin, YMin, XMax, YMax,
				   &MaterialRemoved);

    /* Check if we cut material in G0 fast motion. */
    if (*XMin <= *XMax &&
	*YMin <= *YMax &&
	TheSim -> AnimGCodeInfo != NULL &&
	!TheSim -> AnimGCodeInfo -> IsVerticalUpMotion &&
	 TheSim -> AnimGCodeInfo -> GCodeType ==
					IP_NC_GCODE_LINE_MOTION_G0FAST &&
	 TheSim -> AnimGCodeInfo -> GCodeLineNumber !=
					TheSim -> LastErrorGCodeLineNumber) {
	NCSimPrintf(NCSIM_LOG_ERROR,
		    "Cutting material in G0, \"%s\" line %d, \"%s\"%s.\n",
		    TheSim -> NCFileName,
		    TheSim -> AnimGCodeInfo -> GCodeLineNumber,
		    TheSim -> AnimGCodeInfo -> Line,
		    TheSim -> PauseOnError ? " (paused)" : "");

	TheSim -> LastErrorGCodeLineNumber = 
	    TheSim -> AnimGCodeInfo -> GCodeLineNumber;

	if (TheSim -> PauseOnError)
	    TheSim -> AnimState(NCSIM_ANIM_STATE_PAUSE);
    }

    return RetVal ? true : false;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Update the geometry of the prescribed type.                              M
*                                                                            *
* PARAMETERS:                                                                M
*   PObjs:    New object to update and add their geometry in.                M
*   ObjType:  The type of the new objects.                                   M
*   Append:   TRUE to append to existing geometry of the same type,          M
*	      FALSE to relace it.					     M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* SEE ALSO:                                                                  M
*                                                                            M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimZBufferClass::UpdateObjGeometry                                     M
*****************************************************************************/
void NCSimZBufferClass::UpdateObjGeometry(IPObjectStruct *PObjs,
					  NCSimObjectType ObjType,
					  bool Append)
{
    NCSIM_DEBUG_FUNC_ENTRY

    IPObjectStruct **OldGeom, *OldToolGeom;

    switch (ObjType) {
	case NCSIM_OBJ_GEN:
	    OldGeom = &PObjGnrl;
	    break;
	case NCSIM_OBJ_TOOL:
	    /* The tool is not a slot here so access it with care. */
	    OldToolGeom = TheSim -> Tools[TheSim -> ToolIdx].Geom;
	    if (Append) {
	        IPGetLastObj(OldToolGeom) -> Pnext = IPCopyObjectList(PObjs,
								      FALSE);
	    }
	    else {
	        IPFreeObjectList(OldToolGeom);
		OldToolGeom = IPCopyObjectList(PObjs, FALSE);
	    }
	    TheSim -> Tools[TheSim -> ToolIdx].Geom = OldToolGeom;

	    /* We have a general type tool. */
	    TheSim -> Tools[TheSim -> ToolIdx].Type = NCSIM_TOOL_TYPE_OTHER;
	    return;
	case NCSIM_OBJ_STOCK:
	    OldGeom = &StockWireGeom;
	    /* Update Stock bbox. */ 
	    StockBBox = *GMBBComputeBboxObjectList(PObjs);
	    UpdateObj2ScreenMap();
	    break;
	case NCSIM_OBJ_PART:
	    OldGeom = &FinalPartGeom;
	    break;
	default:
	    NCSimMF -> Printf(NCSIM_LOG_ERROR,
			    "Invalid geometry to update into the simulator\n");
	    return;
    }

    if (*OldGeom == NULL) {
        *OldGeom = IPCopyObjectList(PObjs, FALSE);
    }
    else if (Append) {
        IPGetLastObj(*OldGeom) -> Pnext = IPCopyObjectList(PObjs, FALSE);
    }
    else {
        IPFreeObjectList(*OldGeom);
	*OldGeom = IPCopyObjectList(PObjs, FALSE);
    }
}
