/******************************************************************************
* NCSimulation.cpp - Holds the main modules of the simulation of NC code.     *
*******************************************************************************
* (C) Gershon Elber					                      *
*******************************************************************************
* Written by Gershon Elber, Dec 2006.					      *
******************************************************************************/

#include <wx/combobox.h>
#include <wx/stattext.h>
#include <wx/spinctrl.h>
#include <wx/textctrl.h>
#include <wx/colordlg.h>

#include "irit_sm.h"
#include "grap_lib.h"

#include "NCSimMain.h"
#include "NCSimulation.h"
#include "NCSimOGL.h"

#include "RealSpinCtrl.h"

#define NCSIM_MINIMAL_FRAME_TIME   0.05    /* No more than 20 FPS in anims. */
#define NCSIM_DEF_TOOL_APPROX_REL_TOL  0.2     /* Default tool approx. tol. */
#define NCSIM_ARC_LEN_TRIGGER      (Units == NCSIM_UNIT_MM ? 1.0 : 1.0 / 25.4)

IRIT_STATIC_DATA const IritConfigStruct NCSimCfgSetUpMaster[] = {
        { "RelArcCenters", "", (VoidPtr) NULL, IC_BOOLEAN_TYPE },
        { "ReverseZDir",   "", (VoidPtr) NULL, IC_BOOLEAN_TYPE },
        { "NCZBuffer",     "", (VoidPtr) NULL, IC_BOOLEAN_TYPE },
        { "PauseOnError",  "", (VoidPtr) NULL, IC_BOOLEAN_TYPE },
        { "UpdateFeedRate","", (VoidPtr) NULL, IC_BOOLEAN_TYPE },
	{ "ToolIdx",       "", (VoidPtr) NULL, IC_INTEGER_TYPE },
        { "NCZBufferRes",  "", (VoidPtr) NULL, IC_INTEGER_TYPE },
	{ "ToolFastFactor","", (VoidPtr) NULL, IC_REAL_TYPE },
	{ "MaxRemoveRate", "", (VoidPtr) NULL, IC_REAL_TYPE },
	{ "GeomApprox",    "", (VoidPtr) NULL, IC_REAL_TYPE },
	{ "FeedRate",	   "", (VoidPtr) NULL, IC_REAL_TYPE },
	{ "SpindleSpeed",  "", (VoidPtr) NULL, IC_REAL_TYPE },
	{ "SimFixedStep",  "", (VoidPtr) NULL, IC_REAL_TYPE },
	{ "StockTransp",   "", (VoidPtr) NULL, IC_REAL_TYPE },
	{ "ToolIdleColor", "", (VoidPtr) NULL, IC_STRING_TYPE },
	{ "ToolFastColor", "", (VoidPtr) NULL, IC_STRING_TYPE },
	{ "ToolCutColor",  "", (VoidPtr) NULL, IC_STRING_TYPE },
	{ "StockColor",    "", (VoidPtr) NULL, IC_STRING_TYPE },
	{ "StockMin",      "", (VoidPtr) NULL, IC_STRING_TYPE },
	{ "StockMax",      "", (VoidPtr) NULL, IC_STRING_TYPE },
	{ "StockGeomInit", "", (VoidPtr) NULL, IC_STRING_TYPE },
	{ "SimMode",	   "", (VoidPtr) NULL, IC_STRING_TYPE },
	{ "Units",	   "", (VoidPtr) NULL, IC_STRING_TYPE },
	{ "Tools",	   "", (VoidPtr) NULL, IC_MULTI_STR_TYPE }
    };
IRIT_STATIC_DATA const int
    NCSIM_CFG_NUM_SET_UP =
		       sizeof(NCSimCfgSetUpMaster) / sizeof(IritConfigStruct);
IRIT_STATIC_DATA int
    GCodeIndexChanged = FALSE,
    RealSpinUpdateInCB = TRUE;

static void RealSpinUpdateValCB(double Val, wxWindowID Id);
static void SetButtonBGColor(wxWindow *Btn, int *Clr);
static void TrackGCodeIndicesChangeCB(void);
static void AddRSPinWidget(wxBoxSizer *TopSizer,
			   wxDialog *Dialog,
			   const char *Title,
			   IrtRType *R);

/*****************************************************************************
* AUXILIARY:								     *
*    Event table for NCSimulationClass					     *
*****************************************************************************/
BEGIN_EVENT_TABLE(NCSimulationClass, wxTextCtrl)
    EVT_IDLE(NCSimulationClass::OnIdle)
END_EVENT_TABLE()

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Constructor of the NC simulation class.  initialize the two inherited    M
* classes of the text control and of the Z buffer.                           M
*                                                                            *
* PARAMETERS:                                                                M
*   NCFName:                The file name of the NC G code.                  M
*   Parent:                  Frame window.			             M
*   Id:                     ID assigned to window.                           M
*   Value:                  Default initial text value.                      M
*   Flags:                  Window styles' flags.		             M
*   ZBufSize:   	    Size of Z buffer.                                M
*   StockBBox:              Stock bounding box.                              M
*   SimToolType:            Type of tool to use.                             M
*   SimToolDiameter:        Diameter of used tool.                           M
*   SimUnits:               Mm or inches.                                    M
*                                                                            *
* RETURN VALUE:                                                              M
*   NoValue                                                                  M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimulationClass::NCSimulationClass                                     M
*****************************************************************************/
NCSimulationClass::NCSimulationClass(const char *NCFName,
				     /* Text control flags. */
				     wxWindow *Parent,
				     wxWindowID Id,
				     const wxString &Value,
				     int Flags,
				     /* NCZBuffer flags. */
				     int ZBufSize,
				     GMBBBboxStruct *StockBBox,
				     NCSimToolType SimToolType,
				     IrtRType SimToolDiameter,
				     NCSimUnitsType SimUnits):
	wxTextCtrl(Parent,
		   Id,
		   Value,
		   wxDefaultPosition,
		   wxDefaultSize,
		   Flags | wxTE_MULTILINE | wxTE_READONLY),
	NCSimZBufferClass(this,
			  ZBufSize,
			  StockBBox)
{
    NCSIM_DEBUG_FUNC_ENTRY

    /* A verification we are not invading other name spaces. */
    assert(wxID_NCSIM_PARAM_LAST - wxID_NCSIM_PARAMS < wxID_NCSIM_PARAMS_SIZE);

    NCFileName = NCFName == NULL ? NULL : IritStrdup(NCFName);
    GStream = NULL;
    GPathGeom = NULL;
    MatGenUnitMat(NCSimViewMat);
    HasRelArcCenter = FALSE;
    ReverseZDir = FALSE;

    Units = SimUnits;

    ToolIdx = 0;
    Tools[ToolIdx].Type = SimToolType;
    Tools[ToolIdx].Diameter = SimToolDiameter;
    Tools[ToolIdx].TorusRadius = Tools[ToolIdx].Diameter / 4.0;
    Tools[ToolIdx].Geom = NULL;
    Tools[ToolIdx].Mask = NULL;
    Tools[ToolIdx].MaskX = Tools[ToolIdx].MaskY = 0;

    ToolIdx = 1;
    Tools[ToolIdx].Type = SimToolType;
    Tools[ToolIdx].Diameter = SimToolDiameter;
    Tools[ToolIdx].TorusRadius = Tools[ToolIdx].Diameter / 4.0;
    Tools[ToolIdx].Geom = NULL;
    Tools[ToolIdx].Mask = NULL;
    Tools[ToolIdx].MaskX = Tools[ToolIdx].MaskY = 0;

    ToolFastColor[0] = 20;
    ToolFastColor[1] = 200;
    ToolFastColor[2] = 20;
    ToolCutColor[0] = 200;
    ToolCutColor[1] = 20;
    ToolCutColor[2] = 20;
    ToolIdleColor[0] = 0;
    ToolIdleColor[1] = 0;
    ToolIdleColor[2] = 128;

    FeedRate = 1.0;
    SpindleSpeed = 5000.0;
    NCFileLineNum = -1;
    NCSimMode = NCSIM_MODE_TIME;
    NCZBufferOn = FALSE;
    PauseOnError = FALSE;
    UpdateFeedRates = FALSE;
    GeomApproxTolerance = Tools[ToolIdx].Diameter / 10.0;
    NCSimStep = Tools[ToolIdx].Diameter / 4.0;

    LastErrorGCodeLineNumber = -1;

    GUITextPage = -1;

    ReCreateToolGeometry();

    AnimCrntState = NCSIM_ANIM_STATE_STOP;
    AnimMultFactor = 1.0;
    AnimFastFeedRateMultFactor = 10.0;
    AnimArcLen = AnimRealTime = 0.0;
    AnimTotalArcLen = -1.0;  /* Should be initialized by the G Code stream. */
    IRIT_PT_RESET(AnimToolPosition);
    AnimGCodeInfo = NULL;

    SingleStepActive = FALSE;

    NCFileLineMap = NULL;
    NCFileNumLines = 1;

    MaterialRemovalRate = 0.0;
    MaxMaterialRemovalRate = 10.0;

    IRIT_ZAP_MEM(&StockGeomInit, sizeof(StockGeomInit));
    StockGeomInit.Type = NCSIM_STOCK_GEOM_INIT_NONE;

    LoadSimConfig();            /* If there is a "*_ncsim.cfg", load it in. */

    LoadNCInfo = new NCSimLoadNCInfoClass(this, NCSimMF, wxID_ANY,
					  wxT("NC File load setup"));
    OverwriteDefaults = FALSE;/* Made TRUE if any default value is changed. */

    AllInfo = NULL;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Destructor of the text control class.                                    M
*                                                                            *
* PARAMETERS:                                                                M
*   None					                             M
*                                                                            *
* RETURN VALUE:                                                              M
*   NoValue                                                                  M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimulationClass::~NCSimulationClass                                    M
*****************************************************************************/
NCSimulationClass::~NCSimulationClass()
{
    NCSIM_DEBUG_FUNC_ENTRY

    IritFree(NCFileName);

    if (GStream != NULL)
	IPNCGCodeParserFree(GStream);

    if (GPathGeom != NULL) {
        NCSimMF -> OGLWin -> CleanOGLObject(GPathGeom);
        IPFreeObject(GPathGeom);
    }

    if (NCFileLineMap != NULL)
        IritFree(NCFileLineMap);

    delete LoadNCInfo;

    if (StockGeomInit.Type == NCSIM_STOCK_GEOM_INIT_GEN_SRF) {
	IPFreeObject(StockGeomInit.GenSrf.Geom);
	IritFree(StockGeomInit.GenSrf.FName);
    }

    if (AllInfo != NULL)
	delete AllInfo;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Pop up the all-parameters' setting dialog.                               M
*                                                                            *
* PARAMETERS:                                                                M
*   None                                                                     M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimulationClass::SetSimInfoParams                                      M
*****************************************************************************/
void NCSimulationClass::SetSimInfoParams()
{
    NCSIM_DEBUG_FUNC_ENTRY

    if (AllInfo == NULL) {
	const char *FName;
	char Line[IRIT_LINE_LEN_LONG];

	NCSIM_NAME_FROM_PATH(FName, NCFileName);
	sprintf(Line, "NC Simulation Parameters (%s)", FName);

	AllInfo = new NCSimulationParamClass(this, NCSimMF, wxID_ANY,
					     NCSimChars2wxString(Line));
    }

    AllInfo -> Show();
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Add a line to the provided sizer that contains Title and a real spin     *
* pointing to R.                                                             *
*                                                                            *
* PARAMETERS:                                                                *
*   TopSizer: Top level sizer to add this line to.                           *
*   Dialog:   parent widget.				                     *
*   Title:    The title to add.						     *
*   R:	      Reference to the real location to update.			     *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void AddRSpinWidget(wxBoxSizer *TopSizer,
			   wxDialog *Dialog,
			   char *Title,
			   IrtRType *R)
{
    NCSIM_DEBUG_FUNC_ENTRY

    wxBoxSizer
	*Sizer = new wxBoxSizer(wxHORIZONTAL);
    wxStaticText
	*Ln = new wxStaticText(Dialog, wxID_ANY, NCSimChars2wxString(Title));
    RealSpinCtrlClass
	*RSpin = new RealSpinCtrlClass(R, -IRIT_INFNTY, IRIT_INFNTY, 0.01, 0.1,
				       Dialog, wxID_ANY, wxDefaultPosition,
				       wxSize(45, 20));

    Sizer -> Add(Ln, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);
    Sizer -> Add(RSpin, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);
    TopSizer -> Add(Sizer, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Prompt the user to insert the special geometry stock parameters.         M
*                                                                            *
* PARAMETERS:                                                                M
*   Type:  Type of special stock geometry to fetch parameters for.           M
*                                                                            *
* RETURN VALUE:                                                              M
*   void				                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimulationClass::GetStockGeomInitParams                                M
*****************************************************************************/
void NCSimulationClass::GetStockGeomInitParams(NCSimStockGeomInitType Type)
{
    NCSIM_DEBUG_FUNC_ENTRY

    /* Free old data, if was any. */
    if (StockGeomInit.Type == NCSIM_STOCK_GEOM_INIT_GEN_SRF) {
        IPFreeObject(StockGeomInit.GenSrf.Geom);
	IritFree(StockGeomInit.GenSrf.FName);
    }

    assert(AllInfo != NULL);

    StockGeomInit.Type = Type;

    switch (Type) {
	case NCSIM_STOCK_GEOM_INIT_NONE:
	    break;
	case NCSIM_STOCK_GEOM_INIT_PLANE:
	case NCSIM_STOCK_GEOM_INIT_CYLIN_CAP:
	case NCSIM_STOCK_GEOM_INIT_CYLIN_CAVITY:
	case NCSIM_STOCK_GEOM_INIT_SPHERE_CAP:
	case NCSIM_STOCK_GEOM_INIT_SPHERE_CAVITY:
	    {
	        wxDialog
		    *Dialog = new wxDialog(AllInfo, wxID_ANY,
					   wxT("Get stock geometry params"),
					   wxDefaultPosition, wxDefaultSize,
					   wxCAPTION);
		wxBoxSizer *Sizer,
		    *TopSizer = new wxBoxSizer(wxVERTICAL);
		wxButton *Btn;

		Dialog -> Centre();
		Dialog -> SetSizer(TopSizer);

		switch (Type) {
		    case NCSIM_STOCK_GEOM_INIT_NONE:
		        break;
		    case NCSIM_STOCK_GEOM_INIT_PLANE:
		        AddRSpinWidget(TopSizer, Dialog, "Plane A Coef.:",
				       &StockGeomInit.Plane.ABCD[0]);
		        AddRSpinWidget(TopSizer, Dialog, "Plane B Coef.:",
				       &StockGeomInit.Plane.ABCD[1]);
		        AddRSpinWidget(TopSizer, Dialog, "Plane C Coef.:",
				       &StockGeomInit.Plane.ABCD[2]);
		        AddRSpinWidget(TopSizer, Dialog, "Plane D Coef.:",
				       &StockGeomInit.Plane.ABCD[3]);
		        break;
		    case NCSIM_STOCK_GEOM_INIT_CYLIN_CAP:
		    case NCSIM_STOCK_GEOM_INIT_CYLIN_CAVITY:
		        AddRSpinWidget(TopSizer, Dialog, "Cylinder Point 1, X:",
				       &StockGeomInit.Cylin.Pt1[0]);
		        AddRSpinWidget(TopSizer, Dialog, "Cylinder Point 1, Y:",
				       &StockGeomInit.Cylin.Pt1[1]);
		        AddRSpinWidget(TopSizer, Dialog, "Cylinder Point 1, Z:",
				       &StockGeomInit.Cylin.Pt1[2]);
		        AddRSpinWidget(TopSizer, Dialog, "Cylinder Point 2, X:",
				       &StockGeomInit.Cylin.Pt2[0]);
		        AddRSpinWidget(TopSizer, Dialog, "Cylinder Point 2, Y:",
				       &StockGeomInit.Cylin.Pt2[1]);
		        AddRSpinWidget(TopSizer, Dialog, "Cylinder Point 2, Z:",
				       &StockGeomInit.Cylin.Pt2[2]);
		        AddRSpinWidget(TopSizer, Dialog, "Cylinder Radius:",
				       &StockGeomInit.Cylin.Rad);
		        break;
		    case NCSIM_STOCK_GEOM_INIT_SPHERE_CAP:
		    case NCSIM_STOCK_GEOM_INIT_SPHERE_CAVITY:
		        AddRSpinWidget(TopSizer, Dialog, "Sphere Center, X:",
				       &StockGeomInit.Sphere.Cntr[0]);
		        AddRSpinWidget(TopSizer, Dialog, "Sphere Center, Y:",
				       &StockGeomInit.Sphere.Cntr[1]);
		        AddRSpinWidget(TopSizer, Dialog, "Sphere Center, Z:",
				       &StockGeomInit.Sphere.Cntr[2]);
		        AddRSpinWidget(TopSizer, Dialog, "Sphere Radius:",
				       &StockGeomInit.Sphere.Rad);
		        break;
		}

		Sizer = new wxBoxSizer(wxHORIZONTAL);
		Btn = new wxButton(Dialog, wxID_OK, wxT("OK"));
		Sizer -> Add(Btn, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

		TopSizer -> Add(Sizer, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

		/* Done. */
		TopSizer -> Fit(Dialog);
		TopSizer -> SetSizeHints(Dialog);

		/* Show it! */
		Dialog -> ShowModal();

		delete Dialog;
	    }
	    break;
	case NCSIM_STOCK_GEOM_INIT_GEN_SRF:
	    {
		const char
		    *HelpStr= "Irit text files (*.itd)|*.itd|"
			      "Irit binary files (*.ibd)|*.ibd|"
			      "IGES files (*.igs)|*.igs|"
			      "IGES files (*.iges)|*.iges|"
			      "STL files (*.stl)|*.stl|"
			      "All files (*.*)|*.*";
	        wxFileDialog 
		    *FileDlg = new
		        wxFileDialog(AllInfo,
				     wxT("Select initial stock geometry file"), 
				     wxT(""),           /* Default directory */
				     wxT(""),           /* Default file */
				     NCSimChars2wxString(HelpStr),
				     wxFD_OPEN | wxFD_FILE_MUST_EXIST | wxFD_CHANGE_DIR);
    		FileDlg -> Centre();

		if (wxID_OK == FileDlg -> ShowModal()) {
		    wxString
			Str = FileDlg -> GetPath();
		    const char
		        *FName = NCSimwxString2Chars(Str);

		    SetCursor(wxCURSOR_WAIT);

		    if ((StockGeomInit.GenSrf.Geom =
					 IGLoadGeometry(&FName, 1)) == NULL) {
			StockGeomInit.Type = NCSIM_STOCK_GEOM_INIT_NONE;
			AllInfo -> ShapedStockCBox ->
			   SetValue(NCSimChars2wxString(
			           GetStockGeomInitAsStr(StockGeomInit.Type)));
		    }
		    else
		        StockGeomInit.GenSrf.FName =
						    IritStrdup((char *) FName);
		    SetCursor(wxCURSOR_ARROW);
		}
		else {
		    StockGeomInit.Type = NCSIM_STOCK_GEOM_INIT_NONE;
		    AllInfo -> ShapedStockCBox ->
			SetValue(NCSimChars2wxString(
				   GetStockGeomInitAsStr(StockGeomInit.Type)));
		}

		delete FileDlg;
	    }
	    break;
	default:
	    assert(0);
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Build a map of all positions of all lines so we can track lines fast.    M
* Assumes that the text control holds all its text now.			     M
*                                                                            *
* PARAMETERS:                                                                M
*   None                                                                     M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimulationClass::UpdateLinePosMap                                      M
*****************************************************************************/
void NCSimulationClass::UpdateLinePosMap()
{
    NCSIM_DEBUG_FUNC_ENTRY

    int i;
 
    NCFileNumLines = GetNumberOfLines();
    NCFileLineMap = (int *) IritMalloc(NCFileNumLines * sizeof(int));
    NCFileLineMap[0] = 0;

#ifdef NCSIM_USE_WX_TEXTCTRL_LINE_LEN  /* Very slow! */
    /* Count one more than line length, for the '\n' char. */
    for (i = 1; i < NCFileNumLines; i++)
        NCFileLineMap[i] = NCFileLineMap[i - 1] + GetLineLength(i - 1) + 1;
#else
    int l;
    const wxChar *p;
    wxString
        Str = GetValue();

    for (p = Str.fn_str(), l = i = 0; *p != 0; p++, l++) {
        if (*p == '\n')
	    NCFileLineMap[++i] = l + 1;
    }        
#endif /* NCSIM_USE_WX_TEXTCTRL_LINE_LEN */

    /* For some reason wxtextCtrl returns wrong size. */
    assert(NCFileNumLines >= i);
    NCFileNumLines = i;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Make the requested line visible and highlight it.                        M
*                                                                            *
* PARAMETERS:                                                                M
*   LineNum:   To make visible and highlight.                                M
*   Highlight: TRUE to highlight, FALSE to unhighlight.                      M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimulationClass::HighlightTextLine                                     M
*****************************************************************************/
void NCSimulationClass::HighlightTextLine(int LineNum, int Highlight)
{
    NCSIM_DEBUG_FUNC_ENTRY

    LineNum--;			        /* Here we count from zero not one. */
    if (LineNum < 0 || LineNum >= GetNumberOfLines())
	return;

    SetStyle(NCFileLineMap[LineNum],
	     NCFileLineMap[LineNum] + GetLineLength(LineNum),
	     Highlight ? wxTextAttr(NCSIM_COLOR_HIGHLIGHT)
	               : wxTextAttr(NCSIM_COLOR_NORMAL));

    if (Highlight)
	ShowPosition(NCFileLineMap[IRIT_MAX(LineNum, 0)]);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Recreate the tool geometry from the tools parameters, assuming not a     M
* general tool, in which case this function does nothing.                    M
*   Tool is created at a canonical position (origin) and orientation (+Z),   M
* and is saved in place in the Geom slot of this object. "ToolBottom" and    M
* "ToolProfile" are saved as attributes on Geom for further processing.      M
*                                                                            *
* PARAMETERS:                                                                M
*   Idx:  Tool index to recreate.                                            M
*                                                                            *
* RETURN VALUE:                                                              M
*   bool:    TRUE if geometry was (re)created, false otherwise.              M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimulationClass::ReCreateToolGeometry                                  M
*****************************************************************************/
bool NCSimulationClass::ReCreateToolGeometry(int Idx)
{
    NCSIM_DEBUG_FUNC_ENTRY

    int OldVal;
    IPPolygonStruct *Pl;
    IPObjectStruct *PObj;
    CagdCrvStruct *ToolProfile;
    CagdSrfStruct *ToolSrf, *ToolBottom;
    IPNCGCToolType GMTType;

    assert(ToolIdx >= 0);
    if (Idx < 0)
	Idx = ToolIdx;

    IGGlblPlgnFineness = GeomApproxTolerance;

    switch (Tools[Idx].Type) {
	case NCSIM_TOOL_TYPE_NONE:
	    if (Tools[Idx].Geom != NULL)
	        IPFreeObjectList(Tools[Idx].Geom);
	    Tools[Idx].Geom = NULL;
	    if (Tools[Idx].Mask != NULL)
	        IritFree(Tools[Idx].Mask);
	    Tools[Idx].Mask = NULL;
	    Tools[Idx].MaskX = Tools[Idx].MaskY = 0;
	    return true;
	case NCSIM_TOOL_TYPE_BALL_END:
	    GMTType = IP_NC_GCODE_TOOL_BALL_END;
	    break;
	case NCSIM_TOOL_TYPE_TORUS_END:
	    GMTType = IP_NC_GCODE_TOOL_TORUS_END;
	    break;
	case NCSIM_TOOL_TYPE_FLAT_END:
	    GMTType = IP_NC_GCODE_TOOL_FLAT_END;
	    break;
	case NCSIM_TOOL_TYPE_OTHER:
	default:
	    return false;
    }

    if ((ToolSrf = IPNCGCodeGenToolGeom(GMTType, Tools[Idx].Diameter,
					NCSIM_TOOL_HEIGHT_DIAM *
					                 Tools[Idx].Diameter,
					Tools[Idx].TorusRadius, &ToolProfile,
					&ToolBottom)) != NULL) {

        if (Tools[Idx].Geom != NULL)
	    IPFreeObjectList(Tools[Idx].Geom);
	Tools[Idx].Geom = IPGenSrfObject("Tool", ToolSrf, NULL);

	AttrSetObjectObjAttrib(Tools[Idx].Geom, "ToolProfile",
			       IPGenCrvObject("ToolProfile", ToolProfile,
					      NULL),
			       FALSE);

	/* Convert the Tool Bottom to triangles. */
	OldVal = IPSurface2PolygonsGenTriOnly(TRUE);
	Pl = IPSurface2Polygons(ToolBottom, FALSE, GeomApproxTolerance,
				FALSE, FALSE, TRUE);
	IPSurface2PolygonsGenTriOnly(OldVal);

	PObj = IPGenPolyObject("ToolBottom", Pl, NULL);
	/* Prevent the rndr_lib from applying any transformations. */
	AttrSetObjectIntAttrib(PObj, "_TRANSFORMED", TRUE);
	AttrSetObjectObjAttrib(Tools[Idx].Geom, "ToolBottom", PObj, FALSE);

	ZBufferInitToolMask(&Tools[Idx]);  /* (Re)create tool ZBuffer mask. */
    }

    return true;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Activate this NC simulation page.					     M
*                                                                            *
* PARAMETERS:                                                                M
*   ForceAutoScale:  Force auto scale now, if TRUE.			     M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimulationClass::ActivateSimPage                                       M
*****************************************************************************/
void NCSimulationClass::ActivateSimPage(IrtBType ForceAutoScale)
{
    NCSIM_DEBUG_FUNC_ENTRY

    if (ForceAutoScale || MatIsUnitMatrix(NCSimViewMat, IRIT_EPS)) {
        AutoScaleView();
    }
    else {
	IRIT_HMGN_MAT_COPY(IPViewMat, NCSimViewMat);
    }

    NCSimMF -> AnimNCLineNum -> SetRange(1, NCFileNumLines);

    LastErrorGCodeLineNumber = -1;
    NCFileLineNum = -1;

    UpdateSimDisplayedStatus();
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Auto scale the current simulation based on the bbox of its gemetry.      M
*   Also allows selection of certain basic views.			     M
*                                                                            *
* PARAMETERS:                                                                M
*   View:  The view to set to.                                               M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimulationClass::AutoScaleView                                         M
*****************************************************************************/
void NCSimulationClass::AutoScaleView(NCSimViewDirectionType View)
{
    NCSIM_DEBUG_FUNC_ENTRY

    IrtRType s;
    IrtVecType Center, Scaling;
    IrtHmgnMatType Mat1, Mat2, Mat3;
    GMBBBboxStruct *BBox;

    if (GPathGeom != NULL)
	BBox = GMBBComputeBboxObjectList(GPathGeom);
    else if (PObjGnrl != NULL)
	BBox = GMBBComputeBboxObjectList(PObjGnrl);
    else
        return;

    IRIT_PT_ADD(Center, BBox -> Max, BBox -> Min);
    IRIT_PT_SCALE(Center, 0.5);
    MatGenMatTrans(-Center[0], -Center[1], -Center[2], Mat1);

    IRIT_PT_SUB(Scaling, BBox -> Max, BBox -> Min);
    s = IRIT_MAX(Scaling[0], IRIT_MAX(Scaling[1], Scaling[2]));
    if (s < IRIT_EPS)
	s = IRIT_EPS;
    MatGenMatUnifScale(1.0 / s, Mat2);

    switch (View) {
	case NCSIM_VIEW_CRNT:
	    MatRotateFactorMatrix(IPViewMat, Mat3);
	    MatMultTwo4by4(IPViewMat, Mat1, Mat2);
	    MatMultTwo4by4(IPViewMat, IPViewMat, Mat3);
	    break;
	case NCSIM_VIEW_FRONT:
	    MatMultTwo4by4(IPViewMat, Mat1, Mat2);
	    break;
	case NCSIM_VIEW_SIDE:
	    MatMultTwo4by4(IPViewMat, Mat1, Mat2);
	    MatGenMatRotY1(M_PI * -0.5, Mat1);
	    MatMultTwo4by4(IPViewMat, IPViewMat, Mat1);
	    break;
	case NCSIM_VIEW_UP:
	    MatMultTwo4by4(IPViewMat, Mat1, Mat2);
	    MatGenMatRotX1(M_PI * 0.5, Mat1);
	    MatMultTwo4by4(IPViewMat, IPViewMat, Mat1);
	    break;
    }

    IRIT_HMGN_MAT_COPY(NCSimViewMat, IPViewMat);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Load geometry into the current simulation.				     M
*                                                                            *
* PARAMETERS:                                                                M
*   FullPath:   Name of file to read.                                        M
*   ObjType:    Type of read object.			                     M
*                                                                            *
* RETURN VALUE:                                                              M
*   bool:   True if successful, false otherwise.                             M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimulationClass::LoadGeometry                                          M
*****************************************************************************/
bool NCSimulationClass::LoadGeometry(const char *FullPath,
				     NCSimObjectType ObjType)
{
    NCSIM_DEBUG_FUNC_ENTRY

    IPObjectStruct *PObjs;

    if ((PObjs = IGLoadGeometry(&FullPath, 1)) != NULL) {
        UpdateObjGeometry(PObjs, ObjType);
    }
    else {
        NCSimPrintf(NCSIM_LOG_ERROR, "Failed to read geometry file \"%s\".\n",
		    FullPath);
        return false;
    }

    return true;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Computes the bbox of the tool path and update the stock bbox using that. M
*                                                                            *
* PARAMETERS:                                                                M
*   ExpandAbsolute:  How much to expand beyond the actual BBox, in absolute  M
*		     coordinates.			                     M
*   ExpandRelative:  How much to expand beyond the actual BBox, in relative  M
*		     coordinates.			                     M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimulationClass::SetBBoxFromPathGeom		                     M
*****************************************************************************/
void NCSimulationClass::SetBBoxFromPathGeom(IrtRType ExpandAbsolute,
					    IrtRType ExpandRelative)
{
    NCSIM_DEBUG_FUNC_ENTRY

    int i;
    GMBBBboxStruct
	*BBox = IPNCGCodeBBox(GStream, TRUE);

    StockBBox = *BBox;

    for (i = 0; i < 3; i++) {
	IrtRType
	    d = (StockBBox.Max[i] - StockBBox.Min[i]) * ExpandRelative;

        StockBBox.Min[i] -= ExpandAbsolute + d;
	StockBBox.Max[i] += ExpandAbsolute + d;
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Converts the given value in mm to the currently used units.              M
*                                                                            *
* PARAMETERS:                                                                M
*   Val:    Value im mm to convert to current units.                         M
*                                                                            *
* RETURN VALUE:                                                              M
*   IrtRType:   Val in the current units.                                    M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimulationClass::GetValInMm                                            M
*****************************************************************************/
IrtRType NCSimulationClass::GetValInMm(IrtRType Val)
{
    NCSIM_DEBUG_FUNC_ENTRY

    switch (Units) {
	default:
	    assert(0);
	case NCSIM_UNIT_MM:
	    return Val;
	case NCSIM_UNIT_INCHES:
	    return NCSIM_MM2INCHES(Val);
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Sets the desired simulation mode - real time ot fixed step.              M
*                                                                            *
* PARAMETERS:                                                                M
*   SimMode:   New simulation mode to use.                                   M
*                                                                            *
* RETURN VALUE:                                                              M
*   NCSimModeType:  Old simulation mode.                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimulationClass::SetSimMode                                            M
*****************************************************************************/
NCSimModeType NCSimulationClass::SetSimMode(NCSimModeType SimMode)
{
    NCSIM_DEBUG_FUNC_ENTRY

    NCSimModeType
	OldSimMode = SimMode;

    NCSimMode = SimMode;

    if (NCSimStep == 0.0)
	NCSimStep = Tools[ToolIdx].Diameter / 32.0;

    return OldSimMode;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Update the simulation displayed information/status.                      M
*                                                                            *
* PARAMETERS:                                                                M
*   None			                                             M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* SEE ALSO:                                                                  M
*                                                                            M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimulationClass::UpdateSimDisplayedStatus                              M
*****************************************************************************/
void NCSimulationClass::UpdateSimDisplayedStatus()
{
    NCSIM_DEBUG_FUNC_ENTRY

    char Line[IRIT_LINE_LEN_LONG];

    /* Update progress status. */
    if (AnimArcLen >= 0) { /* Negative values designated end of simulation. */
        switch (NCSimMode) {
	    default:
	        assert(0);
	    case NCSIM_MODE_TIME:
	        sprintf(Line, "%.2f Secs (%.2f%%) X %.3g [%8.5f %8.5f %8.5f] (%s)",
			AnimRealTime, 100.0 * AnimArcLen / AnimTotalArcLen,
			AnimMultFactor, AnimToolPosition[0],
			AnimToolPosition[1], AnimToolPosition[2], 
			GetAnimStateAsStr(AnimCrntState));
		break;
	    case NCSIM_MODE_STEP:
	        sprintf(Line, "%.2f %s (%.2f%%) X %.3g [%8.5f %8.5f %8.5f] (%s)",
			AnimArcLen, GetUnitsAsStr(Units),
			100.0 * AnimArcLen / AnimTotalArcLen,
			AnimMultFactor, AnimToolPosition[0],
			AnimToolPosition[1], AnimToolPosition[2],
			GetAnimStateAsStr(AnimCrntState));
		break;
	}

	NCSimMF -> StatusBar -> SetStatusText(NCSimChars2wxString(Line), 1);
    }

    /* Update feedrate/spindle speed status. */
    switch (NCSimMode) {
	default:
	    assert(0);
	case NCSIM_MODE_TIME:
	    sprintf(Line, "T%d, FR = %g %s, SS = %g RPM, MRR = %.3g %s",
		    ToolIdx, FeedRate,
		    GetFeedRateUnitsAsStr(Units), SpindleSpeed,
		    MaterialRemovalRate,
		    GetMaterialRemovalRateUnitsAsStr(Units));
	    break;
	case NCSIM_MODE_STEP:
	    sprintf(Line, "T%d, Stp = %g %s, SS = %g RPM, MRR = %.3g %s",
		    ToolIdx, NCSimStep,
		    GetUnitsAsStr(Units), SpindleSpeed,
		    MaterialRemovalRate,
		    GetMaterialRemovalRateUnitsAsStr(Units));
	    break;
    }

    NCSimMF -> StatusBar -> SetStatusText(NCSimChars2wxString(Line), 2);

    /* Update the line number/nc text/slider animation bar. */
    if (AnimGCodeInfo != NULL) {
        NCSimMF -> AnimNCLineNum -> SetValue(NCFileLineNum);
	NCSimMF -> AnimNCText -> 
		        SetValue(NCSimChars2wxString(AnimGCodeInfo -> Line));
	NCSimMF -> AnimSlider ->
	    SetValue((int) (0.5 + NCFileLineNum * 10000.0 / NCFileNumLines));
    }
    else {
        NCSimMF -> AnimNCLineNum -> SetValue(0);
	NCSimMF -> AnimNCText -> SetValue(wxT(""));
	NCSimMF -> AnimSlider -> SetValue(0);
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Sets the current animation state to Percent.                             M
*                                                                            *
* PARAMETERS:                                                                M
*   Percent:	Of animation where we want to be at.                         M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimulationClass::SetAnimPercent                                        M
*****************************************************************************/
void NCSimulationClass::SetAnimPercent(IrtRType Percent)
{
    NCSIM_DEBUG_FUNC_ENTRY

    int TargetLineNum = (int) (0.5 + NCFileNumLines * Percent / 100.0);

    SetAnimLine(TargetLineNum);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Sets the current animation state to be at line number TargetLineNum.     M
*                                                                            *
* PARAMETERS:                                                                M
*   TargetLineNum:	Of animation where we want to be at.                 M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimulationClass::SetAnimLine                                           M
*****************************************************************************/
void NCSimulationClass::SetAnimLine(int TargetLineNum)
{
    NCSIM_DEBUG_FUNC_ENTRY

    int DLine;
    IrtRType Dt, Ds;

    if (AnimTotalArcLen <= 0.0 || GStream == NULL)
	return;

    /* Select reasonable incremental steps to use. */
    if (TargetLineNum == NCFileLineNum) {
        return;
    }
    else if (TargetLineNum > NCFileLineNum) {
        Dt = 0.01;
	Ds = Tools[ToolIdx].Diameter / 100.0;
    }
    else {
        Dt = -0.01;
	Ds = -Tools[ToolIdx].Diameter / 100.0;
    }
    DLine = IRIT_ABS(TargetLineNum - NCFileLineNum);

    while (TRUE) {
        switch (NCSimMode) {
	    default:
	        assert(0);
	    case NCSIM_MODE_TIME:
		AnimArcLen = IPNCGCodeTraverseTime(GStream, Dt,
						   &AnimRealTime,
						   AnimToolPosition,
						   &AnimGCodeInfo);
		break;
	    case NCSIM_MODE_STEP:
	        AnimArcLen = IPNCGCodeTraverseStep(GStream, Ds,
						   &AnimRealTime,
						   AnimToolPosition,
						   &AnimGCodeInfo);
		break;
	}

	if (NCFileLineNum != AnimGCodeInfo -> StreamLineNumber) {
	    NCFileLineNum = AnimGCodeInfo -> StreamLineNumber;
	    if (NCFileLineNum == TargetLineNum ||
		IRIT_ABS(NCFileLineNum - TargetLineNum) > DLine)
	        break;
	    DLine = IRIT_ABS(NCFileLineNum - TargetLineNum);
	}
	if (AnimArcLen < 0.0)
	    break;
    }  

    NCSimMF -> Refresh();		                /* Force a redraw. */
    UpdateSimDisplayedStatus();
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   A call back that will be invoked every time the G code changed a line.   *
*                                                                            *
* PARAMETERS:                                                                *
*   None                                                                     *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void TrackGCodeIndicesChangeCB(void)
{
    GCodeIndexChanged = TRUE;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Traps on idle events to be used in animation if animation in active.     M
*                                                                            *
* PARAMETERS:                                                                M
*   Event: N.S.F.I.                                                          M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimulationClass::OnIdle                                                M
*****************************************************************************/
void NCSimulationClass::OnIdle(wxIdleEvent &Event)
{
    NCSIM_DEBUG_FUNC_ENTRY

    int LnNum = -1;
    IrtRType Dt, Ds, AnimCrntTime;

    if  (GStream == NULL)
	return;

    if (SingleStepActive) {
	GCodeIndexChanged = FALSE;
	IPNCGCodeUpdateGCodeIndexCBFunc(GStream, TrackGCodeIndicesChangeCB);
    }
    else
	IPNCGCodeUpdateGCodeIndexCBFunc(GStream, NULL);

    switch (AnimCrntState) {
	default:
	    assert(0);				 /* Should never get here. */
	case NCSIM_ANIM_STATE_STOP:
	    break;
	case NCSIM_ANIM_STATE_PAUSE:
	    break;
	case NCSIM_ANIM_STATE_RUN:
	    switch (NCSimMode) {
	        default:
		    assert(0);
		case NCSIM_MODE_TIME:
		    AnimCrntTime = GetTickCount() * 0.001;
		    Dt = NCSimStep * AnimMultFactor;

		    AnimArcLen = IPNCGCodeTraverseTime(GStream, Dt,
						       &AnimRealTime,
						       AnimToolPosition,
						       &AnimGCodeInfo);
		    break;
		case NCSIM_MODE_STEP:
		    if (NCSimStep == 0.0) {
			NCSimPrintf(NCSIM_LOG_ERROR, "Simulation Step size was zero - adjusted to 1/2 tool radius.\n");
			NCSimStep = Tools[ToolIdx].Diameter / 4.0;
		    }

		    Ds = NCSimStep * AnimMultFactor;
		    AnimCrntTime = GetTickCount() * 0.001;

		    AnimArcLen = IPNCGCodeTraverseStep(GStream, Ds,
						       &AnimRealTime,
						       AnimToolPosition,
						       &AnimGCodeInfo);
		    break;
	    }

#	    ifdef DEBUG_PRINT_SIM_TRACE
	        NCSimPrintf(NCSIM_LOG_ERROR,
			    "Ds/Dt = %f, AnimArcLen = %f, Pos = [%f %f %f]\nLine %d (%d), Len - %f (%f), Ver = %d \"%s\".\n",
			    NCSimMode == NCSIM_MODE_TIME ? Dt: Ds,
			    AnimArcLen,
			    AnimToolPosition[0],
			    AnimToolPosition[1],
			    AnimToolPosition[2],
			    AnimGCodeInfo -> StreamLineNumber,
			    AnimGCodeInfo -> GCodeLineNumber,
			    AnimGCodeInfo -> LenStart,
			    AnimGCodeInfo -> Len,
			    AnimGCodeInfo -> IsVerticalUpMotion,
			    AnimGCodeInfo -> Line);
#	    endif /* DEBUG_PRINT_SIM_TRACE */

	    /* Update state: line number, spindle speed, feedrate, etc. */
	    LnNum = AnimGCodeInfo -> StreamLineNumber;
	    if (FeedRate != IRIT_ABS(AnimGCodeInfo -> UpdatedFeedRate)) {
	        FeedRate = IRIT_ABS(AnimGCodeInfo -> UpdatedFeedRate);
		if (AllInfo != NULL)
		    AllInfo -> FeedRateRSpin -> Refresh();
	    }
	    if (SpindleSpeed != IRIT_ABS(AnimGCodeInfo -> SpindleSpeed)) {
	        SpindleSpeed = IRIT_ABS(AnimGCodeInfo -> SpindleSpeed);
		if (AllInfo != NULL)
		    AllInfo -> SpindleSpeedRSpin -> Refresh();
	    }
	    if (ToolIdx != AnimGCodeInfo -> ToolNumber) {
	        ToolIdx = AnimGCodeInfo -> ToolNumber;
		if (AllInfo != NULL)
		    AllInfo -> ToolIndexSpin -> SetValue(ToolIdx);
		if (Tools[ToolIdx].Type == NCSIM_TOOL_TYPE_NONE)
		    NCSimPrintf(NCSIM_LOG_ERROR,
			       "Selected tool index %d is not defined.\n",
			       ToolIdx);

		else
		    ReCreateToolGeometry(ToolIdx);
	    }

	    if (LnNum != NCFileLineNum) {
	        HighlightTextLine(NCFileLineNum, FALSE);
	        NCFileLineNum = LnNum;
		HighlightTextLine(NCFileLineNum);
	    }

	    if (NCZBufferOn) {
	        /* Render the tool in current position into the ZBuffer. */
	        ZBufferRenderNCTool(&Tools[ToolIdx], AnimToolPosition);
	    }

	    if (AnimArcLen < 0.0) {
	        AnimCrntState = NCSIM_ANIM_STATE_STOP;   /* End of anim. */
		UpdateSimDisplayedStatus();

		/* Make sure we are all normal again. */
		SetStyle(0, GetLastPosition(), wxTextAttr(NCSIM_COLOR_INFO));

		NCSimMF -> Refresh();	                 /* Force a redraw. */
	    }
	    else {
		UpdateSimDisplayedStatus();

		/* skip this rendering (rendering every n'th iteration)? */
		if (!NCSimMF -> OGLWin -> SkipFrame()) {
		    NCSimMF -> Refresh();                /* Force a redraw. */
		}

		ZBufferUpdateCellsToRedraw();

		Event.RequestMore();
	    }
	    break;
    }

    /* We we single step - pause if we changed G code line: */
    if (GCodeIndexChanged) {
        GCodeIndexChanged = FALSE;
        AnimCrntState = NCSIM_ANIM_STATE_PAUSE;
	UpdateSimDisplayedStatus();
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Update the state of the current animation.                               M
*                                                                            *
* PARAMETERS:                                                                M
*   AnimState: New animation state to set to.                                M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimulationClass::AnimState                                             M
*****************************************************************************/
void NCSimulationClass::AnimState(NCSimAnimStateType AnimState)
{
    NCSIM_DEBUG_FUNC_ENTRY

    IrtRType YesNoTime;

    if (AnimCrntState == AnimState)
        return;

    if (GStream == NULL) {
        NCSimMF -> PopupMsg("No NC file to execute simulation for",
			    NCSIM_LOG_WARN);
	return;
    }
    switch (AnimState) {
        case NCSIM_ANIM_STATE_RUN:
	    if (AnimCrntState == NCSIM_ANIM_STATE_STOP) {
	        /* Initiate a whole new animation. */
		if (!HasToolSet()) {
		    NCSimPrintf(NCSIM_LOG_ERROR,
				"No tool was set - at least one tool must be set first.\n");
		    break;
		}

		if (Tools[ToolIdx].Type == NCSIM_TOOL_TYPE_NONE) {
		    NCSimPrintf(NCSIM_LOG_ERROR,
			       "Current tool index %d is not defined.\n",
			       ToolIdx);
		}

		NCSimMF -> OGLWin -> NCSimOGLInitSim();
	        AnimTotalArcLen = IPNCGCodeTraverseInit(GStream, 0.0,
						   AnimFastFeedRateMultFactor,
						   NCSIM_ARC_LEN_TRIGGER);
		if (NCZBufferOn)
		    ZBufferInit(NCSIM_MAX_ZBUF_DISP_LIST, Tools);
		LastErrorGCodeLineNumber = -1;
		NCFileLineNum = -1;
		NCSimPrintf(NCSIM_LOG_INFO,
			    "Starting Simulation in %s mode.\n",
			    NCSimMode == NCSIM_MODE_TIME ? "real time"
							 : "fixed step");

		if (!NCZBufferOn) {
		    NCSimPrintf(NCSIM_LOG_INFO,
				"NC Z buffer is off - only animating tool's motions.\n");
		}
	    }
	    AnimCrntState = NCSIM_ANIM_STATE_RUN;
	    break;
	case NCSIM_ANIM_STATE_STOP:
	    YesNoTime = GetTickCount();
	    if (!NCSimMF -> 
		  PopupYesNo("Are you sure you want to stop the animation?")) {
	        /* Compensate for the break. */
	        YesNoTime = (GetTickCount() - YesNoTime) * 0.001;
	        break;
	    }

	    AnimCrntState = NCSIM_ANIM_STATE_STOP;
	    UpdateSimDisplayedStatus();
	    break;
	case NCSIM_ANIM_STATE_PAUSE:
	    AnimCrntState = NCSIM_ANIM_STATE_PAUSE;
	    UpdateSimDisplayedStatus();
	    break;
	default:
	    assert(0);				 /* Should never get here. */
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Binds the master setup to a dynammically allocated structure that links  M
* the local class's variables.						     M
*                                                                            *
* PARAMETERS:                                                                M
*   ToolIdleColorStr: ToolIdleColor string to bind to.                       M
*   ToolCutColorStr:  ToolCutColor string to bind to.                        M
*   ToolFastColorStr: ToolFastColor string to bind to.                       M
*   StockColorStr:    StockColor string to bind to.                          M
*   StockBBoxMinStr:  StockBBoxMin string to bind to.                        M
*   StockBBoxMaxStr:  StockBBoxMax string to bind to.                        M
*   StockGeomInitStr: Initial special stiock shape such as cylinder, etc.    M 
*   SimModeStr:       SimMode string to bind to.                             M
*   UnitsStr:         Units string to bind to.                               M
*   ToolsStrs:        Tools vector of strings to bind to.                    M
*                                                                            *
* RETURN VALUE:                                                              M
*   IritConfigStruct *: A Dynamically allocated and binded Setup config.     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimulationClass::BindSimConfig                                         M
*****************************************************************************/
IritConfigStruct *NCSimulationClass::BindSimConfig(char **ToolIdleColorStr,
						   char **ToolCutColorStr,
						   char **ToolFastColorStr,
						   char **StockColorStr,
						   char **StockBBoxMinStr,
						   char **StockBBoxMaxStr,
						   char **StockGeomInitStr, 
						   char **SimModeStr,
						   char **UnitsStr,
						   char ***ToolsStrs)
{
    NCSIM_DEBUG_FUNC_ENTRY

    int l = 0;
    IritConfigStruct
	*SetUp = (IritConfigStruct *) IritMalloc(sizeof(IritConfigStruct) *
							NCSIM_CFG_NUM_SET_UP +
						 IRIT_LINE_LEN_LONG * 11);
    char *StrPtrs = (char *) &SetUp[NCSIM_CFG_NUM_SET_UP]; 

    /* Copy master setup so we can set its slots. */
    IRIT_GEN_COPY(SetUp, NCSimCfgSetUpMaster, sizeof(IritConfigStruct) *
							NCSIM_CFG_NUM_SET_UP);
    *ToolIdleColorStr = StrPtrs;
    **ToolIdleColorStr = 0;
    StrPtrs += IRIT_LINE_LEN_LONG;

    *ToolCutColorStr = StrPtrs;
    **ToolCutColorStr = 0;
    StrPtrs += IRIT_LINE_LEN_LONG;

    *ToolFastColorStr = StrPtrs;
    **ToolFastColorStr = 0;
    StrPtrs += IRIT_LINE_LEN_LONG;

    *StockColorStr = StrPtrs;
    **StockColorStr = 0;
    StrPtrs += IRIT_LINE_LEN_LONG;

    *StockBBoxMinStr = StrPtrs;
    **StockBBoxMinStr = 0;
    StrPtrs += IRIT_LINE_LEN_LONG;

    *StockBBoxMaxStr = StrPtrs;
    **StockBBoxMaxStr = 0;
    StrPtrs += IRIT_LINE_LEN_LONG;

    *StockGeomInitStr = StrPtrs;
    **StockGeomInitStr = 0;
    StrPtrs += IRIT_LINE_LEN_LONG;

    *SimModeStr = StrPtrs;
    **SimModeStr = 0;
    StrPtrs += IRIT_LINE_LEN_LONG;

    *UnitsStr = StrPtrs;
    **UnitsStr = 0;
    StrPtrs += IRIT_LINE_LEN_LONG;

    *ToolsStrs = (char **) StrPtrs;
    **ToolsStrs = NULL;

    SetUp[l++].VarData = &HasRelArcCenter;	/* RelArcCenters */
    SetUp[l++].VarData = &ReverseZDir;    	/* ReverseZDir */
    SetUp[l++].VarData = &NCZBufferOn;		/* NCZBuffer */
    SetUp[l++].VarData = &PauseOnError;		/* PauseOnError */
    SetUp[l++].VarData = &UpdateFeedRates;	/* UpdateFeedRates */
    SetUp[l++].VarData = &ToolIdx;		/* ToolIndex */
    SetUp[l++].VarData = &ZBufSize;		/* NCZBufferRes */
    SetUp[l++].VarData = &AnimFastFeedRateMultFactor; /* ToolFastFactor */
    SetUp[l++].VarData = &MaxMaterialRemovalRate; /* MaxRemoveRate */
    SetUp[l++].VarData = &GeomApproxTolerance;	/* GeomApprox */
    SetUp[l++].VarData = &FeedRate;		/* FeedRate */
    SetUp[l++].VarData = &SpindleSpeed;		/* SpindleSpeed */
    SetUp[l++].VarData = &NCSimStep;		/* SimFixedStep */
    SetUp[l++].VarData = &StockTransp;		/* StockTransp */
    SetUp[l++].VarData = ToolIdleColorStr;	/* ToolIdleColor */
    SetUp[l++].VarData = ToolFastColorStr;	/* ToolFastColor */
    SetUp[l++].VarData = ToolCutColorStr;	/* ToolCutColor */
    SetUp[l++].VarData = StockColorStr;		/* StockColor */
    SetUp[l++].VarData = StockBBoxMinStr;	/* StockMin */
    SetUp[l++].VarData = StockBBoxMaxStr;	/* StockMax */
    SetUp[l++].VarData = StockGeomInitStr;	/* StockGeomInit */
    SetUp[l++].VarData = SimModeStr;		/* AnimMode */
    SetUp[l++].VarData = UnitsStr;		/* Units */
    SetUp[l++].VarData = ToolsStrs;		/* Tools */
    assert(NCSIM_CFG_NUM_SET_UP == l);

    return SetUp;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Loads the current simulation configuration file - a file with same name  M
* as the NC Gcode file but with extension '.cfg'.			     M
*                                                                            *
* PARAMETERS:                                                                M
*   None	                                                             M
*                                                                            *
* RETURN VALUE:                                                              M
*   int:  TRUE if successful, FALSE otherwise.                               M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimulationClass::LoadSimConfig                                         M
*****************************************************************************/
int NCSimulationClass::LoadSimConfig()
{
    NCSIM_DEBUG_FUNC_ENTRY

    char *ToolIdleColorStr, *ToolCutColorStr, *ToolFastColorStr,
         *StockColorStr, *StockBBoxMinStr, *StockBBoxMaxStr, *StockGeomInitStr,
	 *SimModeStr, **ToolsStrs, *UnitsStr, FName[IRIT_LINE_LEN_LONG], *p;
    int i, V[3];
    double D[3];
    FILE *f;
    IritConfigStruct *SetUp;
    NCSimStockGeomInitType StockGeomType;

    strncpy(FName, NCFileName, IRIT_LINE_LEN_LONG - 4);
    if ((p = strrchr(FName, '.')) != NULL)
        *p = 0;
    else
        p = &FName[strlen(FName)];
    strcpy(p, "_ncsim.cfg");

    if ((f = fopen(FName, "r")) != NULL)
        fclose(f);
    else
	return FALSE;

    SetUp = BindSimConfig(&ToolIdleColorStr, &ToolCutColorStr,
			  &ToolFastColorStr, &StockColorStr,
			  &StockBBoxMinStr, &StockBBoxMaxStr,
			  &StockGeomInitStr, &SimModeStr, &UnitsStr,
			  &ToolsStrs);

    IritConfig(FName, SetUp, NCSIM_CFG_NUM_SET_UP);

    /* Hard-bind the data of the set up table to this simulation variables. */
    if (sscanf(ToolIdleColorStr, "%d %d %d", &V[0], &V[1], &V[2]) == 3) {
	ToolIdleColor[0] = V[0];
	ToolIdleColor[1] = V[1];
	ToolIdleColor[2] = V[2];
    }
    if (sscanf(ToolCutColorStr, "%d %d %d", &V[0], &V[1], &V[2]) == 3) {
	ToolCutColor[0] = V[0];
	ToolCutColor[1] = V[1];
	ToolCutColor[2] = V[2];
    }
    if (sscanf(ToolFastColorStr, "%d %d %d", &V[0], &V[1], &V[2]) == 3) {
	ToolFastColor[0] = V[0];
	ToolFastColor[1] = V[1];
	ToolFastColor[2] = V[2];
    }
    if (sscanf(StockColorStr, "%d %d %d", &V[0], &V[1], &V[2]) == 3) {
	StockColor[0] = V[0];
	StockColor[1] = V[1];
	StockColor[2] = V[2];
    }

    if (sscanf(StockBBoxMinStr, "%lf %lf %lf", &D[0], &D[1], &D[2]) == 3) {
	StockBBox.Min[0] = D[0];
	StockBBox.Min[1] = D[1];
	StockBBox.Min[2] = D[2];
    }
    if (sscanf(StockBBoxMaxStr, "%lf %lf %lf", &D[0], &D[1], &D[2]) == 3) {
	StockBBox.Max[0] = D[0];
	StockBBox.Max[1] = D[1];
	StockBBox.Max[2] = D[2];
    }

    if (sscanf(StockGeomInitStr, "%d", &StockGeomType) == 1) {
	switch (StockGeomType) {
	    default:
	    case NCSIM_STOCK_GEOM_INIT_NONE:
	        StockGeomInit.Type = NCSIM_STOCK_GEOM_INIT_NONE;
	        break;
	    case NCSIM_STOCK_GEOM_INIT_PLANE:
	        if (sscanf(&StockGeomInitStr[2], "%lf %lf %lf %lf",
			   &StockGeomInit.Plane.ABCD[0],
			   &StockGeomInit.Plane.ABCD[1],
			   &StockGeomInit.Plane.ABCD[2],
			   &StockGeomInit.Plane.ABCD[3]) == 4)
		    StockGeomInit.Type = StockGeomType;
		else
		    StockGeomInit.Type = NCSIM_STOCK_GEOM_INIT_NONE;
	        break;
	    case NCSIM_STOCK_GEOM_INIT_CYLIN_CAP:
	    case NCSIM_STOCK_GEOM_INIT_CYLIN_CAVITY:
	        if (sscanf(&StockGeomInitStr[2], "%lf %lf %lf %lf %lf %lf %lf",
			   &StockGeomInit.Cylin.Pt1[0],
			   &StockGeomInit.Cylin.Pt1[1],
			   &StockGeomInit.Cylin.Pt1[2],
			   &StockGeomInit.Cylin.Pt2[0],
			   &StockGeomInit.Cylin.Pt2[1],
			   &StockGeomInit.Cylin.Pt2[2],
			   &StockGeomInit.Cylin.Rad) == 7)
		    StockGeomInit.Type = StockGeomType;
		else
		    StockGeomInit.Type = NCSIM_STOCK_GEOM_INIT_NONE;
	        break;
	    case NCSIM_STOCK_GEOM_INIT_SPHERE_CAP:
	    case NCSIM_STOCK_GEOM_INIT_SPHERE_CAVITY:
	        if (sscanf(&StockGeomInitStr[2], "%lf %lf %lf %lf",
			   &StockGeomInit.Sphere.Cntr[0],
			   &StockGeomInit.Sphere.Cntr[1],
			   &StockGeomInit.Sphere.Cntr[2],
			   &StockGeomInit.Sphere.Rad) == 4)
		    StockGeomInit.Type = StockGeomType;
		else
		    StockGeomInit.Type = NCSIM_STOCK_GEOM_INIT_NONE;
	        break;
	    case NCSIM_STOCK_GEOM_INIT_GEN_SRF:
	        p = FName;
	        if (sscanf(&StockGeomInitStr[2], "%s", FName) == 1) {
		    if ((StockGeomInit.GenSrf.Geom =
			     IGLoadGeometry((const char **) &p, 1)) == NULL) {
		        NCSimPrintf(NCSIM_LOG_ERROR,
				"Failed to read stock geometry file \"%s\".\n",
				FName);
			StockGeomInit.Type = NCSIM_STOCK_GEOM_INIT_NONE;
		    }
		    else {
		        StockGeomInit.GenSrf.FName = IritStrdup(FName);
			StockGeomInit.Type = StockGeomType;
		    }
		}
		else
		    StockGeomInit.Type = NCSIM_STOCK_GEOM_INIT_NONE;
	        break;
	}
    }

    for (i = 0; i < NCSIM_MAX_NUM_OF_TOOLS && ToolsStrs[i] != NULL; i++) {
        char TType[IRIT_LINE_LEN];
	int TIndex;
	double TDiameter, TTorusRadius;

	if (sscanf(ToolsStrs[i], "%d %s %lf %lf",
		   &TIndex, TType, &TDiameter, &TTorusRadius) == 4) {
	    if (strnicmp(TType, "Ball", 4) == 0)
	        Tools[TIndex].Type = NCSIM_TOOL_TYPE_BALL_END;
	    else if (strnicmp(TType, "Torus", 5) == 0)
	        Tools[TIndex].Type = NCSIM_TOOL_TYPE_TORUS_END;
	    else if (strnicmp(TType, "Flat", 4) == 0)
	        Tools[TIndex].Type = NCSIM_TOOL_TYPE_FLAT_END;
	    else
	        continue;

	    Tools[TIndex].Diameter = TDiameter;
	    Tools[TIndex].TorusRadius = TTorusRadius;
	    Tools[TIndex].Geom = NULL;
	    Tools[TIndex].Mask = NULL;
	    Tools[TIndex].MaskX = Tools[TIndex].MaskY = 0;
	    ReCreateToolGeometry(TIndex);
	}
    }

    if (strnicmp(UnitsStr, "MM", 2) == 0)
	Units = NCSIM_UNIT_MM;
    else if (strnicmp(UnitsStr, "In", 2) == 0)
	Units = NCSIM_UNIT_INCHES;

    if (strnicmp(SimModeStr, "SimTime", 7) == 0)
	NCSimMode = NCSIM_MODE_TIME;
    else if (strnicmp(SimModeStr, "SimStep", 7) == 0)
	NCSimMode = NCSIM_MODE_STEP;

    IritFree(SetUp);

    return TRUE;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Saves the current simulation in a file with same name as the NC Gcode    M
* file but with extension '.cfg'.					     M
*                                                                            *
* PARAMETERS:                                                                M
*   None	                                                             M
*                                                                            *
* RETURN VALUE:                                                              M
*   int:  TRUE if successful, FALSE otherwise.                               M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimulationClass::SaveSimConfig                                         M
*****************************************************************************/
int NCSimulationClass::SaveSimConfig()
{
    NCSIM_DEBUG_FUNC_ENTRY

    char *ToolIdleColorStr, *ToolCutColorStr, *ToolFastColorStr,
	 *StockColorStr, *StockBBoxMinStr, *StockBBoxMaxStr, *StockGeomInitStr,
	 *SimModeStr, **ToolsStrs, *UnitsStr, FName[IRIT_LINE_LEN_LONG], *p;
    int i, j, RetVal;
    IritConfigStruct *SetUp;

    if (!OverwriteDefaults)      /* If no change in defaults - do not save. */
	return TRUE;
    OverwriteDefaults = FALSE;

    SetUp = BindSimConfig(&ToolIdleColorStr, &ToolCutColorStr,
			  &ToolFastColorStr, &StockColorStr,
			  &StockBBoxMinStr, &StockBBoxMaxStr,
			  &StockGeomInitStr, &SimModeStr, &UnitsStr,
			  &ToolsStrs);

    /* Hard-bind the data of the set up table to this simulation variables. */
    sprintf(ToolIdleColorStr, "%d %d %d",
	    ToolIdleColor[0], ToolIdleColor[1], ToolIdleColor[2]);
    sprintf(ToolFastColorStr, "%d %d %d",
	    ToolFastColor[0], ToolFastColor[1], ToolFastColor[2]);
    sprintf(ToolCutColorStr, "%d %d %d",
	    ToolCutColor[0], ToolCutColor[1], ToolCutColor[2]);
    sprintf(StockColorStr, "%d %d %d",
	    StockColor[0], StockColor[1], StockColor[2]);

    sprintf(StockBBoxMinStr, "%f %f %f",
	    StockBBox.Min[0], StockBBox.Min[1], StockBBox.Min[2]);
    sprintf(StockBBoxMaxStr, "%f %f %f",
	    StockBBox.Max[0], StockBBox.Max[1], StockBBox.Max[2]);


    sprintf(StockGeomInitStr, "%d", StockGeomInit.Type);
    switch (StockGeomInit.Type) {
        default:
        case NCSIM_STOCK_GEOM_INIT_NONE:
	    break;
        case NCSIM_STOCK_GEOM_INIT_PLANE:
	    sprintf(&StockGeomInitStr[1], " %lf %lf %lf %lf",
		    StockGeomInit.Plane.ABCD[0],
		    StockGeomInit.Plane.ABCD[1],
		    StockGeomInit.Plane.ABCD[2],
		    StockGeomInit.Plane.ABCD[3]);
	    break;
        case NCSIM_STOCK_GEOM_INIT_CYLIN_CAP:
        case NCSIM_STOCK_GEOM_INIT_CYLIN_CAVITY:
	    sprintf(&StockGeomInitStr[1], " %lf %lf %lf %lf %lf %lf %lf",
		    StockGeomInit.Cylin.Pt1[0],
		    StockGeomInit.Cylin.Pt1[1],
		    StockGeomInit.Cylin.Pt1[2],
		    StockGeomInit.Cylin.Pt2[0],
		    StockGeomInit.Cylin.Pt2[1],
		    StockGeomInit.Cylin.Pt2[2],
		    StockGeomInit.Cylin.Rad);
	    break;
        case NCSIM_STOCK_GEOM_INIT_SPHERE_CAP:
        case NCSIM_STOCK_GEOM_INIT_SPHERE_CAVITY:
	    sprintf(&StockGeomInitStr[1], " %lf %lf %lf %lf",
		    StockGeomInit.Sphere.Cntr[0],
		    StockGeomInit.Sphere.Cntr[1],
		    StockGeomInit.Sphere.Cntr[2],
		    StockGeomInit.Sphere.Rad);
	    break;
	case NCSIM_STOCK_GEOM_INIT_GEN_SRF:
	    sprintf(&StockGeomInitStr[1], " %s",
		    StockGeomInit.GenSrf.FName);
	    break;
    }

    for (i = j = 0;
	 i < NCSIM_MAX_NUM_OF_TOOLS && Tools[i].Type != NCSIM_TOOL_TYPE_NONE;
	 i++) {
        char TLine[IRIT_LINE_LEN_LONG];

	sprintf(TLine, "%d %s %lf %lf",	i, GetToolTypeAsStr(Tools[i].Type),
		Tools[i].Diameter, Tools[i].TorusRadius);
	ToolsStrs[j++] = IritStrdup(TLine);
    }
    ToolsStrs[j] = NULL;

    strncpy(UnitsStr, GetUnitsAsStr(Units), IRIT_LINE_LEN_LONG);
    strncpy(SimModeStr,
	    NCSimMode == NCSIM_MODE_TIME ? "SimTime" : "SimStep",
	    IRIT_LINE_LEN_LONG);

    strncpy(FName, NCFileName, IRIT_LINE_LEN_LONG - 11);
    if ((p = strrchr(FName, '.')) != NULL)
        *p = 0;
    else
        p = &FName[strlen(FName)];
    strcpy(p, "_ncsim.cfg");

    RetVal = IritConfigSave(FName, SetUp, NCSIM_CFG_NUM_SET_UP);

    for (j = 0; ToolsStrs[j] != NULL; j++)
        IritFree(ToolsStrs[j]);
    IritFree(SetUp);

    return RetVal;
}

/*****************************************************************************
* AUXILIARY:								     *
*    Event table for NCSimLoadNCInfoClass				     *
*****************************************************************************/
BEGIN_EVENT_TABLE(NCSimLoadNCInfoClass, wxDialog)
    EVT_CHECKBOX(wxID_NCSIM_PARAM_REL_ARC_CNTR, OnCheckBox)
    EVT_CHECKBOX(wxID_NCSIM_PARAM_REVERSE_Z_DIR, OnCheckBox)
    EVT_TEXT(wxID_NCSIM_PARAM_UNITS, OnComboBox)
    EVT_BUTTON(wxID_NCSIM_PARAM_OK, OnButton)
END_EVENT_TABLE()

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Constructor for the interactive setup control of NC loaded file.         M
*                                                                            *
* PARAMETERS:                                                                M
*   NCSim:       Simulation we should create this page for.		     M
*   Parent:      Frame window.					             M
*   Id:          ID assigned to window. 		                     M
*   Title:       Window's title.		       		             M
*   Flags:       Window styles' flags.				             M
*                                                                            *
* RETURN VALUE:                                                              M
*   NoValue                                                                  M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimLoadNCInfoClass::NCSimLoadNCInfoClass                               M
*****************************************************************************/
NCSimLoadNCInfoClass::NCSimLoadNCInfoClass(NCSimulationClass *NCSim,
					   wxWindow *Parent,
					   wxWindowID Id,
					   const wxString &Title,
					   int Flags):
	wxDialog(Parent,
		 Id,
		 Title,
		 wxDefaultPosition,
		 wxDefaultSize,
		 Flags | wxCAPTION)
{
    NCSIM_DEBUG_FUNC_ENTRY

    wxBoxSizer *Sizer,
	*TopSizer = new wxBoxSizer(wxVERTICAL);
    wxButton *Btn;
    wxCheckBox *CBtn;
    wxStaticText *Ln;
    wxComboBox *CBox;
    wxArrayString Strs1, Strs2;

    ThisNCSim = NCSim;    /* Keep a ref. to NCSimulationClass we belong to. */

    SetSizer(TopSizer);

    /* Row 1. */
    Sizer = new wxBoxSizer(wxHORIZONTAL);

    CBtn = new wxCheckBox(this, wxID_NCSIM_PARAM_REL_ARC_CNTR,
			  wxT("Relative Arc Centers"));
    CBtn -> Set3StateValue(ThisNCSim -> HasRelArcCenter ? wxCHK_CHECKED
						        : wxCHK_UNCHECKED);
    CBtn -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_REL_ARC_CNTR));

    Sizer -> Add(CBtn, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    Ln = new wxStaticText(this, wxID_ANY, wxT("Units:"));
    Ln -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_UNITS));
    Sizer -> Add(Ln, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    Strs2.Add(NCSimChars2wxString(NCSim -> GetUnitsAsStr(NCSIM_UNIT_MM)));
    Strs2.Add(NCSimChars2wxString(NCSim -> GetUnitsAsStr(NCSIM_UNIT_INCHES)));
    CBox = new wxComboBox(this, wxID_NCSIM_PARAM_UNITS,
			  NCSimChars2wxString(
				      NCSim -> GetUnitsAsStr(NCSim -> Units)),
			  wxDefaultPosition, wxDefaultSize,
			  Strs2, wxCB_DROPDOWN | wxCB_READONLY);
    CBox -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_UNITS));

    Sizer -> Add(CBox, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    TopSizer -> Add(Sizer, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    /* Row 2. */
    Sizer = new wxBoxSizer(wxHORIZONTAL);

    CBtn = new wxCheckBox(this, wxID_NCSIM_PARAM_REVERSE_Z_DIR,
			  wxT("Reverse Z direction"));
    CBtn -> Set3StateValue(ThisNCSim -> ReverseZDir ? wxCHK_CHECKED
						    : wxCHK_UNCHECKED);
    CBtn -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_REVERSE_Z_DIR));

    Sizer -> Add(CBtn, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    TopSizer -> Add(Sizer, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    /* Last row. */
    Sizer = new wxBoxSizer(wxHORIZONTAL);

    Btn = new wxButton(this, wxID_NCSIM_PARAM_OK, wxT("OK"));
    Sizer -> Add(Btn, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    TopSizer -> Add(Sizer, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    TopSizer -> Fit(this);
    TopSizer -> SetSizeHints(this);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Traps events from check box.                                             M
*                                                                            *
* PARAMETERS:                                                                M
*   Event: N.S.F.I.                                                          M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimLoadNCInfoClass::OnCheckBox                                         M
*****************************************************************************/
void NCSimLoadNCInfoClass::OnCheckBox(wxCommandEvent &Event)
{
    NCSIM_DEBUG_FUNC_ENTRY

    if (Event.GetId() == wxID_NCSIM_PARAM_REL_ARC_CNTR) {
	ThisNCSim -> HasRelArcCenter = Event.IsChecked();
	ThisNCSim -> OverwriteDefaults = TRUE;
    }
    else if (Event.GetId() == wxID_NCSIM_PARAM_REVERSE_Z_DIR) {
	ThisNCSim -> ReverseZDir = Event.IsChecked();
	ThisNCSim -> OverwriteDefaults = TRUE;
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Traps events from a combo box.	                                     M
*                                                                            *
* PARAMETERS:                                                                M
*   Event: N.S.F.I.                                                          M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimLoadNCInfoClass::OnComboBox                                         M
*****************************************************************************/
void NCSimLoadNCInfoClass::OnComboBox(wxCommandEvent &Event)
{
    NCSIM_DEBUG_FUNC_ENTRY

    if (Event.GetId() == wxID_NCSIM_PARAM_UNITS) {
        wxComboBox
	    *Win = (wxComboBox *) Event.GetEventObject();
	wxString
	    Data = Win -> GetValue();
	int i;

	switch (Event.GetId()) {
	    case wxID_NCSIM_PARAM_UNITS:
	        for (i = NCSIM_UNIT_FIRST;
		     i < NCSIM_UNIT_LAST;
		     i++) {
		    if (strcmp(ThisNCSim -> GetUnitsAsStr((NCSimUnitsType) i),
			       NCSimwxString2Chars(Data)) == 0) {
			ThisNCSim -> Units = (NCSimUnitsType) i;
			ThisNCSim -> OverwriteDefaults = TRUE;
			break;
		    }
		}
		assert(i < NCSIM_UNIT_LAST);
	        break;
	}
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Traps events from a button.		                                     M
*                                                                            *
* PARAMETERS:                                                                M
*   Event: N.S.F.I.                                                          M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimLoadNCInfoClass::OnButton                                           M
*****************************************************************************/
void NCSimLoadNCInfoClass::OnButton(wxCommandEvent &Event)
{
    NCSIM_DEBUG_FUNC_ENTRY

    wxButton
	*Btn = (wxButton *) Event.GetEventObject();

    if (Event.GetId() == wxID_NCSIM_PARAM_OK) {
	Hide();
	ThisNCSim -> SaveSimConfig();
    }
}

/*****************************************************************************
* AUXILIARY:								     *
*    Event table for NCSimulationParamClass				     *
*****************************************************************************/
BEGIN_EVENT_TABLE(NCSimulationParamClass, wxDialog)
    EVT_CHECKBOX(wxID_NCSIM_PARAM_NC_Z_BUF, OnCheckBox)
    EVT_CHECKBOX(wxID_NCSIM_PARAM_PAUSE_ERROR, OnCheckBox)
    EVT_CHECKBOX(wxID_NCSIM_PARAM_UPDATE_FEEDRATES, OnCheckBox)
    EVT_TEXT(wxID_NCSIM_PARAM_TOOL_NUM, OnTextChange)
    EVT_TEXT(wxID_NCSIM_PARAM_ZBUF_RES, OnTextChange)
    EVT_TEXT(wxID_NCSIM_PARAM_UNITS, OnComboBox)
    EVT_TEXT(wxID_NCSIM_PARAM_STOCK_GEOM_INIT, OnComboBox)
    EVT_TEXT(wxID_NCSIM_PARAM_TOOL_TYPE, OnComboBox)
    EVT_BUTTON(wxID_NCSIM_PARAM_SIM_MODE, OnButton)
    EVT_BUTTON(wxID_NCSIM_PARAM_TOOL_ICOLOR, OnButton)
    EVT_BUTTON(wxID_NCSIM_PARAM_TOOL_FCOLOR, OnButton)
    EVT_BUTTON(wxID_NCSIM_PARAM_TOOL_CCOLOR, OnButton)
    EVT_BUTTON(wxID_NCSIM_PARAM_STOCK_COLOR, OnButton)
    EVT_BUTTON(wxID_NCSIM_PARAM_OK, OnButton)
    EVT_BUTTON(wxID_NCSIM_PARAM_SAVE, OnButton)
    EVT_BUTTON(wxID_NCSIM_PARAM_RESET_FEEDRATES, OnButton)
END_EVENT_TABLE()

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Sets the given button background color to Clr (and compensate the        *
* foreground accordingly.)                                                   *
*                                                                            *
* PARAMETERS:                                                                *
*   Btn:  Button to update its background color.                             *
*   Clr:  New color to use.				                     *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void SetButtonBGColor(wxWindow *Btn, int *Clr)
{
    NCSIM_DEBUG_FUNC_ENTRY

    Btn -> SetBackgroundColour(wxColour(Clr[0],
					Clr[1],
					Clr[2]));
    if (Clr[0] + Clr[1] + Clr[2] > 374)
	Btn -> SetForegroundColour(wxColour(0, 0, 0));
    else
	Btn -> SetForegroundColour(wxColour(255, 255, 255));
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Constructor for the interactive setup control of NC loaded file.         M
*                                                                            *
* PARAMETERS:                                                                M
*   NCSim:       Simulation we should create this page for.		     M
*   Parent:      Frame window.					             M
*   Id:          ID assigned to window. 		                     M
*   Title:       Window's title.		       		             M
*   Flags:       Window styles' flags.				             M
*                                                                            *
* RETURN VALUE:                                                              M
*   NoValue                                                                  M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimulationParamClass::NCSimulationParamClass                           M
*****************************************************************************/
NCSimulationParamClass::NCSimulationParamClass(NCSimulationClass *NCSim,
					       wxWindow *Parent,
					       wxWindowID Id,
					       const wxString &Title,
					       int Flags):
	wxDialog(Parent,
		 Id,
		 Title,
		 wxDefaultPosition,
		 wxDefaultSize,
		 Flags | wxCAPTION)
{
    NCSIM_DEBUG_FUNC_ENTRY

    wxString Str;
    wxBoxSizer *Sizer,
	*TopSizer = new wxBoxSizer(wxVERTICAL);
    wxButton *Btn, *BtnToolIdleColor;
    wxCheckBox *CBtn;
    wxStaticText *Ln;
    wxComboBox *UnitsCBox;
    wxSpinCtrl *SCtrl;
    wxStaticBox *SBox;
    RealSpinCtrlClass *RSpin;
    wxArrayString TTypeStrs, UnitsStrs, StockShapeStrs;
    char Line[IRIT_LINE_LEN];

    ThisNCSim = NCSim;    /* Keep a ref. to NCSimulationClass we belong to. */

    SetSizer(TopSizer);

    /* Row. */
    Sizer = new wxBoxSizer(wxHORIZONTAL);

    CBtn = new wxCheckBox(this, wxID_NCSIM_PARAM_NC_Z_BUF,
			  wxT("NC Z-Buffer"));
    CBtn -> Set3StateValue(ThisNCSim -> NCZBufferOn ? wxCHK_CHECKED
						    : wxCHK_UNCHECKED);
    CBtn -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_NC_Z_BUF));
    Sizer -> Add(CBtn, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    SimModeBtn = new wxButton(this, wxID_NCSIM_PARAM_SIM_MODE,
	NCSim -> NCSimMode == NCSIM_MODE_TIME ? wxT("Real Time Sim")
					       : wxT("Fixed Step Sim"));
    SimModeBtn -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_SIM_MODE));
    Sizer -> Add(SimModeBtn, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    SimStepLn = new wxStaticText(this, wxID_ANY,
	NCSim -> NCSimMode == NCSIM_MODE_TIME ? wxT("Simulation\nTime Step")
					      : wxT("Simulation\nFixed Step"));
    SimStepLn -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_SIM_STEP));
    Sizer -> Add(SimStepLn, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    SimStepRSpin = new RealSpinCtrlClass(&NCSim -> NCSimStep,
			IRIT_EPS, IRIT_INFNTY,
			NCSim -> Units == NCSIM_UNIT_MM ? 0.01 : 0.001,
			NCSim -> Units == NCSIM_UNIT_MM ? 0.1 : 0.01,
			this, wxID_NCSIM_PARAM_SIM_STEP,
			wxDefaultPosition, wxSize(45, 20));
    SimStepRSpin -> SetUpdateFunction(RealSpinUpdateValCB);
    SimStepRSpin -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_SIM_STEP));
    Sizer -> Add(SimStepRSpin, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    TopSizer -> Add(Sizer, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    /* Row. */
    Sizer = new wxBoxSizer(wxHORIZONTAL);

    Str.Printf(wxT("   Max MRR\n   (%s)"),
	       NCSim -> GetMaterialRemovalRateUnitsAsStr(NCSim -> Units));
    MRRLn = new wxStaticText(this, wxID_ANY, Str);
    MRRLn -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_MAX_MRR));
    Sizer -> Add(MRRLn, 0, wxALIGN_CENTER_VERTICAL | wxALL, 0);

    RSpin = new RealSpinCtrlClass(&NCSim -> MaxMaterialRemovalRate,
			IRIT_EPS, IRIT_INFNTY,
			NCSim -> Units == NCSIM_UNIT_MM ? 0.01 : 0.001,
			NCSim -> Units == NCSIM_UNIT_MM ? 0.1 : 0.01,
			this, wxID_NCSIM_PARAM_MAX_MRR,
			wxDefaultPosition, wxSize(45, 20));
    RSpin -> SetUpdateFunction(RealSpinUpdateValCB);
    RSpin -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_MAX_MRR));
    Sizer -> Add(RSpin, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    TopSizer -> Add(Sizer, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);


    CBtn = new wxCheckBox(this, wxID_NCSIM_PARAM_UPDATE_FEEDRATES, wxT(""));
    CBtn -> Set3StateValue(NCSim -> UpdateFeedRates ? wxCHK_CHECKED
						    : wxCHK_UNCHECKED);
    CBtn -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_UPDATE_FEEDRATES));
    Sizer -> Add(CBtn, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    Ln = new wxStaticText(this, wxID_ANY, wxT("Update\nFeedrates  "));
    Ln -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_UPDATE_FEEDRATES));
    Sizer -> Add(Ln, 0, wxALIGN_CENTER_VERTICAL | wxALL, 0);

    Btn = new wxButton(this, wxID_NCSIM_PARAM_RESET_FEEDRATES,
		       wxT("Reset Feed Rates"));
    Btn -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_RESET_FEEDRATES));
    Sizer -> Add(Btn, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);


    /* Row. */
    Sizer = new wxBoxSizer(wxHORIZONTAL);

    CBtn = new wxCheckBox(this, wxID_NCSIM_PARAM_PAUSE_ERROR, wxT(""));
    CBtn -> Set3StateValue(ThisNCSim -> PauseOnError ? wxCHK_CHECKED
						     : wxCHK_UNCHECKED);
    CBtn -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_PAUSE_ERROR));
    Sizer -> Add(CBtn, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    Ln = new wxStaticText(this, wxID_ANY, wxT("Pause\non Error  "));
    Ln -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_PAUSE_ERROR));
    Sizer -> Add(Ln, 0, wxALIGN_CENTER_VERTICAL | wxALL, 0);

    Ln = new wxStaticText(this, wxID_ANY, wxT("Shaped\nStock:"));
    Ln -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_STOCK_GEOM_INIT));
    Sizer -> Add(Ln, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    StockShapeStrs.Add(wxT("None"));
    StockShapeStrs.Add(wxT("Plane"));
    StockShapeStrs.Add(wxT("Sphere Cap"));
    StockShapeStrs.Add(wxT("Sphere Cavity"));
    StockShapeStrs.Add(wxT("Cylin. Cap"));
    StockShapeStrs.Add(wxT("Cylin. Cavity"));
    StockShapeStrs.Add(wxT("General Srf"));
    ShapedStockCBox = new wxComboBox(this, wxID_NCSIM_PARAM_STOCK_GEOM_INIT,
	    NCSimChars2wxString(NCSim -> GetStockGeomInitAsStr(NCSim -> 
						         StockGeomInit.Type)),
	    wxDefaultPosition, wxSize(80, -1), StockShapeStrs,
	    wxCB_DROPDOWN | wxCB_READONLY);
    ShapedStockCBox ->
		  SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_STOCK_GEOM_INIT));
    Sizer -> Add(ShapedStockCBox, 0, wxALIGN_CENTER_VERTICAL | wxALL, 0);

    TopSizer -> Add(Sizer, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    Ln = new wxStaticText(this, wxID_ANY, wxT("   ZBuffer\n   Resolution:"));
    Ln -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_ZBUF_RES));
    Sizer -> Add(Ln, 0, wxALIGN_CENTER_VERTICAL | wxALL, 0);

    sprintf(Line, "%d", NCSim -> ZBufSize);
    SCtrl = new wxSpinCtrl(this, wxID_NCSIM_PARAM_ZBUF_RES,
			   NCSimChars2wxString(Line),
			   wxDefaultPosition, wxSize(60, 20),
			   wxTE_RICH | wxSP_ARROW_KEYS | wxSP_WRAP,
			   NCSIM_MIN_ZBUFFER_SIZE,
			   NCSIM_MAX_ZBUFFER_SIZE,
			   ThisNCSim -> ZBufSize);
    SCtrl -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_ZBUF_RES));
    Sizer -> Add(SCtrl, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    /* Row. */
    Sizer = new wxBoxSizer(wxHORIZONTAL);

    Ln = new wxStaticText(this, wxID_ANY, wxT("Units:"));
    Ln -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_UNITS));
    Sizer -> Add(Ln, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    UnitsStrs.Add(NCSimChars2wxString(NCSim -> GetUnitsAsStr(NCSIM_UNIT_MM)));
    UnitsStrs.Add(NCSimChars2wxString(NCSim -> GetUnitsAsStr(NCSIM_UNIT_INCHES)));
    UnitsCBox = new wxComboBox(this, wxID_NCSIM_PARAM_UNITS,
			  NCSimChars2wxString(NCSim -> GetUnitsAsStr(NCSim -> Units)),
			  wxDefaultPosition, wxDefaultSize,
			  UnitsStrs, wxCB_DROPDOWN | wxCB_READONLY);
    UnitsCBox -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_UNITS));
    Sizer -> Add(UnitsCBox, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    sprintf(Line, "Feed Rate:\n(%s)",
	    ThisNCSim -> GetFeedRateUnitsAsStr(NCSim -> Units));
    FeedRateLn = new wxStaticText(this, wxID_ANY, NCSimChars2wxString(Line));
    FeedRateLn -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_FEED_RATE));
    Sizer -> Add(FeedRateLn, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    FeedRateRSpin = new RealSpinCtrlClass(&NCSim -> FeedRate, IRIT_EPS,
					  IRIT_INFNTY, 0.01, 0.1, this,
					  wxID_NCSIM_PARAM_FEED_RATE,
					  wxDefaultPosition, wxSize(50, 20));
    FeedRateRSpin -> SetUpdateFunction(RealSpinUpdateValCB);
    FeedRateRSpin -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_FEED_RATE));
    FeedRateRSpin -> Disable();
    Sizer -> Add(FeedRateRSpin, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    sprintf(Line, "Spindle\nSpeed:");
    Ln = new wxStaticText(this, wxID_ANY, NCSimChars2wxString(Line));
    Ln -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_SPINDLE_SP));
    Sizer -> Add(Ln, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    SpindleSpeedRSpin = new RealSpinCtrlClass(&NCSim -> SpindleSpeed, IRIT_EPS,
				  IRIT_INFNTY, 1, 10, this,
				  wxID_NCSIM_PARAM_SPINDLE_SP,
				  wxDefaultPosition, wxSize(50, 20));
    SpindleSpeedRSpin -> SetUpdateFunction(RealSpinUpdateValCB);
    SpindleSpeedRSpin -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_SPINDLE_SP));
    SpindleSpeedRSpin -> Disable();
    Sizer -> Add(SpindleSpeedRSpin, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    TopSizer -> Add(Sizer, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    /* Row. */
    Sizer = new wxBoxSizer(wxHORIZONTAL);

    Ln = new wxStaticText(this, wxID_ANY, wxT("Fast Speed\nX Factor:"));
    Ln -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_FAST_SPEED_FACTOR));
    Sizer -> Add(Ln, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    FastFeedRateRSpin = new RealSpinCtrlClass(
				&NCSim -> AnimFastFeedRateMultFactor,
				IRIT_EPS, IRIT_INFNTY,
				NCSim -> Units == NCSIM_UNIT_MM ? 0.01 : 0.001,
				NCSim -> Units == NCSIM_UNIT_MM ? 0.1 : 0.01,
				this, wxID_NCSIM_PARAM_FAST_SPEED_FACTOR,
				wxDefaultPosition, wxSize(50, 20));
    FastFeedRateRSpin -> SetUpdateFunction(RealSpinUpdateValCB);
    FastFeedRateRSpin ->
                 SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_FAST_SPEED_FACTOR));
    Sizer -> Add(FastFeedRateRSpin, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    Ln = new wxStaticText(this, wxID_ANY, wxT("Tool\nIdx:"));
    Ln -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_TOOL_NUM));
    Sizer -> Add(Ln, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    sprintf(Line, "%d", NCSim -> ToolIdx);
    ToolIndexSpin = new wxSpinCtrl(this, wxID_NCSIM_PARAM_TOOL_NUM,
			   NCSimChars2wxString(Line),
			   wxDefaultPosition, wxSize(40, 20),
			   wxTE_RICH | wxSP_ARROW_KEYS | wxSP_WRAP, 0,
			   NCSIM_MAX_NUM_OF_TOOLS - 1, 0);
    ToolIndexSpin -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_TOOL_NUM));
    Sizer -> Add(ToolIndexSpin, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    Ln = new wxStaticText(this, wxID_ANY, wxT("Geom.\nApprox:"));
    Ln -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_GEOM_APPROX_TOL));
    Sizer -> Add(Ln, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    GeomApproxTolRSpin = new RealSpinCtrlClass(&NCSim -> GeomApproxTolerance,
			NCSim -> Units == NCSIM_UNIT_MM ? 0.01 : 0.001,
			IRIT_INFNTY,
			NCSim -> Units == NCSIM_UNIT_MM ? 0.01 : 0.001,
			NCSim -> Units == NCSIM_UNIT_MM ? 0.1 : 0.01,
			this, wxID_NCSIM_PARAM_GEOM_APPROX_TOL,
			wxDefaultPosition, wxSize(60, 20));
    GeomApproxTolRSpin -> SetUpdateFunction(RealSpinUpdateValCB);
    GeomApproxTolRSpin ->
		   SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_GEOM_APPROX_TOL));
    Sizer -> Add(GeomApproxTolRSpin, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);


    TopSizer -> Add(Sizer, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    /* Row. */
    Sizer = new wxBoxSizer(wxHORIZONTAL);

    Ln = new wxStaticText(this, wxID_ANY, wxT("Tool\nType:"));
    Ln -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_TOOL_TYPE));
    Sizer -> Add(Ln, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    TTypeStrs.Add(NCSimChars2wxString(NCSim ->
				    GetToolTypeAsStr(NCSIM_TOOL_TYPE_NONE)));
    TTypeStrs.Add(NCSimChars2wxString(NCSim -> 
				GetToolTypeAsStr(NCSIM_TOOL_TYPE_BALL_END)));
    TTypeStrs.Add(NCSimChars2wxString(NCSim -> 
			       GetToolTypeAsStr(NCSIM_TOOL_TYPE_TORUS_END)));
    TTypeStrs.Add(NCSimChars2wxString(NCSim ->
				GetToolTypeAsStr(NCSIM_TOOL_TYPE_FLAT_END)));
    TTypeStrs.Add(NCSimChars2wxString(NCSim ->
				   GetToolTypeAsStr(NCSIM_TOOL_TYPE_OTHER)));
    ToolTypeCBox = new wxComboBox(this, wxID_NCSIM_PARAM_TOOL_TYPE,
	      NCSimChars2wxString(NCSim -> GetToolTypeAsStr(
				      NCSim -> Tools[NCSim -> ToolIdx].Type)),
	      wxDefaultPosition, wxDefaultSize, TTypeStrs,
	      wxCB_DROPDOWN | wxCB_READONLY);
    ToolTypeCBox -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_TOOL_TYPE));
    Sizer -> Add(ToolTypeCBox, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    Ln = new wxStaticText(this, wxID_ANY, wxT("Tool\nDiam.:"));
    Ln -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_TOOL_DIAM));
    Sizer -> Add(Ln, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    ToolDiameterRSpin = new RealSpinCtrlClass(
			NCSim -> Tools[NCSim -> ToolIdx].Diameter,
			RealSpinUpdateValCB, IRIT_EPS, IRIT_INFNTY,
			NCSim -> Units == NCSIM_UNIT_MM ? 0.01 : 0.001,
			NCSim -> Units == NCSIM_UNIT_MM ? 0.1 : 0.01,
			this, wxID_NCSIM_PARAM_TOOL_DIAM,
			wxDefaultPosition, wxSize(50, 20));
    ToolDiameterRSpin -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_TOOL_DIAM));
    Sizer -> Add(ToolDiameterRSpin, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    Ln = new wxStaticText(this, wxID_ANY, wxT("Torus\nRad.:"));
    Ln -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_TOOL_TORUS_RAD));
    Sizer -> Add(Ln, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    ToolTorusRadiusRSpin = new RealSpinCtrlClass(
			NCSim -> Tools[NCSim -> ToolIdx].TorusRadius,
			RealSpinUpdateValCB, IRIT_EPS, IRIT_INFNTY,
			NCSim -> Units == NCSIM_UNIT_MM ? 0.01 : 0.001,
			NCSim -> Units == NCSIM_UNIT_MM ? 0.1 : 0.01,
			this, wxID_NCSIM_PARAM_TOOL_TORUS_RAD,
			wxDefaultPosition, wxSize(45, 20));
    ToolTorusRadiusRSpin ->
		     SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_TOOL_TORUS_RAD));
    Sizer -> Add(ToolTorusRadiusRSpin, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);
    if (NCSim -> Tools[NCSim -> ToolIdx].Type == NCSIM_TOOL_TYPE_TORUS_END)
	ToolTorusRadiusRSpin -> Enable();
    else
        ToolTorusRadiusRSpin -> Disable();
    TopSizer -> Add(Sizer, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    /* Row. */
    Sizer = new wxBoxSizer(wxHORIZONTAL);

    BtnToolIdleColor = new wxButton(this, wxID_NCSIM_PARAM_TOOL_ICOLOR,
				    wxT("Tool Idle Color"));
    SetButtonBGColor(BtnToolIdleColor, NCSim -> ToolIdleColor);
    BtnToolIdleColor -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_TOOL_ICOLOR));
    Sizer -> Add(BtnToolIdleColor, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    Btn = new wxButton(this, wxID_NCSIM_PARAM_TOOL_FCOLOR,
		       wxT("Tool Fast Color"));
    SetButtonBGColor(Btn, NCSim -> ToolFastColor);
    Btn -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_TOOL_FCOLOR));
    Sizer -> Add(Btn, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    Btn = new wxButton(this, wxID_NCSIM_PARAM_TOOL_CCOLOR,
		       wxT("Tool Cut Color"));
    SetButtonBGColor(Btn, NCSim -> ToolCutColor);
    Btn -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_TOOL_CCOLOR));
    Sizer -> Add(Btn, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    TopSizer -> Add(Sizer, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    /* Row. */
    Sizer = new wxBoxSizer(wxHORIZONTAL);

    Ln = new wxStaticText(this, wxID_ANY, wxT("Min XYZ\nDim.:"));
    Ln -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_STOCK_BBOX));
    Sizer -> Add(Ln, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    StockBBoxMinRSpin[0] = new RealSpinCtrlClass(&NCSim -> StockBBox.Min[0],
			-IRIT_INFNTY, IRIT_INFNTY, 0.01, 0.1, this,
			wxID_NCSIM_PARAM_STOCK_BBOX, wxDefaultPosition,
			wxSize(50, 20));
    StockBBoxMinRSpin[0] -> SetUpdateFunction(RealSpinUpdateValCB);
    StockBBoxMinRSpin[0] ->
		         SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_STOCK_BBOX));
    Sizer -> Add(StockBBoxMinRSpin[0], 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    StockBBoxMinRSpin[1] = new RealSpinCtrlClass(&NCSim -> StockBBox.Min[1],
			-IRIT_INFNTY, IRIT_INFNTY, 0.01, 0.1, this,
			wxID_NCSIM_PARAM_STOCK_BBOX, wxDefaultPosition,
			wxSize(50, 20));
    StockBBoxMinRSpin[1] -> SetUpdateFunction(RealSpinUpdateValCB);
    StockBBoxMinRSpin[1] ->
		         SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_STOCK_BBOX));
    Sizer -> Add(StockBBoxMinRSpin[1], 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    StockBBoxMinRSpin[2] = new RealSpinCtrlClass(&NCSim -> StockBBox.Min[2],
			-IRIT_INFNTY, IRIT_INFNTY, 0.01, 0.1, this,
			wxID_NCSIM_PARAM_STOCK_BBOX, wxDefaultPosition,
			wxSize(50, 20));
    StockBBoxMinRSpin[2] -> SetUpdateFunction(RealSpinUpdateValCB);
    StockBBoxMinRSpin[2] ->
			 SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_STOCK_BBOX));
    Sizer -> Add(StockBBoxMinRSpin[2], 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    Btn = new wxButton(this, wxID_NCSIM_PARAM_STOCK_COLOR,
		       wxT("Stock Color"));
    SetButtonBGColor(Btn, NCSim -> StockColor);
    Btn -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_STOCK_COLOR));
    Sizer -> Add(Btn, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);


    TopSizer -> Add(Sizer, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    /* Row. */
    Sizer = new wxBoxSizer(wxHORIZONTAL);

    Ln = new wxStaticText(this, wxID_ANY, wxT("Max XYZ\nDim.:"));
    Ln -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_STOCK_BBOX));
    Sizer -> Add(Ln, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    StockBBoxMaxRSpin[0] = new RealSpinCtrlClass(&NCSim -> StockBBox.Max[0],
			-IRIT_INFNTY, IRIT_INFNTY, 0.01, 0.1, this,
			wxID_NCSIM_PARAM_STOCK_BBOX, wxDefaultPosition,
			wxSize(50, 20));
    StockBBoxMaxRSpin[0] -> SetUpdateFunction(RealSpinUpdateValCB);
    StockBBoxMaxRSpin[0] ->
			 SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_STOCK_BBOX));
    Sizer -> Add(StockBBoxMaxRSpin[0], 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    StockBBoxMaxRSpin[1] = new RealSpinCtrlClass(&NCSim -> StockBBox.Max[1],
			-IRIT_INFNTY, IRIT_INFNTY, 0.01, 0.1, this,
			wxID_NCSIM_PARAM_STOCK_BBOX, wxDefaultPosition,
			wxSize(50, 20));
    StockBBoxMaxRSpin[1] -> SetUpdateFunction(RealSpinUpdateValCB);
    StockBBoxMaxRSpin[1] ->
			 SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_STOCK_BBOX));
    Sizer -> Add(StockBBoxMaxRSpin[1], 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    StockBBoxMaxRSpin[2] = new RealSpinCtrlClass(&NCSim -> StockBBox.Max[2],
			-IRIT_INFNTY, IRIT_INFNTY, 0.01, 0.1, this,
			wxID_NCSIM_PARAM_STOCK_BBOX, wxDefaultPosition,
			wxSize(50, 20));
    StockBBoxMaxRSpin[2] -> SetUpdateFunction(RealSpinUpdateValCB);
    StockBBoxMaxRSpin[2] ->
			 SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_STOCK_BBOX));
    Sizer -> Add(StockBBoxMaxRSpin[2], 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    Ln = new wxStaticText(this, wxID_ANY, wxT("Stock\nTransp:"));
    Ln -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_STOCK_TRANSP));
    Sizer -> Add(Ln, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    RSpin = new RealSpinCtrlClass(&NCSim -> StockTransp, 0.0, 1.0, 0.001, 0.01,
				  this, wxID_NCSIM_PARAM_STOCK_TRANSP,
				  wxDefaultPosition, wxSize(35, 20));
    RSpin -> SetToolTip(NCSimToolTip(wxID_NCSIM_PARAM_STOCK_TRANSP));
    Sizer -> Add(RSpin, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    TopSizer -> Add(Sizer, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    /* Last Row. */
    Sizer = new wxBoxSizer(wxHORIZONTAL);

    Btn = new wxButton(this, wxID_NCSIM_PARAM_OK, wxT("OK"));
    Sizer -> Add(Btn, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    Btn = new wxButton(this, wxID_NCSIM_PARAM_SAVE, wxT("Save"));
    Sizer -> Add(Btn, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    TopSizer -> Add(Sizer, 0, wxALIGN_CENTER_VERTICAL | wxALL, 5);

    /* Done. */
    TopSizer -> Fit(this);
    TopSizer -> SetSizeHints(this);

    /* Add bounding boxes relative to already positioned widgets. */
    int x, y;

    SimModeBtn -> GetPosition(&x, &y);
    SBox = new wxStaticBox(this, -1, wxT("Feedrate control"),
			   wxPoint(4, 28 + y), wxSize(343, 51));

    ShapedStockCBox -> GetPosition(&x, &y);
    SBox = new wxStaticBox(this, -1, wxT("GCode"),
			   wxPoint(4, 28 + y), wxSize(343, 51));

    UnitsCBox -> GetPosition(&x, &y);
    SBox = new wxStaticBox(this, -1, wxT("Tool"),
			   wxPoint(4, 28 + y), wxSize(343, 149));

    BtnToolIdleColor -> GetPosition(&x, &y);
    SBox = new wxStaticBox(this, -1, wxT("Stock"),
			   wxPoint(4, 28 + y), wxSize(343, 100));
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Refresh the values displayed in the control.			     M
*                                                                            *
* PARAMETERS:                                                                M
*   None	                                                             M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimulationParamClass::Refresh                                          M
*****************************************************************************/
void NCSimulationParamClass::Refresh()
{
    NCSIM_DEBUG_FUNC_ENTRY

    SimModeBtn -> SetLabel(
	ThisNCSim -> NCSimMode == NCSIM_MODE_TIME ? wxT("Real Time Sim")
					          : wxT("Fixed Step Sim"));

    SimStepRSpin -> Refresh();

    ShapedStockCBox -> SetValue(NCSimChars2wxString(
	 ThisNCSim -> GetStockGeomInitAsStr(ThisNCSim -> StockGeomInit.Type)));

    FeedRateRSpin -> Refresh();

    SpindleSpeedRSpin -> Refresh();

    FastFeedRateRSpin -> Refresh();

    GeomApproxTolRSpin -> Refresh();

    ToolIndexSpin -> Refresh();

    ToolDiameterRSpin -> Refresh();

    ToolTorusRadiusRSpin -> Refresh();

    StockBBoxMinRSpin[0] -> Refresh();
    StockBBoxMinRSpin[1] -> Refresh();
    StockBBoxMinRSpin[2] -> Refresh();

    StockBBoxMaxRSpin[0] -> Refresh();
    StockBBoxMaxRSpin[1] -> Refresh();
    StockBBoxMaxRSpin[2] -> Refresh();
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Verify we current have a valid bbox.				     M
*                                                                            *
* PARAMETERS:                                                                M
*   None	                                                             M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimulationParamClass::VerifyBBox                                       M
*****************************************************************************/
void NCSimulationParamClass::VerifyBBox()
{
    NCSIM_DEBUG_FUNC_ENTRY

    int i;
    GMBBBboxStruct
        *StockBBox = &ThisNCSim -> StockBBox;

    for (i = 0; i < 3; i++) {
        if (StockBBox -> Max[i] <= StockBBox -> Min[i]) {
	    IRIT_SWAP(IrtRType, StockBBox -> Max[i], StockBBox -> Min[i]);
	    StockBBox -> Max[i] += IRIT_EPS;

	    StockBBoxMinRSpin[i] -> SetRealValue(StockBBox -> Min[i]);
	    StockBBoxMaxRSpin[i] -> SetRealValue(StockBBox -> Max[i]);
	}
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Call back function of a RealSpin object, after its value being modified. *
*                                                                            *
* PARAMETERS:                                                                *
*   Val:   New values.							     *
*   Id:    ID of modified window.					     *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void RealSpinUpdateValCB(double Val, wxWindowID Id)
{
    NCSIM_DEBUG_FUNC_ENTRY

    NCSimulationClass
        *CrntSim = NCSimMF -> CrntSim;

    if (CrntSim == NULL || CrntSim -> AllInfo == NULL)
	return; /* Still building the panel... */

    if (!RealSpinUpdateInCB)
        return;

    switch (Id) {
	case wxID_NCSIM_PARAM_SIM_STEP:
	    break;
	case wxID_NCSIM_PARAM_FEED_RATE:
	    break;
	case wxID_NCSIM_PARAM_SPINDLE_SP:
	    break;
	case wxID_NCSIM_PARAM_FAST_SPEED_FACTOR:
	    break;
	case wxID_NCSIM_PARAM_GEOM_APPROX_TOL:
	    if (CrntSim -> Tools[CrntSim -> ToolIdx].Diameter *
					NCSIM_DEF_TOOL_APPROX_REL_TOL <	Val) {
	        NCSimPrintf(NCSIM_LOG_WARN,
			    "Tool geometry approximation too rough - tool will not be drawn accurately.\n");	        
	    }

	    CrntSim -> ReCreateToolGeometry();
	    NCSimMF -> OGLWin -> NCSimPageChange();
	    NCSimMF -> Refresh();	                 /* Force a redraw. */
	    break;
	case wxID_NCSIM_PARAM_TOOL_DIAM:
	    CrntSim -> Tools[CrntSim -> ToolIdx].Diameter = Val;
	    if (CrntSim -> Tools[CrntSim -> ToolIdx].Diameter <
		CrntSim -> Tools[CrntSim -> ToolIdx].TorusRadius * 2) {
	        RealSpinUpdateInCB = FALSE;
		CrntSim -> AllInfo -> ToolTorusRadiusRSpin ->
						      SetRealValue(Val * 0.5);
		RealSpinUpdateInCB = TRUE;
	    }
	    if (Val * NCSIM_DEF_TOOL_APPROX_REL_TOL <
					     CrntSim -> GeomApproxTolerance) {
	        NCSimPrintf(NCSIM_LOG_WARN,
			    "Tool geometry approximation too rough - updated as well to a reasonable value.\n");	        
		CrntSim -> GeomApproxTolerance =
		    Val * NCSIM_DEF_TOOL_APPROX_REL_TOL;
		CrntSim -> AllInfo -> GeomApproxTolRSpin ->
			        SetRealValue(CrntSim -> GeomApproxTolerance);
	    }

	    CrntSim -> ReCreateToolGeometry();
	    NCSimMF -> Refresh();	                 /* Force a redraw. */
	    break;
	case wxID_NCSIM_PARAM_TOOL_TORUS_RAD:
	    CrntSim -> Tools[CrntSim -> ToolIdx].TorusRadius = Val;
	    /* Verify we have a valid torus tool geometry. */
	    if (CrntSim -> Tools[CrntSim -> ToolIdx].Diameter <
		CrntSim -> Tools[CrntSim -> ToolIdx].TorusRadius * 2) {
	        RealSpinUpdateInCB = FALSE;
		CrntSim -> AllInfo -> ToolDiameterRSpin ->
							SetRealValue(Val * 2);
		RealSpinUpdateInCB = TRUE;
	    }
	    CrntSim -> ReCreateToolGeometry();
	    NCSimMF -> Refresh();	                 /* Force a redraw. */
	    break;
	case wxID_NCSIM_PARAM_STOCK_BBOX:
	    /* Verify we have a valid bbox. */
	    RealSpinUpdateInCB = FALSE;
	    CrntSim -> AllInfo -> VerifyBBox();
	    RealSpinUpdateInCB = TRUE;

	    CrntSim -> ReCreateStockGeometry();
	    CrntSim -> AnimCrntState = NCSIM_ANIM_STATE_STOP;
	    NCSimMF -> Refresh();	                 /* Force a redraw. */
	    break;
    }
    CrntSim -> OverwriteDefaults = TRUE;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Traps events from check box.                                             M
*                                                                            *
* PARAMETERS:                                                                M
*   Event: N.S.F.I.                                                          M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimulationParamClass::OnCheckBox                                       M
*****************************************************************************/
void NCSimulationParamClass::OnCheckBox(wxCommandEvent &Event)
{
    NCSIM_DEBUG_FUNC_ENTRY

    switch (Event.GetId()) {
	case wxID_NCSIM_PARAM_NC_Z_BUF:
            /* We cannot change (no)use of the Z buffer in mid-simulation. */
	    if (ThisNCSim -> NCZBufferOn = Event.IsChecked())	    
	        ThisNCSim -> AnimCrntState = NCSIM_ANIM_STATE_STOP;
	    NCSimMF -> Refresh();	                 /* Force a redraw. */
	    break;
	case wxID_NCSIM_PARAM_PAUSE_ERROR:
	    ThisNCSim -> PauseOnError = Event.IsChecked();
	    break;
	case wxID_NCSIM_PARAM_UPDATE_FEEDRATES:
	    ThisNCSim -> UpdateFeedRates = Event.IsChecked();
	    break;
    }
    ThisNCSim -> OverwriteDefaults = TRUE;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Traps events from text controllers.                                      M
*                                                                            *
* PARAMETERS:                                                                M
*   Event: N.S.F.I.                                                          M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimulationParamClass::OnTextChange                                     M
*****************************************************************************/
void NCSimulationParamClass::OnTextChange(wxCommandEvent &Event)
{
    NCSIM_DEBUG_FUNC_ENTRY

    if (Event.GetId() == wxID_NCSIM_PARAM_TOOL_NUM) {
        int i;
	wxString
	    Str = Event.GetString();
	const char
	    *s = NCSimwxString2Chars(Str);

	if (sscanf(s, "%d", &i) == 1) {
	    NCSimToolClass
		*Tool = &ThisNCSim -> Tools[i];

	    ThisNCSim -> ToolIdx = i;

	    /* Update the displayed parameters of this tool. */
	    ToolTypeCBox -> SetValue(NCSimChars2wxString(ThisNCSim ->
				              GetToolTypeAsStr(Tool -> Type)));
	    RealSpinUpdateInCB = FALSE;
	    ToolDiameterRSpin -> SetRealValue(Tool -> Diameter);
	    if (Tool -> Type == NCSIM_TOOL_TYPE_TORUS_END)
	        ToolTorusRadiusRSpin -> Enable();
	    else
	        ToolTorusRadiusRSpin -> Disable();
	    RealSpinUpdateInCB = TRUE;     /* Will also recreate and redraw. */
	    ToolTorusRadiusRSpin -> SetRealValue(Tool -> TorusRadius);
	}
    }
    else if (Event.GetId() == wxID_NCSIM_PARAM_ZBUF_RES) {
        int i;
	wxString
	    Str = Event.GetString();
	const char
	    *s = NCSimwxString2Chars(Str);

	if (sscanf(s, "%d", &i) == 1) {
	    ThisNCSim -> ZBufSize = i;

	    if (ThisNCSim -> NCZBufferOn = Event.IsChecked())	    
	        ThisNCSim -> AnimCrntState = NCSIM_ANIM_STATE_STOP;
	}
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Traps events from a combo box.	                                     M
*                                                                            *
* PARAMETERS:                                                                M
*   Event: N.S.F.I.                                                          M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimulationParamClass::OnComboBox                                       M
*****************************************************************************/
void NCSimulationParamClass::OnComboBox(wxCommandEvent &Event)
{
    NCSIM_DEBUG_FUNC_ENTRY

    if (Event.GetId() == wxID_NCSIM_PARAM_UNITS ||
	Event.GetId() == wxID_NCSIM_PARAM_STOCK_GEOM_INIT ||
	Event.GetId() == wxID_NCSIM_PARAM_TOOL_TYPE) {
        wxComboBox
	    *Win = (wxComboBox *) Event.GetEventObject();
	wxString
	    Data = Win -> GetValue();
	int i;

	switch (Event.GetId()) {
	    case wxID_NCSIM_PARAM_TOOL_TYPE:
	        for (i = NCSIM_TOOL_TYPE_FIRST;
		     i < NCSIM_TOOL_TYPE_LAST;
		     i++) {
		    if (strcmp(ThisNCSim -> GetToolTypeAsStr((NCSimToolType) i),
			       NCSimwxString2Chars(Data)) == 0) {
			ThisNCSim -> Tools[ThisNCSim -> ToolIdx].Type = 
							    (NCSimToolType) i;
			if (ThisNCSim -> Tools[ThisNCSim -> ToolIdx].Type ==
						    NCSIM_TOOL_TYPE_TORUS_END)
			    ToolTorusRadiusRSpin -> Enable();
			else
			    ToolTorusRadiusRSpin -> Disable();
			break;
		    }
		}
		assert(i < NCSIM_TOOL_TYPE_LAST);
		ThisNCSim -> ReCreateToolGeometry();
		NCSimMF -> Refresh();	                 /* Force a redraw. */
	        break;
	    case wxID_NCSIM_PARAM_UNITS:
	        for (i = NCSIM_UNIT_FIRST;
		     i < NCSIM_UNIT_LAST;
		     i++) {
		    if (strcmp(ThisNCSim -> GetUnitsAsStr((NCSimUnitsType) i),
			       NCSimwxString2Chars(Data)) == 0) {
		        char Line[IRIT_LINE_LEN];

			ThisNCSim -> Units = (NCSimUnitsType) i;

			sprintf(Line, "Feed Rate:\n(%s)",
				ThisNCSim ->
				    GetFeedRateUnitsAsStr((NCSimUnitsType) i));
			FeedRateLn -> SetLabel(NCSimChars2wxString(Line));
			sprintf(Line, "Max MRR\n(%s)",
				ThisNCSim -> 
				    GetMaterialRemovalRateUnitsAsStr(ThisNCSim
								       -> Units));
			MRRLn -> SetLabel(NCSimChars2wxString(Line));
			break;
		    }
		}
		assert(i < NCSIM_UNIT_LAST);
	        break;
	    case wxID_NCSIM_PARAM_STOCK_GEOM_INIT:
	        {
		    NCSimStockGeomInitType Type;
		    const char
			*Str = NCSimwxString2Chars(Data);

		    if (strcmp("None", Str) == 0)
		        Type = NCSIM_STOCK_GEOM_INIT_NONE;
		    else if (strcmp("Plane", Str) == 0)
		        Type = NCSIM_STOCK_GEOM_INIT_PLANE;
		    else if (strcmp("Sphere Cap", Str) == 0)
		        Type = NCSIM_STOCK_GEOM_INIT_SPHERE_CAP;
		    else if (strcmp("Sphere Cavity", Str) == 0)
		        Type = NCSIM_STOCK_GEOM_INIT_SPHERE_CAVITY;
		    else if (strcmp("Cylin. Cap", Str) == 0)
		        Type = NCSIM_STOCK_GEOM_INIT_CYLIN_CAP;
		    else if (strcmp("Cylin. Cavity", Str) == 0)
		        Type = NCSIM_STOCK_GEOM_INIT_CYLIN_CAVITY;
		    else if (strcmp("General Srf", Str) == 0)
		        Type = NCSIM_STOCK_GEOM_INIT_GEN_SRF;
		    else
		        assert(0);

		    ThisNCSim -> GetStockGeomInitParams(Type);
		}
	        break;
	}
    }
    ThisNCSim -> OverwriteDefaults = TRUE;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Traps events from a button.		                                     M
*                                                                            *
* PARAMETERS:                                                                M
*   Event: N.S.F.I.                                                          M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   NCSimulationParamClass::OnButton                                         M
*****************************************************************************/
void NCSimulationParamClass::OnButton(wxCommandEvent &Event)
{
    NCSIM_DEBUG_FUNC_ENTRY

    wxButton
	*Btn = (wxButton *) Event.GetEventObject();

    if (Event.GetId() == wxID_NCSIM_PARAM_TOOL_ICOLOR ||
	Event.GetId() == wxID_NCSIM_PARAM_TOOL_FCOLOR ||
	Event.GetId() == wxID_NCSIM_PARAM_TOOL_CCOLOR ||
	Event.GetId() == wxID_NCSIM_PARAM_STOCK_COLOR) {
        wxColourData ClrData;
	int i, *Clr;
	IrtRType c;

	switch (Event.GetId()) {
	    case wxID_NCSIM_PARAM_TOOL_ICOLOR:
	        Clr = ThisNCSim -> ToolIdleColor;
	        break;
	    case wxID_NCSIM_PARAM_TOOL_FCOLOR:
	        Clr = ThisNCSim -> ToolFastColor;
	        break;
	    case wxID_NCSIM_PARAM_TOOL_CCOLOR:
	        Clr = ThisNCSim -> ToolCutColor;
	        break;
	    case wxID_NCSIM_PARAM_STOCK_COLOR:
	        Clr = ThisNCSim -> StockColor;
	        break;
	}

	c = 255.0 / (IRIT_MAX(IRIT_MAX(Clr[0], Clr[1]), Clr[2]) * 15.0);
	ClrData.SetChooseFull(true);
	ClrData.SetColour(wxColour(Clr[0], Clr[1], Clr[2]));
	for (i = 0; i < 16; i++) {
	    ClrData.SetCustomColour(i,
				    wxColour(Clr[0] * i * c,
					     Clr[1] * i * c,
					     Clr[2] * i * c));
	}

	wxColourDialog
	    ClrDialog(this, &ClrData);

	if (ClrDialog.ShowModal() == wxID_OK) {
	    wxColourData
	        RetData = ClrDialog.GetColourData();
	    wxColour
	        wxColour = RetData.GetColour();

	    Clr[0] = wxColour.Red();
	    Clr[1] = wxColour.Green();
	    Clr[2] = wxColour.Blue();
	}

	SetButtonBGColor(Btn, Clr);
	ThisNCSim -> OverwriteDefaults = TRUE;
    }
    else if (Event.GetId() == wxID_NCSIM_PARAM_SIM_MODE) {
	switch (ThisNCSim -> NCSimMode) {
	    case NCSIM_MODE_TIME:
	        ThisNCSim -> NCSimMode = NCSIM_MODE_STEP;
		Btn -> SetLabel(wxT("Fixed Step Sim"));
		SimStepLn -> SetLabel(wxT("Simulation\nFixed Step"));
		ThisNCSim -> UpdateSimDisplayedStatus();
	        break;
	    case NCSIM_MODE_STEP:
	        ThisNCSim -> NCSimMode = NCSIM_MODE_TIME;
		Btn -> SetLabel(wxT("Real Time Sim"));
		SimStepLn -> SetLabel(wxT("Simulation\nTime Step"));
		ThisNCSim -> UpdateSimDisplayedStatus();
	        break;
	}
	ThisNCSim -> OverwriteDefaults = TRUE;
    }
    else if (Event.GetId() == wxID_NCSIM_PARAM_OK) {
	Hide();
	ThisNCSim -> SaveSimConfig();
    }
    else if (Event.GetId() == wxID_NCSIM_PARAM_SAVE) {
	ThisNCSim -> OverwriteDefaults = TRUE;
	ThisNCSim -> SaveSimConfig();
    }
    else if (Event.GetId() == wxID_NCSIM_PARAM_RESET_FEEDRATES) {
	IPNCGCodeResetFeedRates(ThisNCSim -> GStream);
	NCSimPrintf(NCSIM_LOG_INFO,
		    "Feedrates were reset to original input file values.\n");
    }

    NCSimMF -> Refresh();		                 /* Force a redraw. */
}
