/******************************************************************************
* NCSimulation.h - Holds the main modules of the simulation of NC code.       *
*******************************************************************************
* (C) Gershon Elber					                      *
*******************************************************************************
* Written by Gershon Elber, Dec 2006.					      *
******************************************************************************/

#ifndef NCSIM_TEXT_CTRL
#define NCSIM_TEXT_CTRL

#include "irit_sm.h"
#include "geom_lib.h"
#include "misc_lib.h"
#include "user_lib.h"

#include "wx/dialog.h"
#include "wx/textctrl.h"
#include "wx/gbsizer.h"
#include "wx/combobox.h"
#include "wx/stattext.h"
#include "RealSpinCtrl.h"

#include "NCSimDefs.h"        /* Must be included first in all NCSim files. */
#include "NCSimZBuffer.h"

#define NCSIM_MAX_NUM_OF_TOOLS	100

typedef enum {
    NCSIM_TOOL_TYPE_FIRST = 0,

    NCSIM_TOOL_TYPE_NONE = NCSIM_TOOL_TYPE_FIRST,
    NCSIM_TOOL_TYPE_BALL_END,
    NCSIM_TOOL_TYPE_TORUS_END,
    NCSIM_TOOL_TYPE_FLAT_END,
    NCSIM_TOOL_TYPE_OTHER,

    NCSIM_TOOL_TYPE_LAST
} NCSimToolType;

typedef enum {
    NCSIM_ANIM_FIRST = 1,

    NCSIM_ANIM_STATE_RUN = NCSIM_ANIM_FIRST,
    NCSIM_ANIM_STATE_STOP,
    NCSIM_ANIM_STATE_PAUSE,

    NCSIM_ANIM_LAST
} NCSimAnimStateType;

typedef enum {
    NCSIM_MODE_NONE = 0,

    NCSIM_MODE_TIME,
    NCSIM_MODE_STEP,

    NCSIM_MODE_LAST
} NCSimModeType;

typedef enum {
    NCSIM_VIEW_CRNT = 0,

    NCSIM_VIEW_FRONT = 1,
    NCSIM_VIEW_SIDE = 2,
    NCSIM_VIEW_UP = 3,

    NCSIM_VIEW_LAST
} NCSimViewDirectionType;


class NCSimLoadNCInfoClass : public wxDialog
{
    public:
        NCSimLoadNCInfoClass(class NCSimulationClass *NCSim,
			     wxWindow *Parent,
			     wxWindowID Id,
			     const wxString &Title = wxT(""),
			     int Flags = 0);
	void OnCheckBox(wxCommandEvent &Event);
	void OnComboBox(wxCommandEvent &Event);
	void OnButton(wxCommandEvent &Event);

    private:
	class NCSimulationClass *ThisNCSim;

        DECLARE_EVENT_TABLE()
};

class NCSimulationParamClass : public wxDialog
{
    friend NCSimulationClass;
    public:
        NCSimulationParamClass(class NCSimulationClass *NCSim,
			       wxWindow *Parent,
			       wxWindowID Id,
			       const wxString &Title = wxT(""),
			       int Flags = 0);
	void Refresh();
	void VerifyBBox();
	void OnCheckBox(wxCommandEvent &Event);
	void OnTextChange(wxCommandEvent &Event);
	void OnComboBox(wxCommandEvent &Event);
	void OnButton(wxCommandEvent &Event);

	wxButton *SimModeBtn;
	wxComboBox *ToolTypeCBox, *ShapedStockCBox;
        wxSpinCtrl *ToolIndexSpin;
	RealSpinCtrlClass *ToolDiameterRSpin, *ToolTorusRadiusRSpin,
	    *GeomApproxTolRSpin;

    private:
	class NCSimulationClass *ThisNCSim;

        wxStaticText *FeedRateLn, *SimStepLn, *MRRLn;
	RealSpinCtrlClass *SimStepRSpin, *FeedRateRSpin, *SpindleSpeedRSpin,
	    *FastFeedRateRSpin, *StockBBoxMinRSpin[3], *StockBBoxMaxRSpin[3];

        DECLARE_EVENT_TABLE()
};

class NCSimToolClass
{
    public:
	NCSimToolClass() {
	    Type = NCSIM_TOOL_TYPE_NONE;
	    Geom = NULL;
	    Mask = NULL;
	    MaskX = MaskY = 0;
	    Diameter = TorusRadius = 0.0;
	}
	~NCSimToolClass() {
	    if (Type != NCSIM_TOOL_TYPE_NONE) {
	        if (Geom != NULL)
		    IPFreeObjectList(Geom);
		if (Mask != NULL)
		    IritFree(Mask);
	    }
	}
	NCSimToolType Type;
	IPObjectStruct *Geom;
	IrtRType Diameter;
	IrtRType TorusRadius;
	IrtRType *Mask;        /* ZDepth map of tool of size MaskX * MaskY. */
	int MaskX, MaskY;
};

class NCSimulationClass : public wxTextCtrl, public NCSimZBufferClass
{
    friend NCSimLoadNCInfoClass;
    friend NCSimulationParamClass;
    friend NCSimToolClass;
    public:
	/* Con/Destructor. */
	NCSimulationClass(const char *NCFName,
			  /* Text control flags. */
			  wxWindow *Parent,
			  wxWindowID Id,
			  const wxString &Value = wxT(""),
			  int Flags = 0,
			  /* ZBuffer flags. */
			  int ZBufSize = 1000,
			  GMBBBboxStruct *StockBBox = NULL,
			  NCSimToolType SimToolType = NCSIM_TOOL_TYPE_BALL_END,
			  IrtRType SimToolDiameter = 5.0,
			  NCSimUnitsType SimUnits = NCSIM_UNIT_MM);
	~NCSimulationClass();

	void UpdateLinePosMap();
	void HighlightTextLine(int LineNum, int Highlight = TRUE);
	bool ReCreateToolGeometry(int Idx = -1);
	void ActivateSimPage(IrtBType ForceAutoScale = FALSE);
	void AutoScaleView(NCSimViewDirectionType View = NCSIM_VIEW_CRNT);
	bool LoadGeometry(const char *FullPath,
			  NCSimObjectType ObjType = NCSIM_OBJ_GEN);
	void SetBBoxFromPathGeom(IrtRType ExpandAbsolute = 0.0,
				 IrtRType ExpandRelative = 1.0);
	IrtRType GetValInMm(IrtRType Val);
	NCSimModeType SetSimMode(NCSimModeType SimMode);
        void UpdateSimDisplayedStatus();
	void SetAnimPercent(IrtRType Percent);
	void SetAnimLine(int TargetLineNum);
	void OnIdle(wxIdleEvent &Event);

	const char *GetUnitsAsStr(NCSimUnitsType Units) {
	    return (Units == NCSIM_UNIT_MM ? "MM" : "In");
	}
	const char *GetFeedRateUnitsAsStr(NCSimUnitsType Units) {
	    return (Units == NCSIM_UNIT_MM ? "MM/Sec" : "In/Mnt");
	}
	const char *GetMaterialRemovalRateUnitsAsStr(NCSimUnitsType Units) {
	    return (Units == NCSIM_UNIT_MM ? "MM^3/Sec" : "In^3/Mnt");
	}
	const char *GetToolTypeAsStr(NCSimToolType ToolType) {
	    switch (ToolType) {
	        case NCSIM_TOOL_TYPE_NONE:
		    return "None";
	        case NCSIM_TOOL_TYPE_BALL_END:
		    return "Ball-end";
		case NCSIM_TOOL_TYPE_TORUS_END:
		    return "Torus-end";
		case NCSIM_TOOL_TYPE_FLAT_END:
		    return "Flat-end";
		case NCSIM_TOOL_TYPE_OTHER:
		default:
		    return "General";
	    }
	}
	const char *GetAnimStateAsStr(NCSimAnimStateType AnimState) {
	    switch (AnimState) {
        	case NCSIM_ANIM_STATE_RUN:
		    return "Run";
		case NCSIM_ANIM_STATE_STOP:
		    return "Stop";
		case NCSIM_ANIM_STATE_PAUSE:
		    return "Pause";
		default:
		    assert(1);			 /* Should never get here. */
		    return "";
	    }
	}
	const char *GetStockGeomInitAsStr(NCSimStockGeomInitType Type) {
	    switch (Type) {
		case NCSIM_STOCK_GEOM_INIT_NONE:
		    return "None";
		case NCSIM_STOCK_GEOM_INIT_PLANE:
		    return "Plane";
		case NCSIM_STOCK_GEOM_INIT_SPHERE_CAP:
		    return "Sphere Cap";
		case NCSIM_STOCK_GEOM_INIT_SPHERE_CAVITY:
		    return "Sphere Cavity";
		case NCSIM_STOCK_GEOM_INIT_CYLIN_CAP:
		    return "Cylin. Cap";
		case NCSIM_STOCK_GEOM_INIT_CYLIN_CAVITY:
		    return "Cylin. Cavity";
		case NCSIM_STOCK_GEOM_INIT_GEN_SRF:
		    return "General Srf";
		default:
		    assert(1);			 /* Should never get here. */
		    return "";
	    }
	}
        void SaveView(IrtHmgnMatType ViewMat) {
	    IRIT_HMGN_MAT_COPY(NCSimViewMat, IPViewMat);
	}
	void SetNCFileParams() {
	    LoadNCInfo -> ShowModal();
	    ReCreateToolGeometry();
	}
	void SetSimInfoParams();
        void AnimState(NCSimAnimStateType AnimState);
	int LoadSimConfig();
	int SaveSimConfig();

	char *NCFileName;		   /* Name of NC file to simulate . */
	VoidPtr GStream;	       /* The NC G Code parsed file stream. */
	IPObjectStruct *GPathGeom;	/* The toolpath as a long polyline. */
	int HasRelArcCenter;		     /* State control over NC file. */
	int ReverseZDir;		     /* State control over NC file. */

	/* TOols info. */
	NCSimToolClass Tools[NCSIM_MAX_NUM_OF_TOOLS];         /* All tools. */
	int ToolIdx;				     /* Current tool index. */

        /* More simulation parameters. */
	IrtRType FeedRate;
	IrtRType SpindleSpeed;
	IrtRType NCSimStep;	 /* fixed/time step simulation - step size. */
	NCSimUnitsType Units;
	int NCZBufferOn;		     /* TRUE to use the NC ZBuffer. */
	IrtRType MaterialRemovalRate;  /* Material removal rate during sim. */
	IrtRType MaxMaterialRemovalRate;     /* maximum valid removal rate. */
	int PauseOnError;	 /* TRUE to pause if an error was detected. */
	int UpdateFeedRates;      /* TRUE to correct feedrates if too fast. */
	int LastErrorGCodeLineNumber;

	/* Display info. */
        int ToolFastColor[3];
	int ToolCutColor[3];
	int ToolIdleColor[3];

	/* NCSimulation animation control. */
	IrtRType AnimMultFactor;     /* Anim time multiplier over real time.*/
	NCSimAnimStateType AnimCrntState;
	IrtRType AnimToolPosition[3];
	IPNCGCodeLineStruct *AnimGCodeInfo;

	/* Handle on the text NC GCode page in the GUI, of this simulation. */
	int GUITextPage;

	bool OverwriteDefaults;	      /* Make TRUE if defaults are changed. */

	bool SingleStepActive;

	NCSimulationParamClass *AllInfo;   /* Popup to get all sim. params. */

    private:
	IritConfigStruct *BindSimConfig(char **ToolIdleColorStr,
					char **ToolCutColorStr,
					char **ToolFastColorStr,
					char **StockColorStr,
					char **StockBBoxMinStr,
					char **StockBBoxMaxStr,
					char **StockGeomInitStr,
					char **SimModeStr,
					char **UnitsStr,
					char ***ToolsStr);
	void GetStockGeomInitParams(NCSimStockGeomInitType Type);

        bool HasToolSet() {
	    int i;

	    for (i = 0; i < NCSIM_MAX_NUM_OF_TOOLS; i++)
	        if (Tools[i].Type != NCSIM_TOOL_TYPE_NONE)
		    return TRUE;

	    return FALSE;
	};

	int NCFileNumLines;		     /* Number of lines in NC file. */
	int *NCFileLineMap;	 /* TextCtrl map of positions of all lines. */
	IrtHmgnMatType NCSimViewMat;                   /* Current view mat. */

        /* More simulation parameters. */
	int NCFileLineNum;
	NCSimModeType NCSimMode;     /* Real time or fixed step simulation. */

	/* NCSimulation animation control. */
	IrtRType AnimFastFeedRateMultFactor;   /* Fast time mult. over reg. */
	IrtRType AnimTotalArcLen;
	IrtRType AnimArcLen;
	IrtRType AnimRealTime;

	NCSimLoadNCInfoClass *LoadNCInfo;   /* Popup to get NC load params. */

        DECLARE_EVENT_TABLE()
};

#endif /* NCSIM_TEXT_CTRL */
