/*****************************************************************************
*   Generic Open GL drawing functions, complementing grap_lib.		     *
******************************************************************************
* (C) Gershon Elber, Technion, Israel Institute of Technology                *
******************************************************************************
* Written by:  Gershon Elber			         Ver 0.1, Nov 2006.  *
*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "irit_sm.h"
#include "misc_lib.h"
#include "iritprsr.h"
#include "allocate.h"
#include "attribut.h"
#include "cagd_lib.h"
#include "symb_lib.h"
#include "geom_lib.h"
#include "grap_lib.h"

#ifdef __UNIX__
#   include <Xm/Xm.h>
#   include <Xm/Form.h>
#   include <Xm/DialogS.h>
#   include <X11/cursorfont.h>
#   include <GL/glx.h>
#   include <GL/glu.h>
#   include <GL/gl.h>
#   include "xmtdrvs.h"
#endif /* __UNIX__ */

#if defined (__WINNT__)
#   include <windows.h>
#   include <gl/gl.h>
#   if _MSC_VER <= 1400
#	include <gl/glaux.h> /* VC 2005 or less. */
#   else
#   include <gl/glu.h>   /* VC 2008 or more. */
#endif /* _MSC_VER */

#   ifdef HAVE_OGL_CG_LIB
#       include <gl/glext.h>
#   endif /* HAVE_OGL_CG_LIB */
#elif defined (__WINCE__)
#   include <windows.h>
#   include <GLES/gl.h>
#   include "oglesemu.h"
#   include "wntdrvs.h"
#endif /* __WINNT__ || defined __WINCE__*/

#ifdef IRIT_DOUBLE
#   define OGL_GL_NORMAL	glNormal3d
#   define OGL_GL_VRTX_V	glVertex3dv
#else
#   define OGL_GL_NORMAL	glNormal3f
#   define OGL_GL_VRTX_V	glVertex3fv
#endif /* IRIT_FLOAT */

#define OGL_HANDLE_NORMAL(Normal) { \
			if (IGGlblFlipNormalOrient) \
			    OGL_GL_NORMAL(-Normal[0], \
					  -Normal[1], \
					  -Normal[2]); \
			else \
			    OGL_GL_NORMAL(Normal[0], \
					  Normal[1], \
					  Normal[2]); }
			    
#define TEST_OGL_ERROR  { int i; if ((i = glGetError()) != GL_NO_ERROR) \
			    IRIT_WARNING_MSG_PRINTF("OGL Error: %s\n", \
				                    gluErrorString(i)); }


IRIT_GLOBAL_DATA IrtRType
    IGGlblOpacity = 1.0;

IRIT_STATIC_DATA int
    GlblFontListBase = 0,
    GlblTextureActive = FALSE;

static void HandleVertex(IPVertexStruct *V);
static void SetColorIndex(int c);

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Updates the light source index Index to a new position/location.         M
*                                                                            *
* PARAMETERS:                                                                M
*   LightPos:       New location of light source. (0, 0, 0) disables         M
*		    light source.  If NULL, refresh all light sources color. M
*   LightColor:     Color of light source.				     M
*   LightIndex:     Index of light source in Open GL. -1 will alocate the    M
*		    next available slot.				     M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   IGSetLightSource                                                         M
*****************************************************************************/
void IGSetLightSource(IGLightType LightPos,
		      IrtVecType LightColor,
		      int LightIndex)
{
    IRIT_STATIC_DATA unsigned char
	LightSourceActive[IG_MAX_LIGHT_SOURCES] = { 0 };
    IRIT_STATIC_DATA IrtVecType
	LightSourceClr[IG_MAX_LIGHT_SOURCES] = { 0.0 };
    IRIT_STATIC_DATA IGLightType
	LightSourcePos[IG_MAX_LIGHT_SOURCES] = { 0.0 };
    int l, i;

    glLightModelf(GL_LIGHT_MODEL_TWO_SIDE, 1.0);

    if (LightPos != NULL) {
	if (LightIndex == -1) {
	    for (i = 0; i < IG_MAX_LIGHT_SOURCES; i++)
		if (!LightSourceActive[i])
		    break;
	    if (i < IG_MAX_LIGHT_SOURCES)
		LightIndex = i;
	}
	if (LightIndex >= IG_MAX_LIGHT_SOURCES || LightIndex < 0)
	    return;

	l = GL_LIGHT0 + LightIndex;
	     
	if (IRIT_APX_EQ(LightPos[0], 0.0) &&
	    IRIT_APX_EQ(LightPos[1], 0.0) &&
	    IRIT_APX_EQ(LightPos[2], 0.0)) {
	    LightSourceActive[LightIndex] = FALSE;
	    glDisable(l);
	}
	else {
	    glLightfv(l, GL_POSITION, LightPos);

	    IRIT_VEC_COPY(LightSourceClr[LightIndex], LightColor);
	    IRIT_GEN_COPY(LightSourcePos[LightIndex], LightPos,
		          sizeof(IGLightType));
	    LightSourceActive[LightIndex] = TRUE;
	    glEnable(l);
	}
    }

    /* Update the colors of the light sources. */
    for (i = 0; i < IG_MAX_LIGHT_SOURCES; i++) {
	if (LightSourceActive[i]) {
	    int j;
	    GLfloat Clr[4];

	    Clr[3] = 0.0;
	    l = GL_LIGHT0 + i;

	    glLightfv(l, GL_POSITION, LightSourcePos[i]);

	    for (j = 0; j < 3; j++)
		Clr[j] = (GLfloat) (LightSourceClr[i][j] *
				    IGShadeParam.LightAmbient[j]);
	    glLightfv(l, GL_AMBIENT, Clr);

	    for (j = 0; j < 3; j++)
		Clr[j] = (GLfloat) (LightSourceClr[i][j] *
				    IGShadeParam.LightDiffuse[j]);
	    glLightfv(l, GL_DIFFUSE, Clr);

	    for (j = 0; j < 3; j++)
		Clr[j] = (GLfloat) (LightSourceClr[i][j] *
				    IGShadeParam.LightSpecular[j]);
	    glLightfv(l, GL_SPECULAR, Clr);
	}
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Prepares the texture mapping function of an object.                      M
*                                                                            *
* PARAMETERS:                                                                M
*   PObj:    Object to apply texture to.                                     M
*                                                                            *
* RETURN VALUE:                                                              M
*   int:    TRUE if succesful, FALSE otherwise.                              M
*                                                                            *
* KEYWORDS:                                                                  M
*   IGSetTexture                                                             M
*****************************************************************************/
int IGSetTexture(IPObjectStruct *PObj)
{
    int TexID;

#ifdef HAVE_OGL_CG_LIB
    PFNGLACTIVETEXTUREARBPROC
        glActiveTextureARB = (PFNGLACTIVETEXTUREARBPROC)
	    wglGetProcAddress("glActiveTextureARB");
#endif /* HAVE_OGL_CG_LIB */

    if (PObj == NULL) {
        glDisable(GL_TEXTURE_2D);
	GlblTextureActive = FALSE;
	return FALSE;
    }

    if ((TexID = AttrGetObjectIntAttrib(PObj, "_ImageTexID")) > 0) {
	glEnable(GL_TEXTURE_2D);
	GlblTextureActive = TRUE;
#ifdef HAVE_OGL_CG_LIB
/*	glActiveTextureARB(GL_TEXTURE0_ARB); */
#endif /* HAVE_OGL_CG_LIB */
	glBindTexture(GL_TEXTURE_2D, TexID);    
      	return TRUE;
    }
    else {
	IrtBType *Pwr2Image,
	    *Image = (IrtBType *) AttrGetObjectPtrAttrib(PObj, "_ImageTexture");
	int PwrX, PwrY,
	    Width = AttrGetObjectIntAttrib(PObj, "_ImageWidth"),
	    Height = AttrGetObjectIntAttrib(PObj, "_ImageHeight"),
	    Alpha = AttrGetObjectIntAttrib(PObj, "_ImageAlpha");

	if (Image == NULL)
	    return FALSE;

	/* Find the smallest power of two larger/equal to both axes. */
	for (PwrX = 1; PwrX <= Width; PwrX <<= 1);
	if (PwrX != Width)
	    PwrX >>= 1;
	for (PwrY = 1; PwrY <= Height; PwrY <<= 1);
	if (PwrY != Height)
	    PwrY >>= 1;

	while (PwrX > GL_MAX_TEXTURE_SIZE || PwrY > GL_MAX_TEXTURE_SIZE) {
	    PwrX >>= 1;
	    PwrY >>= 1;
	}

        /* Padding of 4 due to a bug in gluScaleImage six lines below. */
        Pwr2Image = (IrtBType *) IritMalloc(sizeof(IrtBType) * (Alpha ? 4 : 3)
					    * PwrX * PwrY + 1000);

        /* Create a power of two image size out of the input image. */
#ifdef __WINCE__
        {
#else
        if (gluScaleImage(Alpha ? GL_RGBA : GL_RGB, Width, Height,
			  GL_UNSIGNED_BYTE, Image, PwrX, PwrY,
			  GL_UNSIGNED_BYTE, Pwr2Image) != 0) {
#endif /* __WINCE__ */
	    /* Do the best you can with the input image. */
  	    PwrX = Width;
	    PwrY = Height;
	    IritFree(Pwr2Image);
	    Pwr2Image = (IrtBType *) Image;
        }

	/* Get a unique texture ID. */
	glEnable(GL_TEXTURE_2D);
	GlblTextureActive = TRUE;
	glGenTextures(1, (GLuint *) &TexID);

	glBindTexture(GL_TEXTURE_2D, TexID);    
	AttrSetObjectIntAttrib(PObj, "_ImageTexID", TexID);

	glTexImage2D(GL_TEXTURE_2D, 0, Alpha ? GL_RGBA : GL_RGB,
		     PwrX, PwrY, 0, Alpha ? GL_RGBA : GL_RGB,
		     GL_UNSIGNED_BYTE, Pwr2Image);

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
      	return TRUE;
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
* Sets the color according to the given RGB values.			     M
*                                                                            *
* PARAMETERS:                                                                M
*   Color:      An RGB vector of integer values between 0 and 255.           M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   IGSetColorRGB                                                            M
*****************************************************************************/
void IGSetColorRGB(int Color[3])
{
    if (IGGlblDrawStyle) {
	int i;
	GLfloat MatAmbient[4], MatDiffuse[4], MatSpecular[4], MatEmissive[4];

	for (i = 0; i < 3; i++) {
	    MatAmbient[i] = (GLfloat) (IGShadeParam.LightAmbient[i]
 							   * Color[i] / 255.0);
	    MatDiffuse[i] = (GLfloat) (IGShadeParam.LightDiffuse[i]
							   * Color[i] / 255.0);
	    MatSpecular[i] = (GLfloat) IGShadeParam.LightSpecular[i],
	    MatEmissive[i] = (GLfloat) IGShadeParam.LightEmissive[i];
	}
	MatAmbient[3] = MatDiffuse[3] = MatSpecular[3] = MatEmissive[3] = 1.0;

	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, MatAmbient);
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, MatDiffuse);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, MatSpecular);
	glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, MatEmissive);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, &IGShadeParam.Shininess);
    }

    if (IGGlblOpacity < 1.0)
        glColor4f((GLfloat) (Color[0] / 255.0),
		  (GLfloat) (Color[1] / 255.0),
		  (GLfloat) (Color[2] / 255.0),
		  (GLfloat) IGGlblOpacity);
    else
        glColor3f((GLfloat) (Color[0] / 255.0),
		  (GLfloat) (Color[1] / 255.0),
		  (GLfloat) (Color[2] / 255.0));
}

/*****************************************************************************
* DESCRIPTION:                                                               M
* Sets the color of an object according to its color/rgb attributes.	     M
*   If object has an RGB attribute it will be used. Otherwise, if the object M
* has a COLOR attribute it will use. Otherwise, WHITE will be used.	     M
*                                                                            *
* PARAMETERS:                                                                M
*   PObj:      To set the drawing color to its color.                        M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   IGSetColorObj                                                            M
*****************************************************************************/
void IGSetColorObj(IPObjectStruct *PObj)
{
    int c, Color[3];

    if (AttrGetObjectRGBColor(PObj, &Color[0], &Color[1], &Color[2])) {
	IGSetColorRGB(Color);
    }
    else if ((c = AttrGetObjectColor(PObj)) != IP_ATTR_NO_COLOR) {
	SetColorIndex(c);
    }
    else {
	/* No default color: */
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               *
* Sets the color according to the given color index.		    	     *
*                                                                            *
* PARAMETERS:                                                                *
*   color:     Index of color to use. Must be between 0 and IG_MAX_COLOR.    *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void SetColorIndex(int color)
{
    int Color[3];

    if (color < 0 || color > IG_MAX_COLOR)
        color = IG_IRIT_WHITE;

    Color[0] = AttrIritColorTable[color][0];
    Color[1] = AttrIritColorTable[color][1];
    Color[2] = AttrIritColorTable[color][2];

    IGSetColorRGB(Color);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
* Sets the line width to draw the given object, in pixels.		     M
*                                                                            *
* PARAMETERS:                                                                M
*   Width:    In pixels of lines to draw with.                               M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   IGSetWidthObj                                                            M
*****************************************************************************/
void IGSetWidthObj(int Width)
{
    glLineWidth((GLfloat) Width);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
* Sets the line pattern to draw the given object.			     M
*                                                                            *
* PARAMETERS:                                                                M
*   Pattern:    The line pattern to use.	                             M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   IGSetLinePattern                                                         M
*****************************************************************************/
void IGSetLinePattern(int Pattern)
{
}

/*****************************************************************************
* DESCRIPTION:                                                               M
* Sets the opacity level to draw the given object.			     M
*                                                                            *
* PARAMETERS:                                                                M
*   Transparency:    Level of transparency - 1 full, 0 opaque.               M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   IGSetTranspObj                                                           M
*****************************************************************************/
void IGSetTranspObj(IrtRType Transparency)
{

    if (Transparency > 0.0) {
	glEnable(GL_BLEND);
	glDepthMask(GL_FALSE);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	IGGlblOpacity = 1.0 - IRIT_BOUND(Transparency, 0.0, 1.0);
    }
    else {
        IGGlblOpacity = 1.0;

	if ((IGGlblAntiAliasing == IG_STATE_ANTI_ALIAS_ON &&
	     IGGlblDrawStyle == IG_STATE_DRAW_STYLE_WIREFRAME) ||
	    (IGGlblAntiAliasing == IG_STATE_ANTI_ALIAS_BLEND &&
	     IGGlblDrawStyle == IG_STATE_DRAW_STYLE_SOLID)) {
	    glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	    glEnable(GL_LINE_SMOOTH);
	    glEnable(GL_POINT_SMOOTH);
	    glEnable(GL_BLEND);
	}
	else {
	    glBlendFunc(GL_ONE, GL_ZERO);
	    glDisable(GL_LINE_SMOOTH);
	    glDisable(GL_POINT_SMOOTH);
	    glDisable(GL_BLEND);
	}

	glDepthMask(GL_TRUE);
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               *
*   Handle the positional as well as possible RGB attribute of a vertex.     *
*                                                                            *
* PARAMETERS:                                                                *
*   V:      Vertex to process.						     *
*                                                                            *
* RETURN VALUE:                                                              *
*   void                                                                     *
*****************************************************************************/
static void HandleVertex(IPVertexStruct *V) 
{ 
    const char *Str;
    int Color[3];
    float *Uv;

    if ((Str = AttrGetStrAttrib(V -> Attr, "RGB")) != NULL &&
	sscanf(Str, "%d,%d,%d", &Color[0], &Color[1], &Color[2]) == 3)
        glColor3ub((GLubyte) Color[0], (GLubyte) Color[1], (GLubyte) Color[2]);

    if (GlblTextureActive &&
	(Uv = AttrGetUVAttrib(V -> Attr, "uvvals")) != NULL)
        glTexCoord2f((GLfloat) Uv[0], (GLfloat) Uv[1]);

    OGL_GL_VRTX_V(V -> Coord);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
* Draw a single Point/Vector object using current modes and transformations. M
*                                                                            *
* PARAMETERS:                                                                M
*   PObj:     A point/vector object to draw.                                 M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   IGDrawPtVec                                                              M
*****************************************************************************/
void IGDrawPtVec(IPObjectStruct *PObj)
{
    int i;
    IrtRType
	*Pt = PObj -> U.Pt;

    glDisable(GL_LIGHTING);

    if (IGGlblDrawStyle == IG_STATE_DRAW_STYLE_POINTS) {
        float FPt[3];

	FPt[0] = (float) Pt[0];
	FPt[1] = (float) Pt[1];
	FPt[2] = (float) Pt[2];

        glVertexPointer(3, GL_FLOAT, 0, FPt);	

	glEnableClientState(GL_POINTS);
	glDrawArrays(GL_LINES, 0, 1);
	glDisableClientState(GL_POINTS);
    }
    else {
        static float
	    Zero[3] = { 0.0, 0.0, 0.0 };
        float Ends[2][3];

	for (i = 0; i < 3; i++) {
	    Ends[0][0] = Ends[1][0] = (float) Pt[0];
	    Ends[0][1] = Ends[1][1] = (float) Pt[1];
	    Ends[0][2] = Ends[1][2] = (float) Pt[2];

	    Ends[0][i] -= (float) IGGlblPointSize;
	    Ends[1][i] += (float) IGGlblPointSize;

	    glVertexPointer(3, GL_FLOAT, 0, Ends);	

	    glEnableClientState(GL_VERTEX_ARRAY);
	    glDrawArrays(GL_LINES, 0, 2);
	    glDisableClientState(GL_VERTEX_ARRAY);
	}

	if (IP_IS_VEC_OBJ(PObj)) {
	    Ends[0][0] = (float) Pt[0];
	    Ends[0][1] = (float) Pt[1];
	    Ends[0][2] = (float) Pt[2];
	    IRIT_GEN_COPY(Ends[1], Zero, sizeof(float) * 3);

	    glVertexPointer(3, GL_FLOAT, 0, Ends);	

	    glEnableClientState(GL_VERTEX_ARRAY);
	    glDrawArrays(GL_LINES, 0, 2);
	    glDisableClientState(GL_VERTEX_ARRAY);
	}
    }

    if (IGGlblDrawStyle == IG_STATE_DRAW_STYLE_SOLID)
	glEnable(GL_LIGHTING);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
* Draw a single Poly object using current modes and transformations.	     M
*                                                                            *
* PARAMETERS:                                                                M
*   PObj:     A poly object to draw.                                         M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   IGDrawPoly                                                               M
*****************************************************************************/
void IGDrawPoly(IPObjectStruct *PObj)
{
    IPVertexStruct *V;
    IPPolygonStruct
	*Pl = PObj -> U.Pl;

    if (Pl == NULL)
        return;

    if (IP_IS_POLYLINE_OBJ(PObj)) {
        glDisable(GL_LIGHTING);
	for (; Pl != NULL; Pl = Pl -> Pnext) {
	    if (IGGlblCountNumPolys)
		IGGlblNumPolys++;

	    if (IGGlblDrawStyle == IG_STATE_DRAW_STYLE_POINTS)
	        glBegin(GL_POINTS);
	    else
	        glBegin(GL_LINE_STRIP);
	    for (V = Pl -> PVertex; V != NULL; V = V -> Pnext) {
	        HandleVertex(V);
	    }
	    glEnd();
	}
	if (IGGlblDrawStyle == IG_STATE_DRAW_STYLE_SOLID)
	    glEnable(GL_LIGHTING);
    }
    else if (IP_IS_POINTLIST_OBJ(PObj)) {
        glDisable(GL_LIGHTING);
	for (; Pl != NULL; Pl = Pl -> Pnext) {
	    if (IGGlblCountNumPolys)
		IGGlblNumPolys++;

	    if (IGGlblDrawStyle == IG_STATE_DRAW_STYLE_POINTS) {
	        glBegin(GL_POINTS);
	        for (V = Pl -> PVertex; V != NULL; V = V -> Pnext) {
		    HandleVertex(V);
		}
		glEnd();
	    }
	    else {
	        for (V = Pl -> PVertex; V != NULL; V = V -> Pnext) {
		    int i;
		    IrtPtType Ends[6];
		    IrtRType
		        *Pt = V -> Coord;

		    for (i = 0; i < 6; i++)
		        IRIT_PT_COPY(Ends[i], Pt);

		    Ends[0][0] -= IGGlblPointSize;
		    Ends[1][0] += IGGlblPointSize;
		    Ends[2][1] -= IGGlblPointSize;
		    Ends[3][1] += IGGlblPointSize;
		    Ends[4][2] -= IGGlblPointSize;
		    Ends[5][2] += IGGlblPointSize;

		    for (i = 0; i < 6; i += 2) {
		        glBegin(GL_LINE_STRIP);
			OGL_GL_VRTX_V(Ends[i]);
			OGL_GL_VRTX_V(Ends[i + 1]);
			glEnd();
		    }
		}
	    }
	}
	if (IGGlblDrawStyle == IG_STATE_DRAW_STYLE_SOLID)
	    glEnable(GL_LIGHTING);
    }
    else if (IP_IS_POLYGON_OBJ(PObj) || IP_SET_POLYSTRIP_OBJ(PObj)) {
	int PolysFromSrf = AttrGetObjectIntAttrib(PObj, "_srf_polys") == TRUE;
	IrtPtType PNormal, VNormal;

	if (!((PolysFromSrf && !IGGlblDrawSurfacePoly) ||
	      (!PolysFromSrf && !IGGlblDrawPolygons))) {
	    if (IGGlblShadingModel == IG_SHADING_BACKGROUND) {
		glColor3d(IGGlblBackGroundColor[0] / 255.0,
			  IGGlblBackGroundColor[1] / 255.0,
			  IGGlblBackGroundColor[2] / 255.0);
		glDisable(GL_LIGHTING);
	    }
	    else if (IGGlblShadingModel == IG_SHADING_NONE) {
		glDisable(GL_LIGHTING);
	    }
	    else
	        glEnable(GL_LIGHTING);

	    glMatrixMode(GL_PROJECTION);
	    glPushMatrix();
	    glTranslated(0.0, 0.0, -IG_POLYGON_Z_TRANS);
	
	    for (; Pl != NULL; Pl = Pl -> Pnext) {
		int PlDrawn = TRUE;

		if (IGGlblCountNumPolys)
		    IGGlblNumPolys++;

		if (IGGlblDrawStyle == IG_STATE_DRAW_STYLE_SOLID) {
		    glBegin(GL_POLYGON);
		    switch (IGGlblShadingModel) {
		        case IG_SHADING_FLAT:
			    OGL_HANDLE_NORMAL(Pl -> Plane);
			case IG_SHADING_BACKGROUND:
			case IG_SHADING_NONE:
			    for (V = Pl -> PVertex;
				 V != NULL;
				 V = V -> Pnext) {
				HandleVertex(V);
			    }
			    break;
			case IG_SHADING_GOURAUD:
			case IG_SHADING_PHONG:
			    for (V = Pl -> PVertex;
				 V != NULL;
				 V = V -> Pnext) {
				OGL_HANDLE_NORMAL(V -> Normal);
				HandleVertex(V);
			    }
		    }
		    glEnd();
		}
		else if (IGGlblDrawStyle == IG_STATE_DRAW_STYLE_WIREFRAME) {
		    if (IGGlblBackFaceCull) {
			IrtRType P[3];

			MatMultVecby4by4(P, Pl -> Plane, IPViewMat);
			if ((P[2] > 0.0) ^ IGGlblFlipNormalOrient)
			    PlDrawn = FALSE;
		    }

		    if (PlDrawn) {
			glBegin(GL_LINE_STRIP);
			for (V = Pl -> PVertex; V != NULL; V = V -> Pnext) {
			    HandleVertex(V);

			    if (IP_IS_INTERNAL_VRTX(V) && !IGGlblDrawInternal) {
				glEnd();
				glBegin(GL_LINE_STRIP);
			    }
			}
		    }
		    HandleVertex(Pl -> PVertex);
		    glEnd();
		}
		else if (IGGlblDrawStyle == IG_STATE_DRAW_STYLE_POINTS) {
		    glBegin(GL_POINTS);
		    for (V = Pl -> PVertex; V != NULL; V = V -> Pnext) {
		        HandleVertex(V);
		    }
		    glEnd();
		}

		if (PlDrawn && IGGlblDrawPNormal && IP_HAS_PLANE_POLY(Pl)) {
		    IrtRType R;
		    int NumOfVertices = 1;

		    V = Pl -> PVertex;
		    IRIT_PT_COPY(PNormal, V -> Coord);

		    for (V = V -> Pnext; V != NULL; V = V -> Pnext) {
			IRIT_PT_ADD(PNormal, PNormal, V -> Coord);
			NumOfVertices++;
		    }

		    R = 1.0 / NumOfVertices;
		    IRIT_PT_SCALE(PNormal, R);
		    glBegin(GL_LINES);
		    OGL_GL_VRTX_V(PNormal);
		    IG_ADD_ORIENT_NRML(PNormal, Pl -> Plane);
		    OGL_GL_VRTX_V(PNormal);
		    glEnd();
		}

		if (PlDrawn && IGGlblDrawVNormal) {
		    for (V = Pl -> PVertex; V != NULL; V = V -> Pnext) {
			if (IP_HAS_NORMAL_VRTX(V)) {
			    IRIT_VEC_COPY(VNormal, V -> Coord);
			    IG_ADD_ORIENT_NRML(VNormal, V -> Normal);
			    glBegin(GL_LINES);
			    OGL_GL_VRTX_V(V -> Coord);
			    OGL_GL_VRTX_V(VNormal);
			    glEnd();
			}
		    }
		}
	    }

	    glPopMatrix();
	    glMatrixMode(GL_MODELVIEW);

	    glDisable(GL_LIGHTING);
	    IGSetColorObj(PObj);
	}

	if (IGGlblDrawSurfaceSketch && !PolysFromSrf)
	    IGDrawPolygonSketches(PObj);

	if (IGGlblDrawSurfaceSilh || IGGlblDrawSurfaceBndry)
	    IGDrawPolySilhBndry(PObj);

	if (IGGlblDrawStyle == IG_STATE_DRAW_STYLE_SOLID)
	    glEnable(GL_LIGHTING);
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
* Draw a single Poly object using current modes and transformations.	     M
*   This function allows polylines to have normals so we can actually draw   M
* the adaptive isolines to render a surface.				     M
*                                                                            *
* PARAMETERS:                                                                M
*   PObj:         A polygon or a polyline to draw.                           M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   IGDrawPolylineNormal                                                     M
*****************************************************************************/
void IGDrawPolylineNormal(IPObjectStruct *PObj)
{
    IPPolygonStruct
	*Pl = PObj -> U.Pl;

    if (IP_IS_POLYLINE_OBJ(PObj)) {
	for (; Pl != NULL; Pl = Pl -> Pnext) {
	    IPVertexStruct *V;

	    glBegin(GL_LINE_STRIP);
	    for (V = Pl -> PVertex; V != NULL; V = V -> Pnext) {
		OGL_HANDLE_NORMAL(V -> Normal);
		OGL_GL_VRTX_V(V -> Coord);
	    }
	    glEnd();
	}
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
* Draw a single object using current modes and transformations.		     M
*                                                                            *
* PARAMETERS:                                                                M
*   PObj:      Object to draw.                                               M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   IGDrawObject                                                             M
*****************************************************************************/
void IGDrawObject(IPObjectStruct *PObj)
{
    CagdRType *R;
    IrtPtType Pt;

    int HasTexture = FALSE,
	Width = AttrGetObjectIntAttrib(PObj, "DWidth"),
	LPattern = AttrGetObjectIntAttrib(PObj, "LnPtrn");
    IrtRType
	Transp = AttrGetObjectRealAttrib(PObj, "Transp");

    if (IP_ATTR_IS_BAD_INT(Width))
        Width = 1;

    if (!IP_ATTR_IS_BAD_REAL(Transp))
	IGSetTranspObj(Transp);

    if (!IP_ATTR_IS_BAD_INT(LPattern))
        IGSetLinePattern(LPattern);

    IGSetWidthObj(Width * IGGlblLineWidth);
    IGSetColorObj(PObj);

    if (IG_IS_HIGHLIGHT1_OBJ(PObj)) {
        IGSetWidthObj(Width * 2 * IGGlblLineWidth);
	IGSetColorRGB(IGGlblHighlight1Color);
    }
    else if (IG_IS_HIGHLIGHT2_OBJ(PObj)) {
        IGSetWidthObj(Width * 2 * IGGlblLineWidth);
	IGSetColorRGB(IGGlblHighlight2Color);
    }
    else if ((HasTexture = IGInitSrfTexture(PObj)) != FALSE &&
	     IGSetTexture(PObj)) {
    }
#   ifdef HAVE_OGL_CG_LIB
    else if (AttrGetObjectStrAttrib(PObj, "DTexture") != NULL &&
	     IGCGDrawDTexture(PObj))
		    return;
    else if (AttrGetObjectStrAttrib(PObj, "ffd_texture") != NULL &&
	     IGCGFfdDraw(PObj))
        return;
#   endif /* HAVE_OGL_CG_LIB */

    switch (PObj -> ObjType) {
	case IP_OBJ_POLY:
	    IGDrawPoly(PObj);
	    break;
	case IP_OBJ_CTLPT:
	    /* Coerce, in place, a control points to a regular point. */
	    R = PObj -> U.CtlPt.Coords;
	    CagdCoercePointTo(Pt, CAGD_PT_E3_TYPE,
			      &R, -1, PObj -> U.CtlPt.PtType);
	    IRIT_PT_COPY(PObj -> U.Pt, Pt);
	    PObj -> ObjType = IP_OBJ_POINT;
	case IP_OBJ_POINT:
	    if (!IP_ATTR_IS_BAD_INT(AttrGetObjectIntAttrib(PObj,
							   "LIGHT_SOURCE"))) {
		if (IP_ATTR_IS_BAD_INT(AttrGetObjectIntAttrib(PObj,
							      "ACTIVATED"))) {
		    const char *p;
		    int i, Red, Green, Blue,
		        LightIndex = -1;
		    IGLightType LightPos;
		    IrtVecType LightColor;

		    for (i = 0; i < 3; i++)
			LightPos[i] = (float) PObj -> U.Pt[i];
		    LightPos[3] = (float)
			!((p = AttrGetObjectStrAttrib(PObj, "TYPE")) != NULL &&
			  stricmp(p, "POINT_INFTY"));
		    
		    if (AttrGetObjectRGBColor(PObj, &Red, &Green, &Blue)) {
			LightColor[0] = Red / 255.0;
			LightColor[1] = Green / 255.0;
			LightColor[2] = Blue/ 255.0;
		    }
		    else {
			LightColor[0] = LightColor[1] = LightColor[2] = 1.0;
		    }

		    i = AttrGetObjectIntAttrib(PObj, "INDEX");
		    if (!IP_ATTR_IS_BAD_INT(i))
			LightIndex = i;
			
		    IGSetLightSource(LightPos, LightColor, LightIndex);

		    AttrSetObjectIntAttrib(PObj, "ACTIVATED", TRUE);
		}
	    }
	    else
		IGDrawPtVec(PObj);
	    break;
	case IP_OBJ_VECTOR:
	    IGDrawPtVec(PObj);
	    break;
	case IP_OBJ_STRING:
	    IGDrawString(PObj);
	    break;
	case IP_OBJ_CURVE:
	    if (PObj -> U.Crvs == NULL)    /* Can occur in the curve editor. */
	        break;
	    if (CAGD_IS_POWER_CRV(PObj -> U.Crvs)) {
	        CagdCrvStruct
		    *Crv = CagdCnvrtPwr2BzrCrv(PObj -> U.Crvs);

		CagdCrvFree(PObj -> U.Crvs);
		PObj -> U.Crvs = Crv;
	    }
	    if (CAGD_NUM_OF_PT_COORD(PObj -> U.Crvs -> PType) == 1) {
	        CagdCrvStruct
		    *Crv = CagdCoerceCrvTo(PObj -> U.Crvs,
			CAGD_IS_RATIONAL_CRV(PObj -> U.Crvs) ? CAGD_PT_P2_TYPE
							     : CAGD_PT_E2_TYPE,
					   FALSE);

		CagdCrvFree(PObj -> U.Crvs);
		PObj -> U.Crvs = Crv;
	    }

	    IGDrawCurve(PObj);
	    break;
	case IP_OBJ_TRIMSRF:
	    if (CAGD_IS_POWER_SRF(PObj -> U.TrimSrfs -> Srf)) {
	        CagdSrfStruct
		    *Srf = CagdCnvrtPwr2BzrSrf(PObj -> U.TrimSrfs -> Srf);

		CagdSrfFree(PObj -> U.TrimSrfs -> Srf);
		PObj -> U.TrimSrfs -> Srf = Srf;
	    }

	    if (PObj -> U.TrimSrfs -> TrimCrvList == NULL) {
	        CagdSrfStruct
		    *Srf = PObj -> U.TrimSrfs -> Srf;

	        if (IGGlblMore)
		    IRIT_INFO_MSG_PRINTF("Trimmed surface \"%s\" with no trimming curves detected\n\tand converted to regular surface\n",
			                 IP_GET_OBJ_NAME(PObj));
		
		PObj -> U.TrimSrfs -> Srf = NULL;
		TrimSrfFree(PObj -> U.TrimSrfs);
		PObj -> U.Srfs = Srf;
		PObj -> ObjType = IP_OBJ_SURFACE;
	    }
	    else {
	        IGDrawTrimSrf(PObj);
		break;
	    }
	case IP_OBJ_SURFACE:
	    if (PObj -> U.Srfs == NULL)  /* Can occur in the surface editor. */
	        break;
	    if (CAGD_IS_POWER_SRF(PObj -> U.Srfs)) {
	        CagdSrfStruct
		    *Srf = CagdCnvrtPwr2BzrSrf(PObj -> U.Srfs);

		CagdSrfFree(PObj -> U.Srfs);
		PObj -> U.Srfs = Srf;
	    }
	    if (CAGD_NUM_OF_PT_COORD(PObj -> U.Srfs -> PType) == 1) {
	        CagdSrfStruct
		    *Srf = CagdCoerceSrfTo(PObj -> U.Srfs,
			CAGD_IS_RATIONAL_SRF(PObj -> U.Srfs) ? CAGD_PT_P2_TYPE
							     : CAGD_PT_E2_TYPE,
					     FALSE);

		CagdSrfFree(PObj -> U.Srfs);
		PObj -> U.Srfs = Srf;
	    }

	    IGDrawSurface(PObj);
	    break;
	case IP_OBJ_TRIVAR:
	    IGDrawTrivar(PObj);
	    break;
	case IP_OBJ_MULTIVAR:
	    if (PObj -> U.MultiVars -> Dim < 4) { /* Curve, Surface, Trivar. */
		IPObjectStruct *PTmp;

		if ((PTmp = AttrGetObjectObjAttrib(PObj,
						   "_Coerced")) == NULL) {
		    MvarMVStruct
			*MV = PObj -> U.MultiVars;

		    switch (MV -> Dim) {
			case 1:
			    PTmp = IPGenCRVObject(MvarMVToCrv(MV));

			    if (CAGD_IS_POWER_CRV(PTmp -> U.Crvs)) {
			        CagdCrvStruct
				    *Crv = CagdCnvrtPwr2BzrCrv(PTmp -> U.Crvs);

				CagdCrvFree(PTmp -> U.Crvs);
				PTmp -> U.Crvs = Crv;
			    }
			    break;
			case 2:
			    PTmp = IPGenSRFObject(MvarMVToSrf(MV));

			    if (CAGD_IS_POWER_SRF(PTmp -> U.Srfs)) {
			        CagdSrfStruct
				    *Srf = CagdCnvrtPwr2BzrSrf(PTmp -> U.Srfs);

				CagdSrfFree(PTmp -> U.Srfs);
				PTmp -> U.Srfs = Srf;
			    }
			    break;
			case 3:
			    PTmp = IPGenTRIVARObject(MvarMVToTV(MV));
			    break;
		    }
		    PTmp -> Attr = IP_ATTR_COPY_ATTRS(PObj -> Attr);
		    IP_SET_OBJ_NAME2(PTmp, IP_GET_OBJ_NAME(PObj));
		    IGUpdateObjectBBox(PTmp);

		    AttrSetObjectObjAttrib(PObj, "_Coerced", PTmp, FALSE);
		}

		if (IG_IS_HIGHLIGHT1_OBJ(PObj))
		    IG_SET_HIGHLIGHT1_OBJ(PTmp);
		else
		    IG_RST_HIGHLIGHT1_OBJ(PTmp);
		if (IG_IS_HIGHLIGHT2_OBJ(PObj))
		    IG_SET_HIGHLIGHT2_OBJ(PTmp);
		else
		    IG_RST_HIGHLIGHT2_OBJ(PTmp);

		IGDrawObject(PTmp);
	    }
	    break;
	case IP_OBJ_MODEL:
	    IGDrawModel(PObj);
	    break;
	case IP_OBJ_TRISRF:
	    if (TRNG_IS_GREGORY_TRISRF(PObj -> U.TriSrfs)) {
	        TrngTriangSrfStruct
		    *TriSrf = TrngCnvrtGregory2BzrTriSrf(PObj -> U.TriSrfs);

	        TrngTriSrfFree(PObj -> U.TriSrfs);
	        PObj -> U.TriSrfs = TriSrf;
	    }
	    IGDrawTriangSrf(PObj);
	    break;
	case IP_OBJ_LIST_OBJ:
	    IRIT_FATAL_ERROR("Should not have lists at this time");
	    break;
	default:
	    break;
    }

    if (HasTexture)
        IGSetTexture(NULL);	                 /* Disable texture mapping. */

    if (!IP_ATTR_IS_BAD_REAL(Transp))
	IGSetTranspObj(0.0);			    /* Disable transparency. */
}
