/******************************************************************************
* RealSpinCtrl.cpp - Holds numeric entry control.			      *
* This class implements spin control for real values.			      *
*******************************************************************************
* (C) Gershon Elber					                      *
*******************************************************************************
* Written by Gershon Elber, Dec 2006.					      *
******************************************************************************/

#include "NCSimMain.h"
#include "RealSpinCtrl.h"

#define REAL_SPIN_CTRL_BG	220, 255, 160

/*****************************************************************************
* AUXILIARY:								     *
*    Event table for RealSpinCtrlClass					     *
*****************************************************************************/
BEGIN_EVENT_TABLE(RealSpinCtrlClass, wxTextCtrl)
    EVT_TEXT(wxID_ANY, OnTextChange)
    EVT_RIGHT_DOWN(OnMouseEvent)
    EVT_RIGHT_UP(OnMouseEvent)
END_EVENT_TABLE()

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Constructor of the real value spin class.	                             M
*                                                                            *
* PARAMETERS:                                                                M
*   ValueAddr:     Initial value and also location to update on the fly,     M
*		   and as long as the value is valid.			     M
*   MinAllowedValue, MaxAllowedValue:  Domain to consider the input valid.   M
*   HIncrement, VIncrement:  Amounts to inc/decrement the real input value,  M
*		   when the mouse is moving horizontally/vertically.         M
*   Parent:        Parent window.				             M
*   Id:            ID assigned to window.	                             M
*   Pos:	   Default position.					     M
*   Size:	   Default size.					     M
*   Flags:         Window styles' flags.			             M
*                                                                            *
* RETURN VALUE:                                                              M
*   NoValue                                                                  M
*                                                                            *
* KEYWORDS:                                                                  M
*   RealSpinCtrlClass::RealSpinCtrlClass                                     M
*****************************************************************************/
RealSpinCtrlClass::RealSpinCtrlClass(double *ValueAddr,
				     double MinAllowedValue,
				     double MaxAllowedValue,
				     double HIncrement,
				     double VIncrement,
				     wxWindow *Parent,
				     wxWindowID Id,
				     const wxPoint &Pos,
				     const wxSize &Size,
				     int Flags):
	ValAddr(ValueAddr),
	Val(*ValueAddr),
	UpdateValFunc(NULL),
	MinAllowedVal(MinAllowedValue),
	MaxAllowedVal(MaxAllowedValue),
	HInc(HIncrement),
	VInc(VIncrement),
	MouseCaptured(FALSE),
	wxTextCtrl(Parent,
		   Id,
		   wxT(""),
		   Pos,
		   Size,
		   Flags | wxTE_RICH | wxTE_RICH2)
{
    wxString Line;

    SetBackgroundColour(wxColour(REAL_SPIN_CTRL_BG));

    WinId = Id;

    Line.Printf(wxT("%g"), *ValueAddr);
    SetValue(Line);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Constructor of the real value spin class.	                             M
*                                                                            *
* PARAMETERS:                                                                M
*   Value:         Initial value.					     M
*   UpdateValueFunc:  A pointer to a function to update te real value when   M
*		   changed.					             M
*   MinAllowedValue, MaxAllowedValue:  Domain to consider the input valid.   M
*   HIncrement, VIncrement:  Amounts to inc/decrement the real input value,  M
*		   when the mouse is moving horizontally/vertically.         M
*   Parent:        Parent window.				             M
*   Id:            ID assigned to window.	                             M
*   Pos:	   Default position.					     M
*   Size:	   Default size.					     M
*   Flags:         Window styles' flags.			             M
*                                                                            *
* RETURN VALUE:                                                              M
*   NoValue                                                                  M
*                                                                            *
* KEYWORDS:                                                                  M
*   RealSpinCtrlClass::RealSpinCtrlClass                                     M
*****************************************************************************/
RealSpinCtrlClass::RealSpinCtrlClass(double Value,
				     RealSpinCtrlUpdateValFunc UpdateValueFunc,
				     double MinAllowedValue,
				     double MaxAllowedValue,
				     double HIncrement,
				     double VIncrement,
				     wxWindow *Parent,
				     wxWindowID Id,
				     const wxPoint &Pos,
				     const wxSize &Size,
				     int Flags):
	ValAddr(NULL),
	UpdateValFunc(UpdateValueFunc),
	MinAllowedVal(MinAllowedValue),
	MaxAllowedVal(MaxAllowedValue),
	Val(Value),
	HInc(HIncrement),
	VInc(VIncrement),
	MouseCaptured(FALSE),
	wxTextCtrl(Parent,
		   Id,
		   wxT(""),
		   Pos,
		   Size,
		   Flags)
{
    wxString Line;

    SetBackgroundColour(wxColour(REAL_SPIN_CTRL_BG));

    WinId = Id;

    Line.Printf(wxT("%g"), Value);
    SetValue(Line);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Destructor of the real value spin class.                                 M
*                                                                            *
* PARAMETERS:                                                                M
*   None					                             M
*                                                                            *
* RETURN VALUE:                                                              M
*   NoValue                                                                  M
*                                                                            *
* KEYWORDS:                                                                  M
*   RealSpinCtrlClass::~RealSpinCtrlClass                                    M
*****************************************************************************/
RealSpinCtrlClass::~RealSpinCtrlClass()
{
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Update current displayed value based on the value we are referencing.    M
*                                                                            *
* PARAMETERS:                                                                M
*   None		                                                     M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   RealSpinCtrlClass::Refresh                                               M
*****************************************************************************/
void RealSpinCtrlClass::Refresh()
{
    if (ValAddr != NULL) {
        wxString Line;

	Line.Printf(wxT("%g"), *ValAddr);
	SetValue(Line);
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Update the current value of the real spin control.                       M
*                                                                            *
* PARAMETERS:                                                                M
*   NewVal:  New value.                                                      M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   RealSpinCtrlClass::SetRealValue                                          M
*****************************************************************************/
void RealSpinCtrlClass::SetRealValue(char *NewVal)
{
    double R;

    if (sscanf(NewVal, "%lf", &R) == 1)
	SetRealValue(R);
    else {
        SetValue(NCSimChars2wxString(NewVal));
	MarkTextValid(FALSE);
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Update the current value of the real spin control.                       M
*                                                                            *
* PARAMETERS:                                                                M
*   NewVal:  New value.                                                      M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   RealSpinCtrlClass::SetRealValue                                          M
*****************************************************************************/
void RealSpinCtrlClass::SetRealValue(double NewVal)
{
    wxString Line;

    Val = NewVal;

    if (Val < MinAllowedVal)
	Val = MinAllowedVal;
    if (Val > MaxAllowedVal)
	Val = MaxAllowedVal;

    Line.Printf(wxT("%g"), Val);

    /* Set new value only if no other update is being performed now.	    */
    if (m_updatesCount == -1 || m_updatesCount == -2)
	SetValue(Line);

    if (ValAddr != NULL)
	*ValAddr = Val;
    if (UpdateValFunc != NULL)
	UpdateValFunc(Val, WinId);

    MarkTextValid(TRUE);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Fetch the current value of the real spin control.                        M
*                                                                            *
* PARAMETERS:                                                                M
*   None			                                             M
*                                                                            *
* RETURN VALUE:                                                              M
*   double:  Current real value of the control.                              M
*                                                                            *
* KEYWORDS:                                                                  M
*   RealSpinCtrlClass::GetRealValue                                          M
*****************************************************************************/
double RealSpinCtrlClass::GetRealValue()
{
    return Val;
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Mark the text as good or bad, based on it number's validity.             M
*                                                                            *
* PARAMETERS:                                                                M
*   Valid:   TRUE for valid value, FALSE for invalid.                        M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   RealSpinCtrlClass::MarkTextValid                                         M
*****************************************************************************/
void RealSpinCtrlClass::MarkTextValid(int Valid)
{
    if (Valid)
        SetStyle(0, GetLastPosition(), wxTextAttr(REAL_SPIN_COLOR_NORMAL));
    else
	SetStyle(0, GetLastPosition(), wxTextAttr(REAL_SPIN_COLOR_HIGHLIGHT));
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Traps events from text changes.                                          M
*                                                                            *
* PARAMETERS:                                                                M
*   Event: N.S.F.I.                                                          M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   RealSpinCtrlClass::OnTextChange                                          M
*****************************************************************************/
void RealSpinCtrlClass::OnTextChange(wxCommandEvent &Event)
{
    double R;
    wxString
	Str = Event.GetString();
    const char
	*s = NCSimwxString2Chars(Str);

    if (sscanf(s, "%lf", &R) != 1) {
	MarkTextValid(FALSE);
    }
    else {
        if (R >= MinAllowedVal && R <= MaxAllowedVal) {
	    Val = R;
	    if (ValAddr != NULL)
	        *ValAddr = Val;
	    if (UpdateValFunc != NULL)
	        UpdateValFunc(Val, WinId);

	    MarkTextValid(TRUE);
	}
	else
	    MarkTextValid(FALSE);
    }
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Traps events from mouse, on dynamic event handling of mouse motion.      M
*                                                                            *
* PARAMETERS:                                                                M
*   Event: N.S.F.I.                                                          M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   RealSpinCtrlClass::OnMouseMotionEvent                                    M
*****************************************************************************/
void RealSpinCtrlClass::OnMouseMotionEvent(wxCommandEvent &Event)
{
    wxMouseEvent
	*MEvent = (wxMouseEvent *) &Event;
    double DVal;
    int Dx, Dy,
	x = MEvent -> GetX(),
	y = MEvent -> GetY();
    bool
	ShiftDown = MEvent -> ShiftDown(),
	CtrlDown = MEvent -> ControlDown(),
	AltDown = MEvent -> AltDown();

    Dx = x - MouseLastX;
    Dy = MouseLastY - y;
	
    MouseLastX = x;
    MouseLastY = y;

    DVal = Dx * HInc + Dy * VInc;
    if (ShiftDown)
	DVal *= 10;
    if (CtrlDown)
	DVal *= 100;
    if (AltDown)
	DVal *= 0.1;

    SetRealValue(Val + DVal);
}

/*****************************************************************************
* DESCRIPTION:                                                               M
*   Traps events from mouse.		                                     M
*                                                                            *
* PARAMETERS:                                                                M
*   Event: N.S.F.I.                                                          M
*                                                                            *
* RETURN VALUE:                                                              M
*   void                                                                     M
*                                                                            *
* KEYWORDS:                                                                  M
*   RealSpinCtrlClass::OnMouseEvent                                          M
*****************************************************************************/
void RealSpinCtrlClass::OnMouseEvent(wxMouseEvent &Event)
{
    int x = Event.GetX(),
	y = Event.GetY();

    /* Capture the mouse event as long as the mouse is down. */
    if (Event.RightIsDown() && !MouseCaptured) {
        CaptureMouse();
	MouseCaptured = TRUE;
        MouseLastX = x;
        MouseLastY = y;
	Connect(wxID_ANY, wxEVT_MOTION,
		wxCommandEventHandler(RealSpinCtrlClass::OnMouseMotionEvent));
    }
    if (MouseCaptured && !Event.RightIsDown()) {
        ReleaseMouse();
	MouseCaptured = FALSE;
	Disconnect(wxID_ANY, wxEVT_MOTION,
		wxCommandEventHandler(RealSpinCtrlClass::OnMouseMotionEvent));
    }
}
